
/* ************************************************************************
 *            MCMAS -- Model Checker for Multi-Agent Systems
 *
 *     Copyright (C) 2007, 2008
 *     Hongyang Qu and Franco Raimondi and Alessio Lomuscio
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/


#define NAME "mcmas"
#define URL "http://www.cs.ucl.ac.uk/staff/f.raimondi/MCMAS/"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <set>
#include <algorithm>
#include <cctype>
#include <ctime>
#include "types.hh"
#include "utilities.hh"
#include "mcmas-driver.hh"
#include "cuddObj.hh"

#define MAXINISTATES 100
#define MAXTRANSITIONS 100
#define MAXSTACKDEPTH 100


using namespace std;

map<string, basic_agent *> *is_agents;
vector<basic_agent *> *agents;
map<string, bool_expression *> *is_evaluation;
bool_expression *is_istates = NULL;
map<string, set<string> > *is_groups;
vector<modal_formula> *is_formulae;

vector<fairness_expression *> *is_fairness;
int obsvars_bdd_length;

map < string, int > options;
// A map to store global options, e.g. verbosity, etc.
// 'verbosity' -> 0 (default) -- 5

string cex_prefix; // Destination directory for counterexamples


bool global_consistency_check();

Cudd *bddmgr; // The manager
BDDvector *v;
BDDvector *vRT;
ADDvector *addv;
BDDvector *pv;
ADDvector *addpv;
BDDvector *a;
BDD andprot;
BDD preRT;
BDD trans;
BDD RT;
BDD reach;
BDD reachRT;
bool calReachRT;
map<string, BDD> er;
map<string, BDD> deor;
map<string, BDD> gm; // For "Everybody in a group knows".
map<string, BDD> deogm; // deontic group modalities
map<string, BDD> distgm; // distributed knowledge modalities
BDD fair_reach;
BDDvector *vevol, *vprot;
BDD in_st;
BDD* bddOne;

vector< vector<int> * > *countex; // For counterexamples
map<int, BDD *> *idbdd; // Map state id -> bdd (for counterexamples again)
int scount; // a global counter for counterexamples.

vector< vector< transition *> *> *cextr; // instead of state id, counterexample with transitions

bool need_epistemic;
bool need_deontic;
bool need_common_knowledge;
bool need_deontic_knowledge;
bool need_distributed_knowledge;

map<string, BDD> *symmetry_cache;
map<string, string> *symmetry_atom_cache;

void print_help( void ) {
  cout << "Usage: " << NAME << " [OPTIONS] FILE " << endl;
  cout << "Example: " << NAME << " -v 3 -bdd_stats myfile.ispl" <<
    endl;
  cout << endl << "Options: " << endl;
  cout << "  -s \t \t Interactive execution" << endl;
  cout << "  -v Number \t verbosity level ( 0 -- 5, default 0 )" << endl;
  cout << "  -e Number \t Choose the approach to do model checking (0 -- 3, default 1)" << endl;
  cout << "  -bc Number \t Store some BDDs to speed up verification (1: RT*reach only, 2: atomic propositions only, 3: both (default))" << endl;
  cout << "  -fe Number \t A new way to encode evolution functions and initial states (1: evolution only, 2: initial state only, 3: both)" << endl;
  cout << "  -if \t \t Encode initial states first" << endl;
  cout << "  -bdd_stats \t Print BDD statistics " << endl;
  cout << "  -src \t \t Print parsed ISPL source code and exit" << endl;  
  cout << "  -gsc \t \t Check the syntax and exit " << endl;
  cout << "  -c \t \t Print counterexamples or witness executions" << endl;
  cout << "  -cex_prefix \t Destination directory for counterexample files" << endl;
  cout << "  -g \t \t Check formulas globally (i.e., not only on initial states)" << endl;
  cout << "  -d \t \t Run in debug mode" << endl;
  cout << "  -h \t \t This screen" << endl;
  cout << endl << "Report bugs to <hongyang.qu@imperial.ac.uk> or <f.raimondi@cs.ucl.ac.uk>" << endl;

}


void print_banner( void ) {
  cout << "********************************************************************************" 
       << endl;
  cout << "                       MCMAS v0.9.6.2 " << endl << endl;
  cout << " This software comes with ABSOLUTELY NO WARRANTY, to the extent" 
       << endl;
  cout << " permited by applicable law. " << endl << endl;
  cout << " Please check " << endl << " " << URL << endl;
  cout << " for the latest release.\n Report bugs to <hongyang.qu@imperial.ac.uk> or <f.raimondi@cs.ucl.ac.uk>" 
       << endl;
  cout << "********************************************************************************" 
       << endl << endl;
}

void print_ispl_code(void) {
  for(unsigned int i=0; i<agents->size(); i++)
    cout << (*agents)[i]->to_string() << endl;
  cout << "Evaluation" << endl;
  for(map<string, bool_expression *>::iterator i=is_evaluation->begin(); i!=is_evaluation->end(); i++)
    cout << "    " << (*i).first << " if " << (*i).second->to_string() << ";" <<endl;
  cout << "end Evaluation" << endl << endl;
  cout << "InitStates" << endl << "    " << is_istates->to_string() << ";" << endl 
       << "end InitStates" << endl << endl;
  cout << "Group" << endl;
  for(map<string,set<string>  >::iterator i=is_groups->begin(); i!=is_groups->end(); i++) {
    cout << "    " << (*i).first << " = { ";
    int k = 0;
    int bound = ((*i).second).size() - 1;
    for(set<string>::iterator j=((*i).second).begin(); j!=((*i).second).end(); j++) {
      if(k==bound)
	cout << (*j) << " ";
      else
	cout << (*j) << ", ";
      k++;
    }
    cout << "}; " << endl;
  }
  cout << "end Group" << endl << endl;
  cout << "Fairness" << endl;
  for(vector<fairness_expression*>::iterator i=is_fairness->begin(); i!=is_fairness->end(); i++)
    cout << "    " << (*i)->to_string() << ";" << endl;
  cout << "end Fairness" << endl << endl;
  cout << "Formulae" << endl;
  for(vector<modal_formula>::iterator i=is_formulae->begin(); i!=is_formulae->end(); i++)
    cout << "    " << (*i).to_string() << ";" << endl;
  cout << "end Formulae" << endl;
}

void print_state(BDD state, BDDvector v) {
  for(unsigned int i=0; i<agents->size(); i++) {
    cout << "Agent " << (*agents)[i]->get_name() << endl;
    (*agents)[i]->print_state("  ", state, v);
  }
}

string state_to_str(BDD state, BDDvector v) {
    ostringstream s;
	
	for(unsigned int i=0; i<agents->size(); i++) {
		s << "  Agent " << (*agents)[i]->get_name()<< "\n" ;
		s << (*agents)[i]->state_to_str( state, v);
		
	}		
	return s.str();
}

bool find_same_state(map<string, int> *statehash, string state) {
  if(statehash != NULL) {
    map<string,int>::iterator iter = statehash->find(state);  
    if( iter != statehash->end() ) {
      //cout << "find a redundant state" << endl;
      return true;
    }
  }
  return false;
}

bool is_valid_state(BDD state, BDDvector v) {
  for(unsigned int i=0; i<agents->size(); i++) {
    if(!(*agents)[i]->is_valid_state(state, v))
      return false;
  }
  return true;
}

void print_action(BDD state, BDDvector a) {
  for(unsigned int i=0; i<agents->size(); i++) {
    cout << (*agents)[i]->get_name() << " : ";
    (*agents)[i]->print_action(state, a);
    if(i<agents->size()-1)
      cout << "; ";
    else
      cout << endl;
  }
}

void print_action_1(BDD state, BDDvector a) {
  for(unsigned int i=0; i<agents->size(); i++) {
    cout << (*agents)[i]->get_name() << " : ";
    (*agents)[i]->print_action(state, a);
    cout << "; ";
  }
  cout << endl;
}

bool is_valid_action(BDD state, BDDvector a) {
  for(unsigned int i=0; i<agents->size(); i++) {
    if(!(*agents)[i]->is_valid_action(state, a))
      return false;
  }
  return true;
}

BDD append_variable_BDDs(BDD a) {
  for(unsigned int j=0; j<agents->size(); j++) {
    map<string, basictype *> *obsvars = (*agents)[j]->get_obsvars();
    if(obsvars!=NULL && obsvars->size()>0)
      for(map<string, basictype *>::iterator i=obsvars->begin(); i!=obsvars->end(); i++)
	if((*i).second->get_type()==3)
	  a *= ((enumerate *)(*i).second)->build_all_BDDs();
    map<string, basictype *> *vars = (*agents)[j]->get_vars();
    if(vars!=NULL && vars->size()>0)
      for(map<string, basictype *>::iterator i=vars->begin(); i!=vars->end(); i++)
	if((*i).second->get_type()==3)
	  a *= ((enumerate *)(*i).second)->build_all_BDDs();
  }
  return a;
}

int main(int argc, char* argv[]) {
	time_t t0 = time( 0 );

	is_agents = new map<string, basic_agent *>;
	agents = new vector<basic_agent *>;
	is_evaluation = new map<string, bool_expression *>;
	is_groups = new map<string, set<string> >;
	is_formulae = new vector<modal_formula>;

	is_fairness = new vector<fairness_expression *>;

	options["verbosity"] = 0;
	options["pnode_info"] = 0;
	options["bdd_stats"] = 0;
	options["uniform"] = 0;
	options["globalcheck"] = 0;
	options["debug"] = 0;
	options["cex"] = 0;
	options["simulation"] = 0;
	options["source"] = 0;
	options["syncheck"] = 0;
	options["nobanner"] = 0;
	options["quiet"] = 0;
	options["experiment"] = 1;
	options["initialfirst"] = 0;
	options["optimization"] = 0;
	options["symmetry"] = 0;
	options["bddcache"] = 3;
	options["flatexpression"] = 0;


	static const std::string verb("-v");
	static const std::string binfo("-bdd_stats");
	static const std::string uniform("-u");
	static const std::string globalcheck("-g");
	static const std::string help("-h");
	static const std::string help2("--help");
	static const std::string debug("-d");
	static const std::string cex("-c");
	static const std::string simulation("-s");
	static const std::string source("-src");
	static const std::string syncheck("-gsc");
	static const std::string nobanner("-nb");
	static const std::string quiet("-q");
	static const std::string experiment("-e");
	static const std::string bddcache("-bc");
	static const std::string initialfirst("-if");
	static const std::string optimization("-op");
	static const std::string cexoutdir("-cex_prefix");
	static const std::string symmetry("-symmetry");
	static const std::string flatexpression("-fe");

	std::string filename;

	string s;

	mcmas_driver driver;

	calReachRT = true;

	if (argc < 2) {
		print_help();
		exit(1);
	}

	cex_prefix = "";
	for (int i=1; i < argc-1; ++i) {
		if (binfo == argv[i]) {
			options["bdd_stats"] = 1;
		} else if (uniform == argv[i]) {
			options["uniform"] = 1;
		} else if (debug == argv[i]) {
			options["debug"] = 1;
		} else if (cex == argv[i]) {
			options["cex"] = 1;
		} else if (verb == argv[i]) {
			string s;
			s = argv[++i]; // consume the argument
			options["verbosity"] = atoi(s.c_str());
		} else if (globalcheck == argv[i]) {
			options["globalcheck"] = 1;
		} else if ( (help == argv[i] ) || (help2 == argv[i] )) {
			print_help();
			exit(1);
		} else if (simulation == argv[i]) {
			options["simulation"] = 1;
		} else if (source == argv[i]) {
			options["source"] = 1;
		} else if (syncheck == argv[i]) {
			options["syncheck"] = 1;
		} else if (nobanner == argv[i]) {
			options["nobanner"] = 1;
		} else if (quiet == argv[i]) {
			options["quiet"] = 1;
		} else if (experiment == argv[i]) {
			string s;
			s = argv[++i]; // consume the argument
			options["experiment"] = atoi(s.c_str());
		} else if (bddcache == argv[i]) {
		  string s;
		  s = argv[++i]; // consume the argument
		  options["bddcache"] = atoi(s.c_str());
		} else if (initialfirst == argv[i]) {
			options["initialfirst"] = 1;
		} else if (optimization == argv[i]) {
			string s;
			s = argv[++i]; // consume the argument
			options["optimization"] = atoi(s.c_str());
		} else if (cexoutdir == argv[i]) {
			string s;
			s = argv[++i]; // consume the argument
			cex_prefix = s;
		} else if (symmetry == argv[i]) {
			string s;
			s = argv[++i]; // consume the argument
			options["symmetry"] = atoi(s.c_str());
		} else if (flatexpression == argv[i]) {
		  string s;
		  s = argv[++i]; // consume the argument
		  options["flatexpression"] = atoi(s.c_str());
		} else {
			// whether to do anything about unexpected
			// arguments is a matter of taste (see below)
			cout << NAME << " invalid option: " << argv[i] << endl;
			print_help();
			exit(1);
		}
	}

	if(options["nobanner"] == 0 && options["quiet"] == 0)
	  print_banner();

	if ( (help == argv[argc-1] ) || (help2 == argv[argc-1] )) {
		print_help();
		exit(1);
	}

	filename = argv[argc-1];

	if (options["debug"] == 1) {
		cout << "DEBUG: filename" << endl;
		cout << "    " << filename << endl;
		cout << "DEBUG: end filename" << endl;
	}

	driver.parse(filename);
	if (!driver.syntax_error) {
	  if(options["quiet"] == 0 || options["simulation"] == 0)
	    cout << filename << " has been parsed successfully." << endl;
	  if (options["source"] == 1) {
	    print_ispl_code();
	    exit(0);
	  }
	  if(options["quiet"] == 0 || options["simulation"] == 0)
	    cout << "Global syntax checking..." << endl;
	  if (!global_consistency_check()) {
	    cout << filename << " has error(s)." << endl;
	    exit(-1);
	  }
	  if(options["quiet"] == 0)
		cout << "Done" << endl;
		if(options["syncheck"] == 1)
		  exit(0);
	} else {
		cout << filename << " has syntax error(s)." << endl;
		exit(-1);
	}

	/*if (options["debug"] == 1) {
		cout << "DEBUG: begin ISPL code" << endl;
		print_ispl_code();
		cout << "DEBUG: end ISPL code" << endl;
		}*/

	if (options["verbosity"] > 2) {
		cout << "Starting OBDD manager...";
	}

	bddmgr = new ( Cudd );
	bddOne = new BDD(bddmgr->bddOne());

	if (options["verbosity"] > 2) {
		cout << "Done" << endl;
	}

	if(options["simulation"] == 0 || options["quiet"] == 0)
	  cout << "Encoding BDD parameters..." << endl;

	// Calculate the number of BDD variables for states and actions
	int states_count = 0;
	int actions_count = 0;
	for (unsigned int i=0; i<agents->size(); i++) {
		states_count += (*agents)[i]->state_BDD_length();
		actions_count += (*agents)[i]->actions_BDD_length();
	}

	if (options["verbosity"] > 3) {
		cout << endl << "  Building BDDs variables...";
	}

	v = new BDDvector(states_count,bddmgr); // Local states
	addv = new ADDvector(states_count,bddmgr);
	pv = new BDDvector(states_count,bddmgr); // "Next" local states
	addpv = new ADDvector(states_count,bddmgr);
	a = new BDDvector(actions_count,bddmgr); // Actions

	// Allocate BDD variables to states and actions
	for (int i = 0; i < states_count; i++) {
		(*v)[i] = bddmgr->bddVar(i);
		(*pv)[i] = bddmgr->bddVar(states_count + i);
		(*addv)[i] = bddmgr->addVar(i);
		(*addpv)[i] = bddmgr->addVar(states_count + i);
	}
	for (int i = 0; i < actions_count; i++)
		(*a)[i] = bddmgr->bddVar(states_count*2 + i);

	if (options["verbosity"] > 3) {
		cout << "Done: " << v->count() << " for local states, " << a->count()
				<< " for actions." << endl;
	}

	// Calculate the position of every variable and every set of actions
	int k1 = 0;
	int k2 = 0;
	for (unsigned int i=0; i<agents->size(); i++) {
		k1 = (*agents)[i]->allocate_BDD_2_variables(k1);
		k2 = (*agents)[i]->allocate_BDD_2_actions(k2);
	}

	if (options["debug"] == 1) {
		cout << endl << "BDD index and encoding of all values of all variables"
				<< endl;
		for (unsigned int i=0; i<agents->size(); i++) {
			cout << (*agents)[i]->get_name() << ":" << endl;
			(*agents)[i]->print_variable_BDD_encoding();
			cout << endl;
		}
	}

	obsvars_bdd_length = (*agents)[0]->obsvars_BDD_length();

	// Encode initial states
	in_st = bddmgr->bddOne();

	if(options["experiment"] >= 1 || options["simulation"] == 1) {
	  vprot = new BDDvector(agents->size(), bddmgr); 
	  vevol = new BDDvector(agents->size(), bddmgr); 
	  vRT = new BDDvector(agents->size(), bddmgr); 
	}
	if (options["simulation"] == 1) {
	  Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_GROUP_SIFT_CONV);
	  in_st = (options["flatexpression"]<2) ? is_istates->encode_bdd(bddOne) : is_istates->encode_bdd_flat(bddOne);
	  in_st = append_variable_BDDs(in_st);
	  Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_GROUP_SIFT_CONV);
	  if (in_st==bddmgr->bddZero()) {
	    cout << "No initial state. Check your model!" << endl;
	    exit(0);
	  }
	  for (unsigned int i=0; i<agents->size(); i++) {
	    (*vprot)[i] = (*agents)[i]->encode_protocol();
	    (*vevol)[i] = (*agents)[i]->encode_evolution();
	    (*vRT)[i] = (*vprot)[i] * (*vevol)[i];
	  }
	  Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);	
	  BDDvector inistates(MAXINISTATES, bddmgr);
	  int count=0;
	  BDD is = in_st;
	  map<string, int> statehash;
	  while (count<MAXINISTATES && is != bddmgr->bddZero()) {
	    inistates[count] = is.PickOneMinterm(*v);
	    is = is - inistates[count];
	    if (is_valid_state(inistates[count], *v)) {
	      string state = state_to_str(inistates[count], *v);
	      if(!find_same_state(&statehash, state)) {
		statehash[state] = 1;
		count++;
	      }
	    }
	  }
	  if(options["quiet"] == 0) {
	    if (count==MAXINISTATES && is != bddmgr->bddZero())
	      cout
		<< "There are too many initial states. Please specify more initial values."
		<< endl;
	    while (true) {
	      BDDvector stack(MAXSTACKDEPTH, bddmgr);
	      int sp = 0;
	      cout << endl << "--------- Initial state ---------" << endl;
	      int isindex = 0;
	      print_state(inistates[isindex], *v);
	      cout << "----------------------------" << endl;
	      if (count>1) {
		bool choose = false;
		string c;
		while (!choose) {
		  if (isindex==0)
		    cout
		      << "Is this the initial state? [Y(es), N(ext), E(xit)]: ";
		  else if (isindex==count-1)
		    cout
		      << "Is this the initial state? [Y(es), P(revious), E(xit)]: ";
		  else
		    cout
		      << "Is this the initial state? [Y(es), N(ext), P(revious), E(xit)]: ";
		  cin >> c;
		  transform(c.begin(), c.end(), c.begin(),
			    (int(*)(int))tolower);
		  if (c=="y" || c=="yes") {
		    choose = true;
		    break;
		  } else if ((isindex<count-1/* || isindex<99*/) && (c=="n" || c =="next")) {
		    isindex++;
		    cout << endl << "--------- Initial state ---------"
			 << endl;
		    print_state(inistates[isindex], *v);
		    //inistates[isindex].print(2, 2);
		    cout << "----------------------------" << endl;
		  } else if (isindex>0 && (c=="p" || c=="previous")) {
		    isindex--;
		    cout << endl << "--------- Initial state ---------"
			 << endl;
		    print_state(inistates[isindex], *v);
		    //inistates[isindex].print(2, 2);
		    cout << "----------------------------" << endl;
		  } else if (c=="e" || c=="exit")
		    exit(0);
		  else
		    cout << "Please choose one option!" << endl;
		}
	      }
	      stack[sp++] = inistates[isindex]; // initial state
	      while (true) {
		BDDvector enabled(MAXTRANSITIONS, bddmgr);
		int tcount = 0;
		BDD newstates;
		int tran = -2;
		if (sp>=MAXSTACKDEPTH) {
		  while (tran!=-1 && tran!=0) {
		    cout <<"Maximum stack depth is reached. Please type 0 to backtrack or -1 to quit immediately: ";
		    cin >> tran;
		  }
		} else {
		  if (options["debug"] == 1 && options["verbosity"] > 10) {
		    cout << "newstate: ";
		    stack[sp-1].print(2, 2);
		  }
		  newstates = stack[sp-1];// next state
		  for(unsigned int k=0; k<agents->size(); k++)
		    newstates *= (*vRT)[k];
		  if (options["debug"] == 1 && options["verbosity"] > 10) {
		    cout << "newstate + trans: ";
		    newstates.print(2, 2);
		  }
		  BDD transitions = newstates;
		  if(newstates != bddmgr->bddZero()) {
		    while (tcount<MAXTRANSITIONS && transitions != bddmgr->bddZero()) {
		      enabled[tcount] = transitions.PickOneMinterm(*a);
		      transitions-=enabled[tcount];
		      if (is_valid_action(enabled[tcount], *a))
			tcount++;
		    }
		  }
		  if (tcount==MAXTRANSITIONS && transitions != bddmgr->bddZero())
		    cout << "There are too many enabled transitions. We only pick up 100 of them." << endl;
		  if (tcount>0) {
		    cout << "Enabled transtions: "<< endl;
		    for (int i=0; i<tcount; i++) {
		      cout << i+1 << " : ";
		      print_action(enabled[i], *a);
		    }
		    while (tran<-1 || tran>tcount) {
		      cout << "Please choose one, or type 0 to backtrack or -1 to quit: " << endl;
		      cin >> tran;
		    }
		  } else {
		    if(sp > 1 || count > 1) {
		      cout << "There is no enabled transition. Please type 0 to backtrack or -1 to quit immediately: ";
		      cin >> tran;
		      while(tran!=-1 && tran!=0) {
			cout<<"Maximum stack depth is reached. Please type 0 to backtrack or -1 to quit immediately: ";
			cin >> tran;
		      }
		    } else {
		      cout << "There is no enabled transition in the initial state. Simulation exits." << endl;
		      exit(0);
		    }
		  }
		}
		if(tran==-1)
		  exit(0);
		else if(tran==0 && sp>1) {
		  stack[--sp] = bddmgr->bddZero();
		  continue;
		} else if(tran==0 && sp==1) {
		  stack[--sp] = bddmgr->bddZero();
		  break;
		} else {
		  newstates *= enabled[tran-1];
		  if (options["debug"] == 1 && options["verbosity"] > 10) {
		    cout << "newstate: ";
		    newstates.print(2, 2);
		  }
		  BDD tmpstate = Exists(v, newstates); // clear state variables
		  tmpstate = tmpstate.SwapVariables(*v, *pv);
		  tmpstate = Exists(a, tmpstate); // clear action variables
		  tmpstate = append_variable_BDDs(tmpstate);
		  //BDD tmpback = tmpstate;
		  int count1=0;
		  BDDvector succstates(MAXINISTATES, bddmgr);
		  statehash.clear();
		  while (count1<MAXINISTATES && tmpstate != bddmgr->bddZero()) {
		    succstates[count1] = tmpstate.PickOneMinterm(*v);
		    tmpstate = tmpstate - succstates[count1];
		    if (is_valid_state(succstates[count1], *v)) {
		      string state = state_to_str(succstates[count1], *v);
		      if(!find_same_state(&statehash, state)) {
			statehash[state] = 1;
			count1++;
		      }
		    }
		  }
		  if (count1==MAXINISTATES && tmpstate != bddmgr->bddZero())
		    cout
		      << "There are too many states. Please refine your model."
		      << endl;
		  else if(count1>1)
		    cout << "Warning: there are many states." << endl;
		  cout << endl << "--------- Current state ---------" << endl;
		  int succindex = 0;
		  print_state(succstates[succindex], *v);
		  cout << "----------------------------" << endl;
		  if (count1>1) {
		    bool choose = false;
		    string c;
		    while (!choose) {
		      if (succindex==0)
			cout
			  << "Is this the current state? [Y(es), N(ext), E(xit)]: ";
		      else if (succindex==count1 -1)
			cout
			  << "Is this the current state? [Y(es), P(revious), E(xit)]: ";
		      else
			cout
			  << "Is this the current state? [Y(es), N(ext), P(revious), E(xit)]: ";
		      cin >> c;
		      transform(c.begin(), c.end(), c.begin(),
				(int(*)(int))tolower);
		      if (c=="y" || c=="yes") {
			choose = true;
			break;
		      } else if ((succindex<count1-1/* || succindex<99*/) && (c=="n" || c =="next")) {
			succindex++;
			cout << endl << "--------- Current state ---------"
			     << endl;
			print_state(succstates[succindex], *v);
			//succstates[succindex].print(2, 2);
			cout << "----------------------------" << endl;
		      } else if (succindex>0 && (c=="p" || c=="previous")) {
			succindex--;
			cout << endl << "--------- Initial state ---------"
			     << endl;
			print_state(succstates[succindex], *v);
			//succstates[succindex].print(2, 2);
			cout << "----------------------------" << endl;
		      } else if (c=="e" || c=="exit")
			exit(0);
		      else
			cout << "Please choose one option!" << endl;
		    }
		  }
		  stack[sp++] = succstates[succindex];
		  //sp++;
		}
	      }
	    }
	  } else {
	    while (true) {
	      BDDvector stack(MAXSTACKDEPTH, bddmgr);
	      int sp = 0;
	      for(int k=0; k<count; k++) {
		cout << endl << "-- State " << (k+1) << " --" << endl;
		print_state(inistates[k], *v);
	      }
	      cout << "Done." << endl;
	      int isindex = 0;;
	      cin >> isindex;
	      if(isindex==-1)
		exit(0);
	      stack[sp++] = inistates[isindex-1]; // initial state
	      while (true) {
		BDDvector enabled(MAXTRANSITIONS, bddmgr);
		int tcount = 0;
		BDD newstates;
		int tran = -2;
		if (sp>=MAXSTACKDEPTH) {
		  while (tran!=-1 && tran!=0) {
		    cout <<"Maximum stack depth is reached. Please backtrack or quit.";
		    cin >> tran;
		  }
		} else {
		  newstates = stack[sp-1];// next state
		  for(unsigned int k=0; k<agents->size(); k++)
		    newstates *= (*vRT)[k];
		  BDD transitions = newstates;
		  if(newstates != bddmgr->bddZero()) {
		    while (tcount<MAXTRANSITIONS && transitions != bddmgr->bddZero()) {
		      enabled[tcount] = transitions.PickOneMinterm(*a);
		      transitions-=enabled[tcount];
		      if (is_valid_action(enabled[tcount], *a))
			tcount++;
		    }
		  }
		  if (tcount>0) {
		    //cout << "Enabled transtions: "<< endl;
		    for (int i=0; i<tcount; i++) {
		      cout << "-- transition "<< (i+1) << " --" << endl;
		      print_action_1(enabled[i], *a);
		    }
		    cout << "Done." << endl;
		    while (tran<-1 || tran>tcount) {
		      cin >> tran;
		    }
		  } else {
		    if(sp > 1 || count > 1) {
		      cout << "There is no enabled transition. Please backtrack or quit.";
		      cin >> tran;
		    } else {
		      cout << "There is no enabled transition in the initial state." << endl;
		      exit(0);
		    }
		  }
		}
		if(tran==-1)
		  exit(0);
		else if(tran==0 && sp>1) {
		  stack[--sp] = bddmgr->bddZero();
		  continue;
		} else if(tran==0 && sp==1) {
		  stack[--sp] = bddmgr->bddZero();
		  break;
		} else {
		  newstates *= enabled[tran-1];
		  BDD tmpstate = Exists(v, newstates); // clear state variables
		  tmpstate = tmpstate.SwapVariables(*v, *pv);
		  tmpstate = Exists(a, tmpstate); // clear action variables
		  tmpstate = append_variable_BDDs(tmpstate);
		  //BDD tmpback = tmpstate;
		  int count1=0;
		  BDDvector succstates(MAXINISTATES, bddmgr);
		  statehash.clear();
		  while (count1<MAXINISTATES && tmpstate != bddmgr->bddZero()) {
		    succstates[count1] = tmpstate.PickOneMinterm(*v);
		    tmpstate = tmpstate - succstates[count1];
		    if (is_valid_state(succstates[count1], *v)) {
		      string state = state_to_str(succstates[count1], *v);
		      if(!find_same_state(&statehash, state)) {
			statehash[state] = 1;
			count1++;
		      }
		    }
		  }
		  for(int k=0; k<count1; k++) {
		    cout << endl << "-- State " << (k+1) << " --" << endl;
		    print_state(succstates[k], *v);
		  }
		  cout << "Done." << endl;
		  int succindex = 0;
		  cin >> succindex;
		  if(succindex == -1)
		    exit(0);
		  if(succindex > 0)
		    stack[sp++] = succstates[succindex-1];
		  //sp++;
		}
	      }
	    }
	  }
	}
	  
	  // REORDERING OF VARIABLES:
	Cudd_ReorderingType REORDERING = CUDD_REORDER_GROUP_SIFT_CONV;
	if(options["optimization"] == 0 || options["optimization"] == 2)
	  Cudd_AutodynEnable(bddmgr->getManager(), REORDERING);

	if(options["experiment"] == 0) {
	  if(options["initialfirst"] == 1) {
	    if (options["quiet"] == 0 || options["simulation"] == 0) {
	      cout << "  Building OBDD for initial states..." << endl;
	    }
	    in_st = (options["flatexpression"]<2) ? is_istates->encode_bdd(bddOne) : is_istates->encode_bdd_flat(bddOne);
	    in_st = append_variable_BDDs(in_st);
	    if(options["optimization"] == 1 || options["optimization"] == 2)
	      Cudd_AutodynEnable(bddmgr->getManager(), REORDERING);
	    if (options["verbosity"] > 3) {
	      cout << "Done." << endl;
	    }
	  }


	  if (options["quiet"] == 0 || options["simulation"] == 0)
	    cout << "Building OBDDs for protocols..." << endl;

	  andprot = bddmgr->bddOne();
	  for (unsigned int i=0; i<agents->size(); i++) {
	    if (options["verbosity"] > 3) {
	      if(i==0)
		cout << endl;
	      cout << "      Processing " << (*agents)[i]->get_name() << endl;
	    }
	    andprot *= (*agents)[i]->encode_protocol();
	  }
	  
	  if (options["quiet"] == 0 || options["simulation"] == 0) {
	    cout << "Building OBDDs for evolution functions..." << endl;
	  }
	  
	  // Encode evolution functions
	  preRT = bddmgr->bddOne();
	  
	  // REORDERING OF VARIABLES (forcing it):
	  Cudd_AutodynEnable(bddmgr->getManager(), REORDERING);
	  
	  for (unsigned int i=0; i<agents->size(); i++) {
	    if (options["verbosity"] > 3) {
	      if(i==0)
		cout << endl;
	      cout << "      Processing " << (*agents)[i]->get_name() << endl;
	    } 
	    preRT *= (*agents)[i]->encode_evolution();
	    //Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_GROUP_SIFT_CONV);
	  }
	  
	  if (options["quiet"] == 0 || options["simulation"] == 0) {
	    cout << "Building OBDD for global transition relation..." << endl;
	  }
	  
	  Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);
	  
	  //BDD trans = bddmgr->bddOne();
	  trans = preRT*andprot;
	  RT = Exists(a, trans);
	  if (options["verbosity"] > 3) {
	    cout << "Done." << endl;
	  }

	  if(options["initialfirst"] == 0) {
	    if (options["quiet"] == 0 || options["simulation"] == 0) {
	      cout << "Building OBDD for initial states..." << endl;
	    }
	    in_st = (options["flatexpression"]<2) ? is_istates->encode_bdd(bddOne) : is_istates->encode_bdd_flat(bddOne);
	    in_st = append_variable_BDDs(in_st);
	    if (options["verbosity"] > 3) {
	      cout << "Done." << endl;
	    }
	  }

	// REORDERING OF VARIABLES (forcing):
	  REORDERING = CUDD_REORDER_WINDOW4_CONV;
	  Cudd_AutodynEnable(bddmgr->getManager(),REORDERING);
	  //Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);

	  if (options["quiet"] == 0 || options["simulation"] == 0) {
	    cout << "Building OBDD for reachable states..." << endl;
	  }
	  
	  reach = bddmgr->bddZero();
	  BDD q = in_st;
	  BDD next = bddmgr->bddZero();
	  int acounter = 1;
	  
	  if ( options["verbosity"]> 9 ) {
	    cout << endl;
	  }
  
	  while ( q != reach ) {
	    if ( options["verbosity"]> 5 ) {
	      cout << "    Computing depth "<< acounter << endl;
	    }
	    reach = q;
	    if ( options["verbosity"]> 9 ) {
	      cout << "Before evolution: " << endl;
	      Exists(a, reach).print(2,2);
	    }
	    next = Exists(v, q*RT);
	    next = next.SwapVariables(*v, *pv);
	    //next = append_variable_BDDs(next);
	    //next = Exists(a, next); // Clear actions.
	    
	    if ( options["verbosity"]> 9 ) {
	      cout << "Next states: " << endl;
	      next.print(2,2);
	    }
	    //next = next * andprot; // syncronize actions with local states
	    q = q + next;
	    //if ( options["debug"] == 1 ) {
	    if ( options["verbosity"]> 9 ) {
	      cout << "Press enter to continue" << endl;
	      char c;
	      cin.get(c);
	    }
	    acounter++;
	  }
	  
	  reach = Exists(a, reach);
	} else { // ---------------------- experiment  ------------------------
	  Cudd_AutodynEnable(bddmgr->getManager(), REORDERING);
	  if(options["initialfirst"] == 1) {
	    if (options["quiet"] == 0 || options["simulation"] == 0) {
	      cout << "Building OBDD for initial states...";
	      cout.flush();
	    }
	    //options["debug"] = 1;
	    //scount = 0;*/
	    in_st = (options["flatexpression"]<2) ? is_istates->encode_bdd(bddOne) : is_istates->encode_bdd_flat(bddOne);
	    in_st = append_variable_BDDs(in_st);
	    //options["debug"] = 0;
	    //Cudd_AutodynEnable(bddmgr->getManager(), REORDERING);
	    if (options["verbosity"] > 3) 
	      cout << "Done." << endl;
	  }
	  if (options["quiet"] == 0 || options["simulation"] == 0) {
	    cout << "Building partial transition relation..." << endl;
	  }
	  for (unsigned int i=0; i<agents->size(); i++) {
	    if (options["verbosity"] > 3) {
	      if(i==0)
		cout << endl;
	      cout << "      Processing " << (*agents)[i]->get_name() << ": protocol... ";
	      cout.flush();
	    }
	    time_t tmpt0 = time(0);
	    (*vprot)[i] = (*agents)[i]->encode_protocol();
	    if (options["verbosity"] > 3) {
	      cout << "done, evolution...";
	      cout.flush();
	    }
	    (*vevol)[i] = (*agents)[i]->encode_evolution();
	    if (options["verbosity"] > 3) {
	      cout << "done, local transition relation...";
	      cout.flush();
	    }
	    (*vRT)[i] = (*vprot)[i] * (*vevol)[i];
	    time_t tmpt1 = time(0);
	    if (options["verbosity"] > 3) {
	      cout << "done. " << difftime(tmpt1, tmpt0) << " seconds" << endl;
	    }
	    //Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);
	  }	
	  if ( options["verbosity"]> 3 ) 
	    cout << "Done" << endl;
	  if(options["initialfirst"] == 0) {
	    if (options["quiet"] == 0 || options["simulation"] == 0) {
	      cout << "Building OBDD for initial states..." << endl;
	    }
	    in_st = (options["flatexpression"]<2) ? is_istates->encode_bdd(bddOne) : is_istates->encode_bdd_flat(bddOne);
	    in_st = append_variable_BDDs(in_st);
	    if (options["verbosity"] > 3) 
	      cout << "Done." << endl;
	  }
	  //REORDERING = CUDD_REORDER_WINDOW4_CONV;
	  //Cudd_AutodynEnable(bddmgr->getManager(),REORDERING);
	  Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);
	  if (options["quiet"] == 0 || options["simulation"] == 0) {
	    cout << "Building reachable state space..." << endl;
	  }
	  if(options["experiment"] == 1) {
	    reach = bddmgr->bddZero();
	    BDD q1 = in_st;
	    BDD next1 = bddmgr->bddZero();
	    int acounter1 = 1;
	    
	    //time_t t21 = time( 0 );
	    while ( q1 != reach ) {
	      if ( options["verbosity"]> 5 ) 
		cout << "    Computing depth "<< acounter1 << endl;
	      
	      reach = q1;
	      if ( options["verbosity"]> 9 ) {
		cout << "Before evolution: " << endl;
		Exists(a, reach).print(2,2);
	      }
	      //next1 = Exists(v, q*RT);
	      next1 = q1;
	      for(unsigned int k=0; k<agents->size(); k++)
		next1 *= (*vRT)[k];
	      next1 = Exists(v, next1);
	      next1 = next1.SwapVariables(*v, *pv);
	      //next = append_variable_BDDs(next);
	      next1 = Exists(a, next1); // Clear actions.
	      
	      if ( options["verbosity"]> 9 ) {
		cout << "Next states: " << endl;
		next1.print(2,2);
	      }
	      //next = next * andprot; // syncronize actions with local states
	      //if(!(next1<=q1))
		q1 = q1 + next1;
	      //Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);
	      //if ( options["debug"] == 1 ) {
	      if ( options["verbosity"]> 9 ) {
		cout << "Press enter to continue" << endl;
		char c;
		cin.get(c);
	      }
	      acounter1++;
	    }
	    
	    reach = Exists(a, reach);

	    if ( options["verbosity"]> 3 ) {
	      cout << "Done" << endl;
	    }
	  } else if(options["experiment"] == 2) { // BMC goes here
	    reach = in_st;
	    BDD q1 = bddmgr->bddZero();
	    BDD next1 = in_st;
	    int acounter1 = 1;
	    while ( true ) {
	      if ( options["verbosity"]> 5 ) 
		cout << "    Computing depth "<< acounter1 << endl;
	      
	      if ( options["verbosity"]> 9 ) {
		cout << "Before evolution: " << endl;
		Exists(a, reach).print(2,2);
	      }
	      for(unsigned int k=0; k<agents->size(); k++)
		next1 *= (*vRT)[k];
	      next1 = Exists(v, next1);
	      next1 = next1.SwapVariables(*v, *pv);
	      next1 = Exists(a, next1); // Clear actions.
	      Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);
	      
	      if ( options["verbosity"]> 9 ) {
		cout << "Next states: " << endl;
		next1.print(2,2);
	      }
	      q1 = reach + next1;
	      Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);
	      
	      if ( options["verbosity"]> 9 ) {
		cout << "Press enter to continue" << endl;
		char c;
		cin.get(c);
	      }
	      acounter1++;
	      
	      if(q1 == reach)
		break;
	      else
		reach = q1;
	    }

	    if ( options["verbosity"]> 3 ) {
	      cout << "Done" << endl;
	    }
	    
	  } else { 
	    reach = in_st;
	    BDD next1 = in_st;
	    int acounter1 = 1;
	    while ( true ) {
	      if ( options["verbosity"]> 5 ) 
		cout << "    Computing depth "<< acounter1 << endl;
	      
	      if ( options["verbosity"]> 9 ) {
		cout << "Before evolution: " << endl;
		Exists(a, reach).print(2,2);
	      }
	      for(unsigned int k=0; k<agents->size(); k++)
		next1 *= (*vRT)[k];
	      next1 = Exists(v, next1);
	      next1 = next1.SwapVariables(*v, *pv);
	      next1 = Exists(a, next1); // Clear actions.
	      Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);
	      
	      if ( options["verbosity"]> 9 ) {
		cout << "Next states: " << endl;
		next1.print(2,2);
	      }
	      next1 = next1 - reach;
	      Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);
	      
	      if ( options["verbosity"]> 9 ) {
		cout << "Press enter to continue" << endl;
		char c;
		cin.get(c);
	      }
	      acounter1++;
	      
	      if(next1 == bddmgr->bddZero())
		break;
	      else
		reach = reach + next1;
	      if ( options["verbosity"]> 3 ) {
		cout << "Done" << endl;
	      }
	    }
	  } 
	}
	Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);

	if ( options["verbosity"]> 6 ) {
		cout << "Reachable states: " << endl;
		reach.print(2,2);
	}

	if ( options["verbosity"]> 4 )
	cout << "Done." << endl;

	need_epistemic = false;
	need_deontic = false;
	need_common_knowledge = false;
	need_deontic_knowledge = false;
	need_distributed_knowledge = false;

	for(vector<modal_formula>::iterator i=is_formulae->begin(); i!=is_formulae->end(); i++)
	i->look_flags(&need_epistemic, &need_deontic, &need_common_knowledge,
	&need_deontic_knowledge, &need_distributed_knowledge);

	if(need_epistemic) {
	  //obsvars_bdd_length = (*agents)[0]->obsvars_BDD_length();
		if (options["quiet"] == 0 || options["simulation"] == 0)
		  cout << "Building OBDD for epistemic relations..." << endl;

		//BDD epistemic;// = bddmgr->bddOne();
		int length = 0; //is_environment->state_BDD_length();
		int position = 0;
		//for(int i=position; i<position+length; i++)
		//  epistemic = epistemic * ( ( !v[i] + pv[i] ) * 
		//			( v[i] + !pv[i] ) );
		//pair<string, BDD> p("Environment", epistemic); 
		//er.insert(p);
		//position += length;
		//    epistemic.print(2,2);
		for(unsigned int k=0; k<agents->size(); k++) {
		  BDD epistemic = bddmgr->bddOne();
		  length = (*agents)[k]->state_BDD_length();
		  for(int i=position; i<position+length; i++)
		    epistemic = epistemic * ( ( !(*v)[i] + (*pv)[i] ) *
					      ( (*v)[i] + !(*pv)[i] ) );
		  if(k > 0)
		    for(int i=0; i<obsvars_bdd_length; i++)
		      epistemic = epistemic * ( ( !(*v)[i] + (*pv)[i] ) *
						( (*v)[i] + !(*pv)[i] ) );
		  er.insert(pair<string, BDD>((*agents)[k]->get_name(), epistemic));
		  position += length;
		}

		if ( options["verbosity"]> 4 )
		cout << "Done." << endl;
	}

	if(need_deontic) {
		if (options["quiet"] == 0 || options["simulation"] == 0)
		  cout << "Building OBDD for deontic relations..." << endl;
		//BDD deontic; // = is_environment->encode_greenstates(pv, a, bddmgr);
		//pair<string, BDD> p("Environment", deontic);
		//deor.insert(p);
		for(unsigned int k=0; k<agents->size(); k++) {
			BDD deontic = (*agents)[k]->encode_greenstates();
			deor.insert(pair<string, BDD>((*agents)[k]->get_name(), deontic));
		}

		if ( options["verbosity"]> 4 ) {
			cout << "Done." << endl;
		}
	}

	// Building group modalities
	if (is_groups!=NULL && is_groups->size()>0) {
		if (options["quiet"] == 0 || options["simulation"] == 0) {
			cout << "Building OBDD for group modalities..." << endl;
		}

		BDD egm; // Epistemic group modality
		BDD dgm; // Deontic group modality
		BDD tmpdistgm; // Distributed knowledge
		set<string> names_g;
		BDD swapreach = reach.SwapVariables(*v, *pv);

		if(need_common_knowledge) {
			if ( options["verbosity"]> 4 )
			cout << "    Building OBDD for everybody knows...";

			for(map<string, set<string> >::iterator gi=is_groups->begin(); gi!=is_groups->end(); gi++) {
				// Epistemic first
				egm = bddmgr->bddZero();
				names_g = gi->second;
				for (set<string>::iterator igs=names_g.begin(); igs!=names_g.end(); igs++)
				egm = (egm + er[*igs])*reach*swapreach;
				/** Can it be the following code
				 for (set<string>::iterator igs=names_g.begin(); igs!=names_g.end(); igs++)
				 egm += er[*igs];
				 egm *= reach*swapreach;
				 */
				gm[gi->first] = egm;
				//egm.print(2,2);
			}

			if ( options["verbosity"]> 4 )
			cout << "Done." << endl;
		}

		if(need_deontic_knowledge) {
			if ( options["verbosity"]> 4 )
			cout << "    Building OBDD for deontic knowledge...";

			for(map<string, set<string> >::iterator gi=is_groups->begin(); gi!=is_groups->end(); gi++) {
				// Deontic
				dgm = bddmgr->bddOne();
				names_g = gi->second;
				for (set<string>::iterator igs=names_g.begin(); igs!=names_g.end(); igs++)
				dgm = (dgm * deor[*igs])*reach*swapreach;

				/** Can it be the following code
				 for (set<string>::iterator igs=names_g.begin(); igs!=names_g.end(); igs++)
				 dgm *= deor[*igs];
				 dgm *= reach*swapreach;
				 */

				deogm[gi->first] = dgm;
			}

			if ( options["verbosity"]> 4 )
			cout << "Done." << endl;
		}

		if(need_distributed_knowledge) {
			if ( options["verbosity"]> 4 )
			cout << "    Building OBDD for distributed knowledge...";

			for(map<string, set<string> >::iterator gi=is_groups->begin(); gi!=is_groups->end(); gi++) {
				// Deontic
				tmpdistgm = bddmgr->bddOne();
				names_g = gi->second;
				for (set<string>::iterator igs=names_g.begin(); igs!=names_g.end(); igs++)
				tmpdistgm = (tmpdistgm * er[*igs])*reach*swapreach;

				/** Can it be the following code
				 for (set<string>::iterator igs=names_g.begin(); igs!=names_g.end(); igs++)
				 tmpdistgm *= er[*igs];
				 tmpdistgm *= reach*swapreach;
				 */

				distgm[gi->first] = tmpdistgm;
			}
			if ( options["verbosity"]> 4 )
			cout << "Done." << endl;
		}

		if ( options["verbosity"]> 4 )
		cout << "  Done: group modalities encoded successfully" << endl;
	}
	Cudd_AutodynEnable(bddmgr->getManager(), CUDD_REORDER_SIFT);

	if(options["quiet"] == 0)
	  cout << "Done." << endl;
	
	if(options["quiet"] == 0)
	  cout << "Checking formulae..." << endl;
	
	string str = "_init";
	(*is_evaluation)[str] = is_istates;
	modal_formula *init = new modal_formula(new atomic_proposition(&str));

	if(is_fairness->size()>0) {
	  if(options["quiet"] == 0 || options["simulation"] == 0)
	    cout << "Building set of fair states..." << endl;

	  if((options["bddcache"] == 2 || options["bddcache"] == 3) && symmetry_cache==NULL) {
	    symmetry_cache = new map<string, BDD>();
	    (*symmetry_cache)["_init"] = in_st;
	  }
	  
	  modal_formula *neg_init = new modal_formula(3,init);
	  
	  modal_formula *taut = new modal_formula(2,init,neg_init);
	  
	  for(vector<fairness_expression *>::iterator i=is_fairness->begin(); i!=is_fairness->end(); i++) {
	    BDD fairset = (*i)->check_formula();
	    (*i)->set_bdd_representation(fairset);
	  }
	  
	  modal_formula *eg_true = new modal_formula(11, taut); // This is EG(True)
	  
	  // This is a set of "fair" states (the beginning of a fair computation)
	  fair_reach = eg_true->check_formula_fair(bddmgr->bddOne());
	  
	  if (options["debug"] == 1) {
	    cout << "DEBUG: the set of fair states (beginning of a fair computation)" << endl;
	    fair_reach.print(v->count(),2);
	    cout << "DEBUG end" << endl;
	  }
	} // end if
	
	countex = new vector< vector<int> *>;
	idbdd = new map< int, BDD *>;
	cextr = new vector< vector<transition *> * >;

	if(options["quiet"] == 0 || options["simulation"] == 0)
	    cout << "Verifying properties..." << endl;

	BDD result;
	if ( options["globalcheck"] == 1) { // Check formula on all states
	  for(unsigned int i=0; i<is_formulae->size(); i++) {
	    result= is_fairness->size()>0 ?
	      (*is_formulae)[i].check_formula_fair(fair_reach)
	      :(*is_formulae)[i].check_formula();
	    if(options["quiet"] == 0)
	      cout << "  Formula number "<< i << ": " << (*is_formulae)[i].to_string()  << ", is " <<
		(result == reach ? "TRUE" : "FALSE") << " in the model" << endl;
	    else
	      cout << (result == reach ? "TRUE" : "FALSE") << endl;
	  }
	} if ( options["symmetry"] > 0) { // Check formula by symmetry reduction
	  symmetry_cache = new map<string, BDD>();
	  symmetry_atom_cache = new map<string, string>();
	  set<string>* formulas;
	  for(map<string, bool_expression*>::iterator i=is_evaluation->begin(); i!=is_evaluation->end(); i++)
	    if((i->first).find(".")==string::npos) {
	      if((i->first).compare("_init")!=0) {
		//cout << "encoding " << i->first << endl;
		BDD tmp=i->second->encode_bdd(bddOne)*reach;
		(*symmetry_cache)[i->first] = tmp;
		(*symmetry_atom_cache)[i->second->to_string()] = i->first;
	      } else {
		(*symmetry_cache)[i->first] = in_st;
	      }
	    }
	  vector<map<string, string>*>* permutations = (options["symmetry"] % 2) == 1 ? 
	    compute_round_permutation() : 
	    compute_arbitrary_permutation();
	  for(unsigned int i=0; i<is_formulae->size(); i++) {
	    bool flag = true;
	    if((options["symmetry"] % 2) == 0)
	      formulas = new set<string>;
	    for(unsigned int j=0; j<permutations->size(); j++) {
	      modal_formula *newformula = ((*is_formulae)[i]).permute_agents((*permutations)[j]);
	      string fstr = newformula->to_string();
	      if((options["symmetry"] % 2) == 0) {
		if(formulas->find(fstr)!=formulas->end()) {
		  //cout << "formula found:" << fstr << endl;
		  continue;
		}
		else
		  formulas->insert(fstr);
	      }
	      modal_formula * f = new modal_formula(4, init, newformula);
	      result = f->check_formula_symmetry(permutations);
	      if(result != reach) {
		flag = false;
		break;
	      }
	    }
	    cout << (flag ? "TRUE" : "FALSE") << endl;
	  }
	} else {
	  if((options["bddcache"] == 2 || options["bddcache"] == 3) && symmetry_cache==NULL) {
	    symmetry_cache = new map<string, BDD>();
	    (*symmetry_cache)["_init"] = in_st;
	  }

	  for(unsigned int i=0; i<is_formulae->size(); i++) {
	    scount = 0;
	    modal_formula * f = new modal_formula(4, init, &((*is_formulae)[i]));
	    result=is_fairness->size()>0 ?
	      f->check_formula_fair(fair_reach)
	      : f->check_formula();
	    
	    if ( options["debug"] == 1 ) {
	      cout << "DEBUG RESULT: " << endl;
	      //result.print(2,2);
	      cout << "REACH" << endl;
	      reach.print(2,2);
	      cout << "DEBUG end" << endl;
	    }
	    
	    if(options["quiet"] == 0)
	      cout << "  Formula number "<< i << ": " << (*is_formulae)[i].to_string()  << ", is " <<
		(result == reach ? "TRUE" : "FALSE") << " in the model" << endl;
	    else
	      cout <<(result == reach ? "TRUE" : "FALSE") << endl;

	    if ( (options["cex"] == 1) ) {
	      
	      // Computing counterexample / witness
	      
	      BDD * is = new BDD(in_st);
	      
	      countex->clear();
	      idbdd->clear();
	      scount = 0;
	      cextr->clear();
	      
	      modal_formula * g = f->push_negations(0);
	      
	      cextr->clear();
	      
	      
	      if ( ( result==reach ) && !(g->is_NoCTL()) && ( g->is_ECTLK_new() ) ) {
		
		// True and ECTLK: can print witness
		
		bool done = false;
		
		if(options["quiet"] == 0)
		  cout << "  The following is a witness for the formula: " << endl;
		
		while ( (*is) != bddmgr->bddZero()) {
		  
		  BDD curinitstate = is->PickOneMinterm(*v);
		  
		  if ( options["debug"] == 1 ) {
		    cout << "DEBUG Computing witness with init. state : " << endl;
		    cout << " ------------------------------" << endl;
		    print_state(curinitstate,*v);
		    cout << " ------------------------------" << endl;
		    cout << "DEBUG end" << endl;
		  }
		  
		  if (is_valid_state(curinitstate, *v)) {
		    //idbdd->insert(pair<int, BDD*>(0,new BDD(curinitstate)));
		    if ( options["debug"] == 1 ) {							
		      cout << "DEBUG Computing witness for formula: " << endl << g->to_string() << endl;
		      cout << "DEBUG end" << endl;
		    }	
		    bool cexresult = (is_fairness->size()>0)? g->build_cex_fair(&curinitstate, 0, fair_reach) : 
		      g->build_cex(&curinitstate, 0);
		    if ( cexresult ) {
		      done = true;
		      break;
		    }
		    else {
		      *is = *is - curinitstate;
		      idbdd->clear();
		      countex->clear();
		      cextr->clear();
		      scount = 0;
		    }
		  }
		}
		
		if ( !done ) {
		  if(options["quiet"] == 0) {
		    cout << " Uhm there is a problem here: I could not compute a witness." << endl;	
		    cout << " Please report this problem to the authors." << endl;
		  }
		}
		else {
		  if(options["quiet"] == 0) {
		    
		    cout << " Witness: " << endl;
		    for(int ac = 0;ac < (int)countex->size();ac++){
		      cout << "   < ";
		      for (int j=0; j< (int)countex->at(ac)->size(); j++) {
			cout << countex->at(ac)->at(j) << " ";
		      }							
		      cout << "> " << endl;
		    }
		    
		    cout << "  States description: " << endl;
		    for(map<int,BDD *>::iterator iter = idbdd->begin(); iter != idbdd->end(); iter++)
		      {
			
			cout << "------------- State: " << (*iter).first << " -----------------" <<  endl;    
			print_state(*(iter->second),*v);
			cout << "----------------------------------------" << endl;
		      }
		  } else {
		    // Writing full witness/counterexample to text file
		    
		    ostringstream s, s2;
		    s << "formula" << (i+1) << ".dot";
		    s2 << "formula" << (i+1) << ".info";
		    string filename  = cex_prefix + s.str();
		    string filename2 = cex_prefix + s2.str();
		    ofstream myfile(filename.c_str());
		    ofstream myfile2(filename2.c_str());
		    
		    if (myfile.is_open() && myfile2.is_open())
		    {
		      myfile << "digraph witness" << i << " {\n";

		    	// print states and their description
		    	for(map<int,BDD *>::iterator iter = idbdd->begin(); iter != idbdd->end(); iter++)
		    	{
		    		myfile << " " << (*iter).first << " [label=\""<<(*iter).first<<"\",shape=ellipse]" << endl;
		    		myfile2 << "-- State " << (*iter).first << " --" << endl;
		    		myfile2 << state_to_str(*(iter->second),*v) <<  endl;    
		    	}
		    	
		    	// printing transitions
		    	// HERE
				for(int ac = 0;ac < (int)cextr->size();ac++){
				  for (int j=0; j< (int)cextr->at(ac)->size(); j++) {
				    cextr->at(ac)->at(j)->to_str(&myfile);
				  }							
				}

		    	myfile << "}\n";

		    	myfile.close();
		    	myfile2.close();
		    }
		    else cout << "Unable to open file";

		    
		    
		  }
		}
		
	      }
	      else if ( ( result!=reach ) && ( g->is_ACTLK_new() ) ) {
		
		// False and ACTLK: can print counterexample
		
		if(options["quiet"] == 0)
		  cout << "  The following is a counterexampe for the formula " << endl;
		bool done = false;
		
		
		// Negation of the formula:
		modal_formula * fcex = new modal_formula(3, g);
		fcex = fcex->push_negations(0);

		//cout << fcex->to_string() << endl;
		
		
		while ( (*is) != bddmgr->bddZero()) {
		  
		  BDD curinitstate = is->PickOneMinterm(*v);
		  
		  if ( options["debug"] == 1 ) {
		    cout << "DEBUG Computing counterexample with init. state : " << endl;
		    cout << " ------------------------------" << endl;
		    print_state(curinitstate,*v);
		    cout << " ------------------------------" << endl;
		    cout << "DEBUG end" << endl;
		  }
		  
		  if (is_valid_state(curinitstate, *v)) {
		    //idbdd->insert(pair<int, BDD*>(0,new BDD(curinitstate)));
		    
		    
		    if ( options["debug"] == 1 ) {							
		      cout << "DEBUG Computing counterexample for formula: " << endl << fcex->to_string() << endl;
		      cout << "DEBUG end" << endl;
		    }
		    bool cexresult = (is_fairness->size()>0)? fcex->build_cex_fair(&curinitstate, 0, fair_reach) : 
		      fcex->build_cex(&curinitstate, 0);
		    if ( cexresult ) {
		      done = true;
		      break;
		    }
		    else {
		      *is = *is - curinitstate;
		      idbdd->clear();
		      countex->clear();
		      cextr->clear();
		      scount = 0;
		    }
		  }
		}
		
		if ( !done ) {
		  if(options["quiet"] == 0) {
		    cout << " Uhm there is a problem here: I could not compute a counterexample." << endl;	
		    cout << " Please report this problem to the authors." << endl;
		  }
		}
		else {
		  
		  if(options["quiet"] == 0) {
		    cout << " Counterexample: " << endl;
		    for(int ac = 0;ac < (int)countex->size();ac++){
		      cout << "   < ";
		      for (int j=0; j< (int)countex->at(ac)->size(); j++) {
			cout << countex->at(ac)->at(j) << " ";
		      }							
		      cout << "> " << endl;
		    }
		    
		    cout << "  States description: " << endl;
		    for(map<int,BDD *>::iterator iter = idbdd->begin(); iter != idbdd->end(); iter++)
		      {
			
			cout << "------------- State: " << (*iter).first << " -----------------" <<  endl;    
			print_state(*(iter->second),*v);
			cout << "----------------------------------------" << endl;
		      }
		    
		  } else {
		    
		    // Writing full witness/counterexample to text file
		    
		    ostringstream s, s2;
		    s << "formula" << (i+1) << ".dot";
		    s2 << "formula" << (i+1) << ".info";
		    string filename  = cex_prefix + s.str();
		    string filename2 = cex_prefix + s2.str();
		    ofstream myfile(filename.c_str());
		    ofstream myfile2(filename2.c_str());
		    
		    if (myfile.is_open() && myfile2.is_open())
		    {
		    	myfile << "digraph witness" << i << " {\n";
		    
		    	// print states and their description
		    	for(map<int,BDD *>::iterator iter = idbdd->begin(); iter != idbdd->end(); iter++)
		    	{
		    		myfile << " " << (*iter).first << " [label=\""<<(*iter).first<<"\",shape=ellipse]" << endl;
		    		myfile2 << "-- State " << (*iter).first << " --" << endl;
		    		myfile2 << state_to_str(*(iter->second),*v) <<  endl;    
		    	}
		    	
		    	// printing transitions
		    	// HERE
				for(int ac = 0;ac < (int)cextr->size();ac++){
				  for (int j=0; j< (int)cextr->at(ac)->size(); j++) {
				    cextr->at(ac)->at(j)->to_str(&myfile);
				  }							
				}

		    	myfile << "}\n";

		    	myfile.close();
		    	myfile2.close();
		    }
		    else cout << "Unable to open file";

		    
		    
		  }
		}
	      }
	      else if (result==reach) {
		if(options["quiet"] == 0)
		  cout << "    -- Sorry it is not possible to compute witnesses for non-ECTLK formulae" << endl;
	      }
	      else if (result!=reach) {
		if(options["quiet"] == 0)
		  cout << "    -- Sorry it is not possible to compute counterexamples for non-ACTLK formulae" << endl;
	      }
	    }
	    
	    if ( (options["cex"] == 1) && ( result != reach )) {
	      BDD c = reach*(!result);
	      
	      if ( !(is_fairness->size()>0) ) {
		fair_reach = reach;
	      }
	      
	      //f.build_cex();
	    }
	    
	  }
	}
	
	if(options["quiet"] == 0)
	  cout << "done, " << is_formulae->size() <<
	    " formulae successfully read and checked" << endl;
	time_t t1 = time( 0 );
	//if(options["verbosity"] > 4) {
	  cout << "execution time = " << difftime(t1, t0) << endl;
	  double reachablestates = reach.CountMinterm(v->count());
	  cout << "number of reachable states = " << reachablestates << endl;
	  unsigned long memoryinuse = bddmgr->ReadMemoryInUse();
	  cout << "BDD memory in use = " << memoryinuse << endl;
	  //}
	if ( options["bdd_stats"] == 1 ) {
	  bddmgr->info();
	}
	
}
