%skeleton "lalr1.cc"                          /*  -*- C++ -*- */
%require "2.3"
%defines
%define "parser_class_name" "mcmas_parser"

%{
#define USE(VALUE) /*empty*/

# include <string>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <algorithm>
#include "types.hh"

class mcmas_driver;
%}

%parse-param { mcmas_driver& driver }
%lex-param   { mcmas_driver& driver }

%locations
%initial-action
{
  // Initialize the initial location.
  @$.begin.filename = @$.end.filename = &driver.file;
};

%debug
%error-verbose

%union {
  int ival;
  std::string* sval;
  map<string, basictype *>* map_basictype;
  basictype* _basictype;
  set<string>* set_string;
  vector<evolution_line*>* vector_evolution_line;
  evolution_line* _evolution_line;
  vector<assignment*>* vector_assignment;
  bool_expression* _bool_expression;
  arithmetic_expression* _arithmetic_expression;
  expression* _expression;
  basic_agent* _basic_agent;
  vector<protocol_line*>* vector_protocol_line;
  protocol_line* _protocol_line;
  unsigned char _unsigned_char;
  bool_value* _bool_value;
  modal_formula* _modal_formula;
  fairness_expression* _fairness_expression;
  pair<modal_formula*, modal_formula*>* dual_modal_formula;
  pair<string*, modal_formula*>* ID_modal_formula;
  pair<pair<string*, string*>*, modal_formula*>* ID_ID_modal_formula;
  pair<fairness_expression*, fairness_expression*>* dual_fairness_expression;
  pair<string*, fairness_expression*>* ID_fairness_expression;
  pair<pair<string*, string*>*, fairness_expression*>* ID_ID_fairness_expression;
}

%{
#include "mcmas-driver.hh"

#define yyclearin (yychar == yyempty_)

  //extern Environment *is_environment;
  extern map<string, basic_agent *> *is_agents;
  extern vector<basic_agent *> *agents;
  extern map<string, bool_expression *> *is_evaluation;
  extern bool_expression *is_istates;
  extern map<string, set<string> > *is_groups;
  extern vector<modal_formula> *is_formulae;
  extern vector<fairness_expression *> *is_fairness;
%}

%token        END_OF_FILE      0 "end of file"
%token AGENT ACTIONS PROTOCOL EVOLUTION END LB RB LCB RCB LT GT IF EQ AND
%token ENVIRONMENT OBSVARS VARS BOOLEAN BTRUE BFALSE TO LE GE NEQ MINUS PLUS TIMES DIVIDE /* new symbols */
%token ACTION REDSTATES GREENSTATES FAIRNESS OTHER /* new symbols */
%token OR NOT COLON COMMA DOT SEMICOLON EVALUATION INITSTATES GROUPS
%token FORMULAE IMPLIES AG EG AX EX AF EF A E UNTIL K GK GCK O KH DK
%token DKH X F G 
%token <sval> ID "identifier"
%token <ival> NUMBER "number"

%left LB RB LT GT LE GE NEQ
%left ID
%right IMPLIES
%left OR 
%left AND
%left NOT
%left EX AX EF AF EG AG E A UNTIL 
%left K GK GCK O KH DK DKH X F G
%left EQ 
%left DOT

%type <map_basictype> obsvardef
%type <map_basictype> varidlist
%type <_basictype> onevardef
%type <ival> integer
%type <set_string> enumlist
%type <map_basictype> vardef
%type <map_basictype> envardef
%type <vector_evolution_line> envevdef
%type <vector_evolution_line> envevdeflist
%type <_evolution_line> envevline
%type <vector_assignment> boolresult
%type <vector_assignment> boolresult1
%type <_expression> expr1
%type <_expression> term1
%type <_expression> element1
%type <_expression> expr2
%type <_expression> term2
%type <_expression> element2
%type <_expression> expr3
%type <_expression> term3
%type <_expression> element3
%type <_bool_expression> lboolcond
%type <_bool_expression> eboolcond
%type <_bool_expression> rboolcond
%type <_bool_expression> reddef
%type <_expression> varvalue1
%type <_expression> varvalue2
%type <_expression> varvalue3
%type <_expression> varvalue4
%type <_basic_agent> agent
%type <set_string> actiondef
%type <set_string> enactiondef
%type <set_string> actionidlist
%type <vector_protocol_line> protdef
%type <vector_protocol_line> enprotdef
%type <vector_protocol_line> protdeflist
%type <_protocol_line> protline
%type <_protocol_line> otherbranch
%type <set_string> enabledidlist
%type <vector_evolution_line> evdef
%type <vector_evolution_line> evdeflist
%type <_evolution_line> evline
%type <_bool_expression> gboolcond
%type <_unsigned_char> logicop
%type <_bool_value> boolvalue
%type <_bool_expression> evaboolcond
%type <_bool_expression> isboolcond
%type <set_string> namelist
%type <sval> agentname
%type <_modal_formula> formula
%type <_fairness_expression> fformula

%type <ival> redprefix
%type <ival> redsuffix
%type <ival> envprefix
%type <ival> envsuffix
%type <ival> groupprefix
%type <ival> groupsuffix
%type <ival> actprefix
%type <ival> actsuffix
%type <ival> evprefix
%type <ival> evsuffix
%type <ival> iniprefix
%type <ival> inisuffix
%type <ival> forprefix
%type <ival> forsuffix
%type <ival> fairprefix
%type <ival> fairsuffix
%type <ival> evaprefix
%type <ival> evasuffix
%type <ival> protprefix
%type <ival> protsuffix
%type <ival> plprefix
%type <ival> plsuffix
%type <ival> obsprefix
%type <ival> obssuffix
%type <ival> varprefix
%type <ival> varsuffix
%type <ival> agprefix
%type <ival> agsuffix
%type <sval> atlprefix
%type <dual_modal_formula> untilprefix
%type <ID_modal_formula> epistemicprefix
%type <ID_modal_formula> gepistemicprefix
%type <ID_ID_modal_formula> khprefix
%type <ID_ID_modal_formula> dkhprefix
%type <dual_fairness_expression> untilprefixfair
%type <ID_fairness_expression> epistemicprefixfair
%type <ID_fairness_expression> gepistemicprefixfair
%type <ID_ID_fairness_expression> khprefixfair
%type <ID_ID_fairness_expression> dkhprefixfair
%type <ival> gplprefix
%type <ival> gplsufffix


%printer    { debug_stream () << *$$; } "identifier"
%destructor { delete $$; } "identifier"
%destructor { delete $$; } obsvardef
%destructor { delete $$; } varidlist
%destructor { delete $$; } onevardef
%destructor { delete $$; } enumlist
%destructor { delete $$; } vardef
%destructor { delete $$; } envardef
%destructor { delete $$; } envevdef
%destructor { delete $$; } envevdeflist
%destructor { delete $$; } envevline
%destructor { delete $$; } boolresult
%destructor { delete $$; } lboolcond
%destructor { delete $$; } eboolcond
%destructor { delete $$; } rboolcond
%destructor { delete $$; } reddef
%destructor { delete $$; } varvalue1
%destructor { delete $$; } varvalue2
%destructor { delete $$; } varvalue3
%destructor { delete $$; } varvalue4
%destructor { delete $$; } agent
%destructor { delete $$; } actiondef
%destructor { delete $$; } enactiondef
%destructor { delete $$; } actionidlist
%destructor { delete $$; } protdef
%destructor { delete $$; } enprotdef
%destructor { delete $$; } protdeflist
%destructor { delete $$; } protline
%destructor { delete $$; } enabledidlist
%destructor { delete $$; } evdef
%destructor { delete $$; } evdeflist
%destructor { delete $$; } evline
%destructor { delete $$; } gboolcond
%destructor { delete $$; } boolvalue
%destructor { delete $$; } evaboolcond
%destructor { delete $$; } isboolcond
%destructor { delete $$; } namelist
%destructor { delete $$; } agentname
%destructor { delete $$; } formula

%printer    { debug_stream () << $$; } "number"

%%
%start is;
is: environment agents evaluation istates groups fairness formulae;

/* definition of agent Enviroment */
envprefix: AGENT ENVIRONMENT {
  $$ = 1;
 }
| error ENVIRONMENT {
  $$ = 0;
  driver.syntax_error = true;
 }
| AGENT error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
envsuffix: END AGENT {
  $$ = 1;
 }
| error AGENT {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };

environment: envprefix obsvardef envardef reddef enactiondef enprotdef envevdef envsuffix {
  if($1==1 && $2!=NULL && $3!=NULL && /* $4!=NULL && */ $5!=NULL && $6!=NULL && $7!=NULL && $8==1) { // redstates is NULL if no red state at all
    //is_environment = new Environment($2, $3, $4, $5, $6, $7);
    basic_agent *env = new basic_agent(new string("Environment"), $2, $3, $4, $5, $6, $7);
    pair<string, basic_agent *> *a = new pair<string, basic_agent *>("Environment", env);
    is_agents->insert(*a);
    agents->push_back(env);
  }
 }
| error {
  driver.syntax_error = true;
  yyclearin;
  };

/* definitio of observable variables */
obsprefix: OBSVARS COLON {
  $$ = 1;
 }
| error COLON {
  $$ = 0;
  driver.syntax_error = true;
 }
| OBSVARS error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
obssuffix: END OBSVARS {
  $$ = 1;
 }
| error OBSVARS {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
obsvardef: obsprefix varidlist obssuffix {
  if($1==1 && $3==1)
    $$ = $2;
  else
    $$ = NULL;
 }
| obsprefix obssuffix {
  if($1==1 && $2==1)
    $$ = new map<string, basictype *>;
  else
    $$ = NULL;
 };
varidlist: onevardef SEMICOLON {
  if($1!=NULL) {
    $$ = new map<string, basictype *>;
    pair<string, basictype *> *a = new pair<string, basictype *>($1->get_name(), $1);
    $$->insert(*a);
  } else
    $$ = NULL;
 }
| onevardef error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| varidlist onevardef SEMICOLON {
  if($1!=NULL && $2!=NULL) {
    pair<string, basictype *> *a = new pair<string, basictype *>($2->get_name(), $2);
    pair<map<string, basictype *>::iterator,bool> p = $1->insert(*a);
    if(!p.second)
      cout << "warning: type " << $2->get_name() << " has already been defined." << endl;
    $$ = $1;
  } else
    $$ = NULL;
 }
| varidlist onevardef error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
  USE($2);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

onevardef: ID COLON BOOLEAN {
  if($1!=NULL)
    $$ = new basictype($1);
  else
    $$ = NULL;
 }
| ID COLON error {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID error BOOLEAN {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID COLON integer TO integer {
  if($1!=NULL) {
    if($3>$5)
      cout << "warning: the lower bound " << $3 << " is greater than the upper bound " << $5 << endl;
    $$ = new rangedint($1, $3, $5);
  }
  else
    $$ = NULL;
 }
| ID error integer TO integer {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID COLON integer error integer {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID error integer error integer {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID COLON LCB enumlist RCB {
  if($1!=NULL && $4!=NULL)
    $$ = new enumerate($1, $4);
  else
    $$ = NULL;
 }
| ID error enumlist RCB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($3);
 }
| ID COLON LCB enumlist error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| ID error enumlist error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($3);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
 }; 
enumlist: ID {
  if($1!=NULL) {
    $$ = new set<string>;
    $$->insert(*$1);
  } else
    $$ = NULL;
 }
| enumlist COMMA ID {
  if($1!=NULL && $3!=NULL) {
    pair<set<string>::iterator, bool> ret = $1->insert(*$3);
    if(!ret.second)
      cout << "warning: " << *$3 << " has been defined for the enum type." << endl;
    $$ = $1;
  } else
    $$ = NULL;
	USE($1);
 }
| enumlist error ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
integer: NUMBER {
  $$ = $1;
 }
| MINUS NUMBER {
  $$ = 0-$2;
 }
| error NUMBER {
  $$ = -1;
  driver.syntax_error = true;
 }
| error {
  $$ = -1;
  driver.syntax_error = true;
  yyclearin;
  };
/* definition of non-observable variables in Environment.
   can be null*/
varprefix: VARS COLON {
  $$ = 1;
 }
| error COLON {
  $$ = 0;
  driver.syntax_error = true;
 }
| VARS error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
varsuffix: END VARS {
  $$ = 1;
 }
| error VARS {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
envardef: varprefix varidlist varsuffix {
  if($1==1 && $3==1)
    $$ = $2;
  else
    $$ = NULL;
 }
| varprefix varsuffix {
  if($1==1 && $2==1)
    $$ = new map<string, basictype *>;
  else
    $$ = NULL;
 } 
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

/* definition of non-observable variables */
vardef: varprefix varidlist varsuffix {
  if($1==1 && $3==1)
    $$ = $2;
  else
    $$ = NULL;
 } 
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

/* Definition of red states */
redprefix: REDSTATES COLON {
  $$ = 1;
 }
| REDSTATES error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error COLON {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
redsuffix: END REDSTATES {
  $$ = 1;
 }
| error REDSTATES {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };

reddef: redprefix rboolcond redsuffix {
  if($1==1 && $3==1)
    $$ = $2;
  else
    $$ = NULL;
 } 
| redprefix redsuffix {
  $$ = NULL; // Be careful when processing red states!!!!!!!!!!!!!!
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
rboolcond: lboolcond SEMICOLON {
  $$ = $1;
  }
| lboolcond error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

/* Evolution definition for Environment.
   can be null. */
evprefix: EVOLUTION COLON {
  $$ = 1;
 }
| error COLON {
  $$ = 0;
  driver.syntax_error = true;
 }
| EVOLUTION error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
evsuffix: END EVOLUTION {
  $$ = 1;
 }
| error EVOLUTION {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
envevdef: evprefix envevdeflist evsuffix {
  if($1==1 && $3==1)
    $$ = $2;
  else
    $$ = NULL;
 }
| evprefix evsuffix {
  if($1==1 && $2==1)
    $$ = new vector<evolution_line *>;
  else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
envevdeflist: envevline {
  if($1!=NULL) {
    $$ = new vector<evolution_line *>;
    $$->push_back($1);
  } else
    $$ = NULL;
 }
| envevdeflist envevline {
  if($1!=NULL && $2!=NULL) {
    $1->push_back($2);
    $$ = $1;
  } else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
envevline: boolresult IF eboolcond SEMICOLON {
  if($1!=NULL && $3!=NULL) 
    $$ = new evolution_line($1, $3);
  else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
boolresult: LB boolresult RB {
  $$ = $2;
 }
| boolresult AND boolresult {
  if($1!=NULL && $3!=NULL) {
    $$ = $1;
    vector<assignment*>::iterator i;
    for(i=$3->begin(); i != $3->end(); i++)
      $$->push_back(*i);
  } else
    $$ = NULL;
 }
| ID EQ expr1 {
  if($1!=NULL && $3!=NULL) {
    assignment *a = new assignment($1, $3);
    $$ = new vector<assignment*>;
    $$->push_back(a);
  } else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  }; 

boolresult1: LB boolresult1 RB {
  $$ = $2;
 }
| boolresult1 AND boolresult1 {
  if($1!=NULL && $3!=NULL) {
    $$ = $1;
    vector<assignment*>::iterator i;
    for(i=$3->begin(); i != $3->end(); i++)
      $$->push_back(*i);
  } else
    $$ = NULL;
 }
| ID EQ expr2 {
  if($1!=NULL && $3!=NULL) {
    assignment *a = new assignment($1, $3);
    $$ = new vector<assignment*>;
    $$->push_back(a);
  } else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  }; 

agents: agent {
  if($1!=NULL) {
    pair<string, basic_agent *> *a = new pair<string, basic_agent *>($1->get_name(), $1);
    is_agents->insert(*a);
    agents->push_back($1);
  } 
 }
| agents agent {
  if($2!=NULL) {
    pair<string, basic_agent *> *a = new pair<string, basic_agent *>($2->get_name(), $2);
    pair<map<string, basic_agent *>::iterator,bool> p = is_agents->insert(*a);
    if(!p.second)
      cout << "warning: agent " << $2->get_name() << " already exists." << endl;
    else
      agents->push_back($2);
  }
 }
| error {
  driver.syntax_error = true;
  yyclearin;
  };

agprefix: AGENT  {
  $$ = 1;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
agsuffix: END AGENT {
  $$ = 1;
 }
| error AGENT {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };

agent: agprefix ID vardef reddef actiondef protdef evdef agsuffix {
  if($1==1 && $2!=NULL && $3!=NULL && /* $4!=NULL && */ 
     $5!=NULL && $6!=NULL && $7!=NULL && $8==1) // redstates is NULL if no red state at all
    $$ = new basic_agent($2, NULL, $3, $4, $5, $6, $7);
  else
    $$ = NULL;
 } 
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

/* ACTIONS in Environment.
   can be null */
actprefix: ACTIONS EQ LCB {
  $$ = 1;
 }
| error EQ LCB {
  $$ = 0;
  driver.syntax_error = true;
 }
| ACTIONS error LCB {
  $$ = 0;
  driver.syntax_error = true;
 }
| ACTIONS EQ error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error LCB {
  $$ = 0;
  driver.syntax_error = true;
 }
| ACTIONS error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error EQ error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
actsuffix: RCB SEMICOLON {
  $$ = 1;
 }
| RCB error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error SEMICOLON {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
enactiondef: actprefix actionidlist actsuffix {
  if($1==1 && $3==1)
    $$ = $2;
  else
    $$ = NULL;
 }
| actprefix actsuffix {
  if($1==1 && $2==1)
    $$ = new set<string>;
  else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

/* ACTIONS */
actiondef: actprefix actionidlist actsuffix {
  if($1==1 && $3==1) 
    $$ = $2;
  else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
actionidlist: ID {
  if($1!=NULL) {
    $$ = new set<string>;
    $$->insert(*$1);
  } else
    $$ = NULL;
 }
| actionidlist COMMA ID {
  if($1!=NULL && $3!=NULL) {
    pair<set<string>::iterator, bool> ret = $1->insert(*$3);
    if(!ret.second)
      cout << "warning: action " << *$3 << " has been defined." << endl;
    $$ = $1;
  } else
    $$ = NULL;
 }
| actionidlist error ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

/* PROTOCOL in Environment.
   can be null */
protprefix: PROTOCOL COLON {
  $$ = 1;
 }
| error COLON {
  $$ = 0;
  driver.syntax_error = true;
 }
| PROTOCOL error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
protsuffix: END PROTOCOL {
  $$ = 1;
 }
| error PROTOCOL {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };

enprotdef: protprefix protdeflist protsuffix {
  if($1==1 && $3==1) 
    $$ = $2;
  else
    $$ = NULL;
 }
| protprefix protdeflist otherbranch protsuffix {
  if($1==1 && $4==1) {
    $$ = $2;
    $$->push_back($3);
  }
  else
    $$ = NULL;
 }
| protprefix otherbranch protsuffix {
  if($1==1 && $3==1) {
    $$ = new vector<protocol_line*>;
    $$->push_back($2);;
  }
  else
    $$ = NULL;
 }
| protprefix protsuffix {
  if($1==1 && $2==1) 
    $$ = new vector<protocol_line*>;
  else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

/* PROTOCOL */
protdef: protprefix  protdeflist  protsuffix {
  if($1==1 && $3==1) 
    $$ = $2;
  else
    $$ = NULL;
 }
| protprefix protdeflist otherbranch protsuffix {
  if($1==1 && $4==1) {
    $$ = $2;
    $$->push_back($3);
  }
  else
    $$ = NULL;
 }
| protprefix otherbranch protsuffix {
  if($1==1 && $3==1) {
    $$ = new vector<protocol_line*>;
    $$->push_back($2);;
  }
  else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

protdeflist: protline {
  if($1!=NULL) {
    $$ = new vector<protocol_line*>;
    $$->push_back($1);
  } else
    $$ = NULL;
 }
| protdeflist protline {
  if($1!=NULL && $2!=NULL) {
    $1->push_back($2);
    $$ = $1;
  } else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

plprefix: COLON LCB {
  $$ = 1;
 }
| error LCB {
  $$ = 0;
  driver.syntax_error = true;
 }
| COLON error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
plsuffix: RCB SEMICOLON {
  $$ = 1;
 }
| error SEMICOLON {
  $$ = 0;
  driver.syntax_error = true;
 }
| RCB error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 };
protline: lboolcond plprefix enabledidlist plsuffix {
  if($1!=NULL && $2==1 && $3!=NULL && $4==1) 
    $$ = new protocol_line($1, $3);
  else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

otherbranch: OTHER plprefix enabledidlist plsuffix {
  if($2==1 && $3!=NULL && $4==1) 
    $$ = new protocol_line(new bool_expression(), $3);
  else
    $$ = NULL;
 }
| error plprefix enabledidlist plsuffix {
  $$ = NULL;
  driver.syntax_error = true;
  USE($3);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
 };

enabledidlist: ID {
  if($1!=NULL) {
    $$ = new set<string>;
    $$->insert(*$1);
  } else
    $$ = NULL;
 }
| enabledidlist COMMA ID {
  if($1!=NULL && $3!=NULL) {
    pair<set<string>::iterator, bool> ret = $1->insert(*$3);
    if(!ret.second)
      cout << "warning: action " << *$3 << " is repeated." << endl;
    $$ = $1;
  } else
    $$ = NULL;
 }
| enabledidlist error ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };


/* EVOLUTION DEFINITION */
evdef: evprefix evdeflist evsuffix {
  if($1==1 && $3==1)
    $$ = $2;
  else
    $$ = NULL;
}
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
evdeflist: evline {
  if($1!=NULL) {
    $$ = new vector<evolution_line*>;
    $$->push_back($1);
  } else
    $$ = NULL;
 }
| evdeflist evline {
  if($1!=NULL && $2!=NULL) {
    $1->push_back($2);
    $$ = $1;
  } else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

evline: boolresult1 IF gboolcond SEMICOLON {
  if($1!=NULL)
    $$ = new evolution_line($1, $3);
  else
    $$ = NULL;
 }
| boolresult1 error gboolcond SEMICOLON {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
  USE($3);
 }
| boolresult1 IF gboolcond error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
  USE($3);
 }
| boolresult1 error gboolcond error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
  USE($3);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

lboolcond: LB lboolcond RB { 
  $$ = $2; 
 }
| error lboolcond RB { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB lboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| error lboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| lboolcond AND lboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(1, $1, $3);
  else
    $$ = NULL;
 }
| lboolcond OR lboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(2, $1, $3);
  else
    $$ = NULL;
 }
| lboolcond error lboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
  USE($3);
 }
| NOT lboolcond {
  if($2!=NULL)
    $$ = new bool_expression(3, $2);
  else
    $$ = NULL;
 }
| error lboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| expr1 logicop expr1 {
  if($1!=NULL && $3!=NULL) {
    logic_expression *expr = new logic_expression($2, $1, $3);
    $$ = new bool_expression(0,  expr);
  } else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

expr1: expr1 PLUS term1 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(5, $1, $3);
  } else
    $$ = NULL;
 }
| expr1 MINUS term1 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(6, $1, $3);
  } else
    $$ = NULL;
 }
| expr1 error term1 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| term1 {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  };
term1: term1 TIMES element1 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(7, $1, $3);
  } else
    $$ = NULL;
 }
| term1 DIVIDE element1 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(8, $1, $3);
  } else
    $$ = NULL;
 }
| term1 error element1 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| element1 {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  };
element1: LB expr1 RB {
  $$ = $2;
  }
| error expr1 RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB expr1 error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| varvalue1  {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  };

expr2: expr2 PLUS term2 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(5, $1, $3);
  } else
    $$ = NULL;
 }
| expr2 MINUS term2 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(6, $1, $3);
  } else
    $$ = NULL;
 }
| expr2 error term2 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| term2 {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  };
term2: term2 TIMES element2 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(7, $1, $3);
  } else
    $$ = NULL;
 }
| term2 DIVIDE element2 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(8, $1, $3);
  } else
    $$ = NULL;
 }
| term2 error element2 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| element2 {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  };
element2: LB expr2 RB {
  $$ = $2;
  }
| error expr2 RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB expr2 error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| varvalue2  {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  };
expr3: expr3 PLUS term3 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(5, $1, $3);
  } else
    $$ = NULL;
 }
| expr3 MINUS term3 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(6, $1, $3);
  } else
    $$ = NULL;
 }
| expr3 error term3 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| term3 {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  };
term3: term3 TIMES element3 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(7, $1, $3);
  } else
    $$ = NULL;
 }
| term3 DIVIDE element3 {
  if($1!=NULL && $3!=NULL) {
    $$ = new arithmetic_expression(8, $1, $3);
  } else
    $$ = NULL;
 }
| term3 error element3 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| element3 {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  };
element3: LB expr3 RB {
  $$ = $2;
  }
| error expr3 RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB expr3 error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| varvalue3 {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  };



eboolcond: LB eboolcond RB { 
  $$ = $2; 
 }
| error eboolcond RB { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB eboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| error eboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| eboolcond AND eboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(1, $1, $3);
  else
    $$ = NULL;
 }
| eboolcond OR eboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(2, $1, $3);
  else
    $$ = NULL;
 }
| eboolcond error eboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
  USE($3);
 }
| NOT eboolcond {
  if($2!=NULL)
    $$ = new bool_expression(3, $2);
  else
    $$ = NULL;
 }
| error eboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| expr1 logicop expr1 {
  if($1!=NULL && $3!=NULL) {
    logic_expression *expr = new logic_expression($2, $1, $3);
    $$ = new bool_expression(0,  expr);
  } else
    $$ = NULL;
 }
| ACTION EQ ID {
  if($3!=NULL) {
    laction *a = new laction();
    variable *v = new variable($3);
    logic_expression *expr = new logic_expression(0, a, v);
    $$ = new bool_expression(0,  expr);
  } else
    $$ = NULL;
 }
| error EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ACTION error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID DOT ACTION EQ ID {
  if($1!=NULL && $5!=NULL) {
    laction *a = new laction($1);
    variable *v = new variable($1, $5);
    logic_expression *expr = new logic_expression(0, a, v);
    $$ = new bool_expression(0,  expr);
  } else
    $$ = NULL;
 }
| ENVIRONMENT DOT ACTION EQ ID {
  if($5!=NULL) {
    laction *a = new laction(new string("Environment"));
    variable *v = new variable(new string("Environment"), $5);
    logic_expression *expr = new logic_expression(0, a, v);
    $$ = new bool_expression(0,  expr);
  } else
    $$ = NULL;
 }
| ID error ACTION EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID DOT error EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID DOT ACTION error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID error EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID DOT error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID error ACTION error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

logicop: LT {
  $$ = 2;
 }
| LE {
  $$ = 3;
  }
| GT {
  $$ = 4;
  }
| GE {
  $$ = 5;
  }
| EQ {
  $$ = 0;
  }
| NEQ {
  $$ = 1;
  }
| error {
  $$ = 0;
  driver.syntax_error = true;
  };

/* global boolean condition definition used in normal agents */
gboolcond: LB gboolcond RB { 
  $$ = $2; 
 }
| error gboolcond RB { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB gboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| error gboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| gboolcond AND gboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(1, $1, $3);
  else
    $$ = NULL;
 }
| gboolcond OR gboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(2, $1, $3);
  else
    $$ = NULL;
 }
| gboolcond error gboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
  USE($3);
 }
| NOT gboolcond {
  if($2!=NULL)
    $$ = new bool_expression(3, $2);
  else
    $$ = NULL;
 }
| error gboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| expr2 logicop expr2 {	
  if($1!=NULL && $3!=NULL) {
    logic_expression *expr = new logic_expression($2, $1, $3);
    $$ = new bool_expression(0,  expr);
  } else
    $$ = NULL;
 }
| ACTION EQ ID {
  if($3!=NULL) {
    laction *a = new laction();
    variable *v = new variable($3);
    logic_expression *expr = new logic_expression(0, a, v);
    $$ = new bool_expression(0,  expr);
  } else
    $$ = NULL;
 }
| error EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ACTION error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID DOT ACTION EQ ID {
  if($1!=NULL && $5!=NULL) {
    laction *a = new laction($1);
    variable *v = new variable($1, $5);
    logic_expression *expr = new logic_expression(0, a, v);
    $$ = new bool_expression(0,  expr);
  } else 
    $$ = NULL;
 }
| ID error ACTION EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID DOT error EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID DOT ACTION error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID error EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID DOT error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID error ACTION error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ENVIRONMENT DOT ACTION EQ ID {
  if($5!=NULL) {
    laction *a = new laction(new string("Environment"));
    variable *v = new variable(new string("Environment"), $5);
    logic_expression *expr = new logic_expression(0, a, v);
    $$ = new bool_expression(0,  expr);
  } else
    $$ = NULL;
 }
| error ACTION EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ENVIRONMENT DOT error EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ENVIRONMENT DOT ACTION error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ENVIRONMENT error EQ ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ENVIRONMENT DOT error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ENVIRONMENT error ACTION error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| ENVIRONMENT error ID {
  $$ = NULL;
  driver.syntax_error = true;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

/* Variable values */
varvalue1: boolvalue {
  $$ = $1;
 }
| ID {
  $$ = new variable($1); // can be enumerate value. Will be decided during global syntax checking
  }
| integer {
  $$ = new int_value($1);
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
boolvalue: BTRUE {
  $$ = new bool_value(true);
  }
| BFALSE {
  $$ = new bool_value(false);
  };
varvalue2: boolvalue {
  $$ = $1;
 }
| ID {
  $$ = new variable($1); // can be enumerate value. Will be decided during global syntax checking
  }
| ENVIRONMENT DOT ID {
  $$ = new variable(new string("Environment"), $3);
  }
| error DOT ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($3);
  }
| ENVIRONMENT error ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($3);
  }
| error ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
  }
| integer {
  $$ = new int_value($1);
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
varvalue3: boolvalue {
  $$ = $1;
 }
| ID {
  $$ = new variable($1); // can be enumerate value. Will be decided during global syntax checking
  }
| ID DOT ID {
  $$ = new variable($1, $3);
  }
| ID error ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($3);
  }
| ENVIRONMENT DOT ID {
  $$ = new variable(new string("Environment"), $3);
  }
| error DOT ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($3);
  }
| ENVIRONMENT error ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($3);
  }
| error ID {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
  }
| integer {
  $$ = new int_value($1);
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
/* Initial value of a variable */
varvalue4: boolvalue {
  $$ = $1;
 }
| ID {
  $$ = new enum_value($1); // can be enumerate value. Will be decided during global syntax checking
  }
| integer {
  $$ = new int_value($1);
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

/* EVALUATION */
evaprefix: EVALUATION {
  $$ = 1;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 }
evasuffix: END EVALUATION {
  $$ = 1;
 }
| error EVALUATION {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
  }
evaluation: evaprefix evalist evasuffix
| error {
  driver.syntax_error = true;
  yyclearin;
  };

evalist: ID IF evaboolcond SEMICOLON {
  if($1!=NULL && $3!=NULL) {
    pair<string, bool_expression *> *a = new pair<string, bool_expression *>(*$1, $3);
    is_evaluation->insert(*a);
  } 
 }
| ID error evaboolcond SEMICOLON {
  driver.syntax_error = true;
  USE($3);
 }
| ID IF evaboolcond error {
  driver.syntax_error = true;
  USE($3);
 }
| ID error evaboolcond error {
  driver.syntax_error = true;
  USE($3);
 }
| evalist ID IF evaboolcond SEMICOLON {
  if($2!=NULL && $4!=NULL) {
    pair<string, bool_expression *> *a = new pair<string, bool_expression *>(*$2, $4);
    pair<map<string, bool_expression *>::iterator, bool> p = is_evaluation->insert(*a);
    if(!p.second)
      cout << "warning: atomic proposition " << *$2 << " has already been defined." << endl;
  }
 }
| evalist ID error evaboolcond SEMICOLON {
  driver.syntax_error = true;
  USE($4);
 }
| evalist ID IF evaboolcond error {
  driver.syntax_error = true;
  USE($4);
 }
| evalist ID error evaboolcond error {
  driver.syntax_error = true;
  USE($4);
 }
| error {
  driver.syntax_error = true;
  yyclearin;
  };

evaboolcond: LB evaboolcond RB { 
  $$ = $2; 
 }
| error evaboolcond RB { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB evaboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| error evaboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| evaboolcond AND evaboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(1, $1, $3);
  else
    $$ = NULL;
 }
| evaboolcond OR evaboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(2, $1, $3);
  else
    $$ = NULL;
 }
| evaboolcond error evaboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);USE($3);
 }
| NOT evaboolcond {
  if($2!=NULL)
    $$ = new bool_expression(3, $2);
  else
    $$ = NULL;
 }
| error evaboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| expr3 logicop expr3 {	
  if($1!=NULL && $3!=NULL) {
    logic_expression *e = new logic_expression($2, $1, $3);
    $$ = new bool_expression(0, e);
  } else
    $$ = NULL;
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };



/* INITIAL STATES */
iniprefix: INITSTATES {
  $$ = 1;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 }
inisuffix: SEMICOLON END INITSTATES {
  $$ = 1;
 }
| error END INITSTATES {
  $$ = 0;
  driver.syntax_error = true;
 }
| SEMICOLON error INITSTATES {
  $$ = 0;
  driver.syntax_error = true;
 }
| SEMICOLON END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error INITSTATES {
  $$ = 0;
  driver.syntax_error = true;
 }
| SEMICOLON error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
  }
istates: iniprefix isboolcond inisuffix {
  if($1==1 && $3==1)
    is_istates = $2; 
 }
| error {
  driver.syntax_error = true;
  yyclearin;
  };

isboolcond: LB isboolcond RB { 
  $$ = $2; 
}
| error isboolcond RB { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB isboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| error isboolcond error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2)
 }
| isboolcond AND isboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(1, $1, $3);
  else
    $$ = NULL;
 }
| isboolcond OR isboolcond {
  if($1!=NULL && $3!=NULL)
    $$ = new bool_expression(2, $1, $3);
  else
    $$ = NULL;
 }
| isboolcond error isboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);USE($3);
 }
| NOT isboolcond {
  if($2!=NULL)
    $$ = new bool_expression(3, $2);
  else
    $$ = NULL;
 }
| error isboolcond {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| ID DOT ID EQ varvalue4 {
  if($1!=NULL && $3!=NULL && $5!=NULL) {
    variable *v = new variable($1, $3);
    logic_expression *e = new logic_expression(0, v, $5);
    $$ = new bool_expression(0, e);
  } else
    $$ = NULL;
 }
| ID error ID EQ varvalue4 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($5);
 }
| ID DOT ID EQ varvalue4 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($5);
 }
| ID error ID error varvalue4 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($5);
 }
| ENVIRONMENT DOT ID EQ varvalue4 {	
  if($3!=NULL && $5!=NULL) {
    variable *v = new variable(new string("Environment"), $3);
    logic_expression *e = new logic_expression(0, v, $5);
    $$ = new bool_expression(0, e);
  } else
    $$ = NULL;
 }
| ENVIRONMENT error ID EQ varvalue4 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($5);
 }
| ENVIRONMENT DOT ID EQ varvalue4 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($5);
 }
| ENVIRONMENT error ID error varvalue4 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($5);
 }
| error ID EQ varvalue4 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error DOT ID error varvalue4 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($5);
 }
| error ID error varvalue4 {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };


/* Groups (not necessary) */
groupprefix: GROUPS {
  $$ = 1;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 }
groupsuffix: END GROUPS {
  $$ = 1;
 }
| error GROUPS {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
  }
groups: groupprefix groupdeflist groupsuffix
| groupprefix  groupsuffix // allow not to define groups
| error {
  driver.syntax_error = true;
  yyclearin;
  };

gplprefix: EQ LCB {
  $$ = 1;
 }
| error LCB {
  $$ = 0;
  driver.syntax_error = true;
 }
| EQ error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
  };
gplsufffix: RCB SEMICOLON {
  $$ = 1;
 }
| error SEMICOLON {
  $$ = 0;
  driver.syntax_error = true;
 }
| RCB error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
  };
groupdeflist: ID gplprefix namelist gplsufffix {
  if($1!=NULL && $2==1 && $3!=NULL && $4==1) {
    pair<string, set<string> > *a = new pair<string, set<string> >(*$1, *$3);
    is_groups->insert(*a);
  }
 }
| groupdeflist ID gplprefix namelist gplsufffix {
  if($2!=NULL && $3==1 && $4!=NULL && $5==1) {
    pair<string, set<string> > *a = new pair<string, set<string> >(*$2, *$4);
    pair<map<string, set<string> >::iterator,bool> p = is_groups->insert(*a);
    if(!p.second)
      cout << "warning: group " << *$2 << " already exists." << endl;    
  }
 }
| error {
  driver.syntax_error = true;
  yyclearin;
  };

namelist: agentname {
  if($1!=NULL) {
    $$ = new set<string>;
    $$->insert(*$1);
  } else
    $$ = NULL;
 }
| namelist COMMA agentname {
  if($1!=NULL && $3!=NULL) {
    pair<set<string>::iterator, bool> ret = $1->insert(*$3);
    if(!ret.second)
      cout << "warning: agent " << *$3 << " in group definition is redundant." << endl;
    $$ = $1;
  } else
    $$ = NULL;
 }
| namelist error agentname {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);USE($3);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };
agentname: ENVIRONMENT {
  $$ = new string("Environment");
 }
| ID {
  $$ = $1;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };


/* FAIRNES FORMULAE (boolean) */
fairprefix: FAIRNESS {
  $$ = 1;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 }
fairsuffix: END FAIRNESS {
  $$ = 1;
 }
| error FAIRNESS {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
  }
fairness: fairprefix fairsuffix
| fairprefix fformlist fairsuffix;

fformlist: fformula SEMICOLON { 
  if($1!=NULL)
    is_fairness->push_back($1);
 }
| fformula error {
  driver.syntax_error = true;
  USE($1);
 }
| fformlist fformula SEMICOLON { 
  if($2!=NULL)
    is_fairness->push_back($2); 
 }
| fformlist fformula error {
  driver.syntax_error = true;
  USE($2);
 }
| error {
  driver.syntax_error = true;
  yyclearin;
  };

untilprefixfair: LB fformula UNTIL fformula RB{
  if($2!=NULL && $4!=NULL)
    $$ = new pair<fairness_expression *, fairness_expression *>($2, $4);
  else
    $$ = NULL;
 }
| error fformula UNTIL fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| LB fformula error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| LB fformula UNTIL fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| error fformula error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| error fformula UNTIL fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| LB fformula error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| error fformula error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 };
epistemicprefixfair: LB ID COMMA fformula RB{
  if($2!=NULL && $4!=NULL)
    $$ = new pair<string *, fairness_expression *>($2, $4);
  else
    $$ = NULL;
 }
| error ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ENVIRONMENT COMMA fformula RB{
  if($4!=NULL)
    $$ = new pair<string *, fairness_expression *>(new string("Environment"), $4);
  else
    $$ = NULL;
 }
| error ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 };
gepistemicprefixfair: LB ID COMMA fformula RB{
  if($2!=NULL && $4!=NULL)
    $$ = new pair<string *, fairness_expression *>($2, $4);
  else
    $$ = NULL;
 }
| error ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 };
khprefixfair: LB ID COMMA ID COMMA fformula RB {
  if($2!=NULL && $4!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>($2, $4);
    $$ = new pair<pair<string *, string *>*, fairness_expression *>(two, $6);
  } else
    $$ = NULL;
 }
| error ID COMMA ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ID COMMA fformula RB {
  if($4!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>(new string("Environment"), $4);
    $$ = new pair<pair<string *, string *>*, fairness_expression *>(two, $6);
  } else
    $$ = NULL;
 }
| error ENVIRONMENT COMMA ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT COMMA fformula RB {
  if($2!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>($2, new string("Environment"));
    $$ = new pair<pair<string *, string *>*, fairness_expression *>(two, $6);
  } else
    $$ = NULL;
 }
| error ID COMMA ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ENVIRONMENT COMMA fformula RB {
  if($6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>(new string("Environment"), new string("Environment"));
    $$ = new pair<pair<string *, string *>*, fairness_expression *>(two, $6);
  } else
    $$ = NULL;
 }
| error ENVIRONMENT COMMA ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 };
dkhprefixfair: LB ID COMMA ID COMMA fformula RB {
  if($2!=NULL && $4!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>($2, $4);
    $$ = new pair<pair<string *, string *>*, fairness_expression *>(two, $6);
  } else
    $$ = NULL;
 }
| error ID COMMA ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT COMMA fformula RB {
  if($2!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>($2, new string("Environment"));
    $$ = new pair<pair<string *, string *>*, fairness_expression *>(two, $6);
  } else
    $$ = NULL;
 }
| error ID COMMA ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT COMMA fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT error fformula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT COMMA fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT error fformula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 };

fformula: LB fformula RB { 
  $$ = $2; 
 }
| error fformula RB { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB fformula error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| error fformula error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| fformula AND fformula {
  if($1!=NULL && $3!=NULL)
    $$ = new fairness_expression(1, $1, $3);
  else
    $$ = NULL;
 }
| fformula OR fformula {
  if($1!=NULL && $3!=NULL)
    $$ = new fairness_expression(2, $1, $3);
  else
    $$ = NULL;
 }
| fformula error fformula {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);USE($3);
 }
| NOT fformula {
  if($2!=NULL)
    $$ = new fairness_expression(3, $2);
  else
    $$ = NULL;
 }
| error fformula {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| fformula IMPLIES fformula {
  if($1!=NULL && $3!=NULL)
    $$ = new fairness_expression(4, $1, $3);
  else
    $$ = NULL;
 }
| AG fformula {
  if($2!=NULL)
    $$ = new fairness_expression(10, $2);
  else
    $$ = NULL;
 }
| EG fformula {
  if($2!=NULL)
    $$ = new fairness_expression(11, $2);
  else
    $$ = NULL;
  }
| AX fformula {
  if($2!=NULL)
    $$ = new fairness_expression(12, $2);
  else
    $$ = NULL;
 }
| EX fformula {
  if($2!=NULL)
    $$ = new fairness_expression(13, $2);
  else
    $$ = NULL;
 }
| AF fformula {
  if($2!=NULL)
    $$ = new fairness_expression(14, $2);
  else
    $$ = NULL;
 }
| EF fformula {
  if($2!=NULL)
    $$ = new fairness_expression(15, $2);
  else
    $$ = NULL;
 }
| A untilprefixfair {
  if($2!=NULL)
    $$ = new fairness_expression(16, $2->first, $2->second);
  else
    $$ = NULL;
 }
| E untilprefixfair {
  if($2!=NULL)
    $$ = new fairness_expression(17, $2->first, $2->second);
  else
    $$ = NULL;
 }
| K epistemicprefixfair {
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    fairness_expression *a = new fairness_expression(ap);
    $$ = new fairness_expression(30, a, $2->second);
  } else
    $$ = NULL;
 }
| GK gepistemicprefixfair {
  // Knowledge in a group (everybody knows)
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    fairness_expression *a = new fairness_expression(ap);
    $$ = new fairness_expression(31, a, $2->second);
  } else
    $$ = NULL;
 }
| GCK gepistemicprefixfair {
  // Common knowledge in a group
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    fairness_expression *a = new fairness_expression(ap);
    $$ = new fairness_expression(32, a, $2->second);
  } else
    $$ = NULL;
 }
| O epistemicprefixfair {
  // deontic operator
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    fairness_expression *a = new fairness_expression(ap);
    $$ = new fairness_expression(40, a, $2->second);
  } else
    $$ = NULL;
 }
| KH khprefixfair {
  // Knowledge assuming correct behaviour
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first->first);
    atomic_proposition *ap1 = new atomic_proposition($2->first->second);
    fairness_expression *a = new fairness_expression(ap);
    fairness_expression *b = new fairness_expression(ap1);
    $$ = new fairness_expression(41, a, b, $2->second);
  } else
    $$ = NULL;
 }
| DKH dkhprefixfair {
  // Distributed knowledge assuming correct behaviour
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first->first);
    atomic_proposition *ap1 = new atomic_proposition($2->first->second);
    fairness_expression *a = new fairness_expression(ap);
    fairness_expression *b = new fairness_expression(ap1);
    $$ = new fairness_expression(43, a, b, $2->second);
  } else
    $$ = NULL;
 }
| error khprefixfair {
  $$ = NULL;
  driver.syntax_error = true;
 }
| error dkhprefixfair {
  $$ = NULL;
  driver.syntax_error = true;
 }
| DK gepistemicprefixfair {
  // Distributed knowledge in a group
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    fairness_expression *a = new fairness_expression(ap);
    $$ = new fairness_expression(42, a, $2->second);
  } else
    $$ = NULL;
 }
| error epistemicprefixfair { 
  $$ = NULL;
  driver.syntax_error = true;
 }
| error gepistemicprefixfair { 
  $$ = NULL;
  driver.syntax_error = true;
 }
| ID {	
  if($1!=NULL) {
    atomic_proposition *ap = new atomic_proposition($1);
    $$ = new fairness_expression(ap);
  } else
    $$ = NULL;
  }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };


/* FORMULAE TO CHECK */
forprefix: FORMULAE {
  $$ = 1;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
 }
forsuffix: END FORMULAE {
  $$ = 1;
 }
| error FORMULAE {
  $$ = 0;
  driver.syntax_error = true;
 }
| END error {
  $$ = 0;
  driver.syntax_error = true;
 }
| error {
  $$ = 0;
  driver.syntax_error = true;
  }
formulae: forprefix formlist forsuffix;

formlist: formula SEMICOLON { 
  if($1!=NULL)
    is_formulae->push_back(*$1);
 }
| formula error {
  driver.syntax_error = true;
  USE($1);
 }
| formlist formula SEMICOLON { 
  if($2!=NULL)
    is_formulae->push_back(*$2); 
 }
| formlist formula error {
  driver.syntax_error = true;
  USE($2);
 }
| error {
  driver.syntax_error = true;
  yyclearin;
  };

atlprefix: LT ID GT {
  $$ = $2;
 }
| LT ID error {
  $$ = NULL;
  driver.syntax_error = true;
 }
| error ID GT {
  $$ = NULL;
  driver.syntax_error = true;
 };
untilprefix: LB formula UNTIL formula RB{
  if($2!=NULL && $4!=NULL)
    $$ = new pair<modal_formula *, modal_formula *>($2, $4);
  else
    $$ = NULL;
 }
| error formula UNTIL formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| LB formula error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| LB formula UNTIL formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| error formula error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| error formula UNTIL formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| LB formula error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 }
| error formula error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);USE($4);
 };
epistemicprefix: LB ID COMMA formula RB{
  if($2!=NULL && $4!=NULL)
    $$ = new pair<string *, modal_formula *>($2, $4);
  else
    $$ = NULL;
 }
| error ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ENVIRONMENT COMMA formula RB{
  if($4!=NULL)
    $$ = new pair<string *, modal_formula *>(new string("Environment"), $4);
  else
    $$ = NULL;
 }
| error ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 };
gepistemicprefix: LB ID COMMA formula RB{
  if($2!=NULL && $4!=NULL)
    $$ = new pair<string *, modal_formula *>($2, $4);
  else
    $$ = NULL;
 }
| error ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| LB ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 }
| error ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($4);
 };
khprefix: LB ID COMMA ID COMMA formula RB {
  if($2!=NULL && $4!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>($2, $4);
    $$ = new pair<pair<string *, string *>*, modal_formula *>(two, $6);
  } else
    $$ = NULL;
 }
| error ID COMMA ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ID COMMA formula RB {
  if($4!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>(new string("Environment"), $4);
    $$ = new pair<pair<string *, string *>*, modal_formula *>(two, $6);
  } else
    $$ = NULL;
 }
| error ENVIRONMENT COMMA ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT COMMA formula RB {
  if($2!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>($2, new string("Environment"));
    $$ = new pair<pair<string *, string *>*, modal_formula *>(two, $6);
  } else
    $$ = NULL;
 }
| error ID COMMA ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ENVIRONMENT COMMA formula RB {
  if($6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>(new string("Environment"), new string("Environment"));
    $$ = new pair<pair<string *, string *>*, modal_formula *>(two, $6);
  } else
    $$ = NULL;
 }
| error ENVIRONMENT COMMA ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT COMMA ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT COMMA ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ENVIRONMENT error ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ENVIRONMENT error ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 };
dkhprefix: LB ID COMMA ID COMMA formula RB {
  if($2!=NULL && $4!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>($2, $4);
    $$ = new pair<pair<string *, string *>*, modal_formula *>(two, $6);
  } else
    $$ = NULL;
 }
| error ID COMMA ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ID error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT COMMA formula RB {
  if($2!=NULL && $6!=NULL) {
    pair<string *, string *> *two = new pair<string *, string *>($2, new string("Environment"));
    $$ = new pair<pair<string *, string *>*, modal_formula *>(two, $6);
  } else
    $$ = NULL;
 }
| error ID COMMA ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT COMMA formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID COMMA ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT error formula RB {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT COMMA formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID COMMA ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| LB ID error ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 }
| error ID error ENVIRONMENT error formula error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($6);
 };

formula: LB formula RB { 
  $$ = $2; 
 }
| error formula RB { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| LB formula error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| error formula error { 
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| formula AND formula {
  if($1!=NULL && $3!=NULL)
    $$ = new modal_formula(1, $1, $3);
  else
    $$ = NULL;
 }
| formula OR formula {
  if($1!=NULL && $3!=NULL)
    $$ = new modal_formula(2, $1, $3);
  else
    $$ = NULL;
 }
| formula error formula {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);USE($3);
 }
| NOT formula {
  if($2!=NULL)
    $$ = new modal_formula(3, $2);
  else
    $$ = NULL;
 }
| error formula {
  $$ = NULL;
  driver.syntax_error = true;
  USE($2);
 }
| formula IMPLIES formula {
  if($1!=NULL && $3!=NULL)
    $$ = new modal_formula(4, $1, $3);
  else
    $$ = NULL;
 }
| AG formula {
  if($2!=NULL)
    $$ = new modal_formula(10, $2);
  else
    $$ = NULL;
 }
| EG formula {
  if($2!=NULL)
    $$ = new modal_formula(11, $2);
  else
    $$ = NULL;
  }
| AX formula {
  if($2!=NULL)
    $$ = new modal_formula(12, $2);
  else
    $$ = NULL;
 }
| EX formula {
  if($2!=NULL)
    $$ = new modal_formula(13, $2);
  else
    $$ = NULL;
 }
| AF formula {
  if($2!=NULL)
    $$ = new modal_formula(14, $2);
  else
    $$ = NULL;
 }
| EF formula {
  if($2!=NULL)
    $$ = new modal_formula(15, $2);
  else
    $$ = NULL;
 }
| A untilprefix {
  if($2!=NULL)
    $$ = new modal_formula(16, $2->first, $2->second);
  else
    $$ = NULL;
 }
| E untilprefix {
  if($2!=NULL)
    $$ = new modal_formula(17, $2->first, $2->second);
  else
    $$ = NULL;
 }
| K epistemicprefix {
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    modal_formula *a = new modal_formula(ap);
    $$ = new modal_formula(30, a, $2->second);
  } else
    $$ = NULL;
 }
| GK gepistemicprefix {
  // Knowledge in a group (everybody knows)
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    modal_formula *a = new modal_formula(ap);
    $$ = new modal_formula(31, a, $2->second);
  } else
    $$ = NULL;
 }
| GCK gepistemicprefix {
  // Common knowledge in a group
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    modal_formula *a = new modal_formula(ap);
    $$ = new modal_formula(32, a, $2->second);
  } else
    $$ = NULL;
 }
| O epistemicprefix {
  // deontic operator
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    modal_formula *a = new modal_formula(ap);
    $$ = new modal_formula(40, a, $2->second);
  } else
    $$ = NULL;
 }
| KH khprefix {
  // Knowledge assuming correct behaviour
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first->first);
    atomic_proposition *ap1 = new atomic_proposition($2->first->second);
    modal_formula *a = new modal_formula(ap);
    modal_formula *b = new modal_formula(ap1);
    $$ = new modal_formula(41, a, b, $2->second);
  } else
    $$ = NULL;
 }
| DKH dkhprefix {
  // Distributed knowledge assuming correct behaviour
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first->first);
    atomic_proposition *ap1 = new atomic_proposition($2->first->second);
    modal_formula *a = new modal_formula(ap);
    modal_formula *b = new modal_formula(ap1);
    $$ = new modal_formula(43, a, b, $2->second);
  } else
    $$ = NULL;
 }
| error khprefix {
  $$ = NULL;
  driver.syntax_error = true;
 }
| error dkhprefix {
  $$ = NULL;
  driver.syntax_error = true;
 }
| DK gepistemicprefix {
  // Distributed knowledge in a group
  if($2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($2->first);
    modal_formula *a = new modal_formula(ap);
    $$ = new modal_formula(42, a, $2->second);
  } else
    $$ = NULL;
 }
| error epistemicprefix { 
  $$ = NULL;
  driver.syntax_error = true;
 }
| error gepistemicprefix { 
  $$ = NULL;
  driver.syntax_error = true;
 }
| atlprefix X formula {
  // ATL X operator
  if($1!=NULL && $3!=NULL) {
    atomic_proposition *ap = new atomic_proposition($1);
    modal_formula *a = new modal_formula(ap);
    $$ = new modal_formula(45, a, $3);
  } else
    $$ = NULL;
 }
| atlprefix F formula {
  // ATL F operator
  if($1!=NULL && $3!=NULL) {
    atomic_proposition *ap = new atomic_proposition($1);
    modal_formula *a = new modal_formula(ap);
    $$ = new modal_formula(46, a, $3);
  } else
    $$ = NULL;
 }
| atlprefix G formula {
  // ATL G operator
  if($1!=NULL && $3!=NULL) {
    atomic_proposition *ap = new atomic_proposition($1);
    modal_formula *a = new modal_formula(ap);
    $$ = new modal_formula(47, a, $3);
  } else
    $$ = NULL;
 }
| atlprefix error formula {
  $$ = NULL;
  driver.syntax_error = true;
  USE($3);
 }
| atlprefix untilprefix {
  // ATL U operator
  if($1!=NULL && $2!=NULL) {
    atomic_proposition *ap = new atomic_proposition($1);
    modal_formula *a = new modal_formula(ap);
    $$ = new modal_formula(48, a, $2->first, $2->second);
  } else
    $$ = NULL;
 }
| ID {	
  if($1!=NULL) {
    atomic_proposition *ap = new atomic_proposition($1);
    $$ = new modal_formula(ap);
  } else
    $$ = NULL;
  }
| ID DOT GREENSTATES {
  if($1!=NULL) {
    string *newname = new string(*$1 + ".GreenStates");
    atomic_proposition *ap = new atomic_proposition(newname);
    $$ = new modal_formula(ap);
  } else
    $$ = NULL;
 }
| ID DOT REDSTATES {
  if($1!=NULL) {
    string *newname = new string(*$1 + ".RedStates");
    atomic_proposition *ap = new atomic_proposition(newname);
    $$ = new modal_formula(ap);
  } else
    $$ = NULL;
 }
| ID DOT error {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| ID error GREENSTATES {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| ID error REDSTATES {
  $$ = NULL;
  driver.syntax_error = true;
  USE($1);
 }
| error {
  $$ = NULL;
  driver.syntax_error = true;
  yyclearin;
  };

%%
void
yy::mcmas_parser::error (const yy::mcmas_parser::location_type& l,
                          const std::string& m)
{
  driver.error (l, m);
}
