package org.mcmas.ui.editors;


import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.text.rules.*;

/**
 * This scanner recognizes agents, evalualtion, initstates, groups, fariness and formulae.
 */
public class IsplAgentScanner extends RuleBasedPartitionScanner {

	public final static String ISPL_OBSVARS= "__ispl_obsvars"; 
	public final static String ISPL_VARS= "__ispl_vars";
	public final static String ISPL_REDSTATES= "__ispl_RedStates"; 
	public final static String ISPL_ACTIONS= "__ispl_actions";
	public final static String ISPL_PROTOCOL= "__ispl_protocol"; 
	public final static String ISPL_EVOLUTION= "__ispl_evolution";
	
	public final static String[] ISPL_PARTITION_TYPES= new String[] { ISPL_OBSVARS, ISPL_VARS, ISPL_REDSTATES,
		ISPL_ACTIONS, ISPL_PROTOCOL, ISPL_EVOLUTION };


	/**
	 * Creates the partitioner and sets up the appropriate rules.
	 */
	public IsplAgentScanner() {
		super();

		IToken IsplObsvars= new Token(ISPL_OBSVARS);
		IToken IsplVars= new Token(ISPL_VARS);
		IToken IsplRedstates= new Token(ISPL_REDSTATES);
		IToken IsplActions= new Token(ISPL_ACTIONS);
		IToken IsplProtocol= new Token(ISPL_PROTOCOL);
		IToken IsplEvolution= new Token(ISPL_EVOLUTION);

		List<IPredicateRule> rules= new ArrayList<IPredicateRule>();

		// Add rule for single line comments.
		rules.add(new EndOfLineRule("--", Token.UNDEFINED)); //$NON-NLS-1$


		// Add rules for multi-line obsvars, vars, actions, redstates, protocol and evoliution.
		rules.add(new MultiLineRule("Obsvars", "Obsvars", IsplObsvars, (char) 0, true)); 
		rules.add(new MultiLineRule("Vars", "Vars", IsplVars, (char) 0, true)); 
		rules.add(new MultiLineRule("RedStates", "RedStates", IsplRedstates, (char) 0, true)); 
		rules.add(new MultiLineRule("Actions", ";", IsplActions, (char) 0, true)); 
		rules.add(new MultiLineRule("Protocol", "Protocol", IsplProtocol, (char) 0, true)); 
		rules.add(new MultiLineRule("Evolution", "Evolution", IsplEvolution, (char) 0, true)); 

		IPredicateRule[] result= new IPredicateRule[rules.size()];
		rules.toArray(result);
		setPredicateRules(result);
	}
}
