package org.mcmas.ui.editors;


import java.util.ArrayList;
import java.util.Iterator;

import org.eclipse.core.resources.IMarker;
import org.eclipse.jface.text.*;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.IAnnotationHover;
import org.eclipse.jface.text.source.IAnnotationModel;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.projection.ProjectionAnnotation;
import org.eclipse.ui.texteditor.MarkerAnnotation;

/** 
 * The JavaAnnotationHover provides the hover support for java editors.
 */
 
public class IsplAnnotationHover implements IAnnotationHover {
	
	private MCMASEditor editor;

	public String getHoverInfo(ISourceViewer aViewer, int aLine) {
			String info = null;
			IMarker[] marker = getMarkerForLine(aViewer, aLine);
			if (marker != null) {
				for(int i=0; i<marker.length; i++) {
					String message = marker[i].getAttribute(IMarker.MESSAGE, (String)null);
					if (message != null && message.trim().length() > 0) {
						if(info == null)
							info = message.trim();
						else
							info += "\n" + message.trim();
					}
				}
			}
			return info;
		}
		
		protected IMarker[] getMarkerForLine(ISourceViewer aViewer, int aLine) {
			IAnnotationModel model = aViewer.getAnnotationModel();
			if (model != null) {
				ArrayList<IMarker> markerList = new ArrayList<IMarker>();
				Iterator e = model.getAnnotationIterator();
				while(e.hasNext()) {
					Object o = e.next();
					if (o instanceof MarkerAnnotation) {
						MarkerAnnotation a = (MarkerAnnotation)o;
						IMarker marker = a.getMarker();
						int line = marker.getAttribute(IMarker.LINE_NUMBER, 0);
						if(line == aLine+1) 
							markerList.add(marker);
					}	
				}
				if(markerList.size() > 0){
					IMarker[] markers = new IMarker[markerList.size()];
					markerList.toArray(markers);
					return markers;
				}
			}
			if (editor.annotationModel != null) {
				Iterator e = editor.annotationModel.getAnnotationIterator();
				while(e.hasNext()) {
					Object o = e.next();
					if (o instanceof ProjectionAnnotation) {
						ProjectionAnnotation a = (ProjectionAnnotation)o;
						Position p = editor.annotationModel.getPosition(a);
						IDocument document = editor.getDocumentProvider().getDocument(editor.getEditorInput());
						try {
							int startLine = document.getLineOfOffset(p.offset);
							if(startLine == aLine) {
								ArrayList<IMarker> markerlist = new ArrayList<IMarker>();
								if (model != null) {
									Iterator e1 = model.getAnnotationIterator();
									while(e1.hasNext()) {
										Object o1 = e1.next();
										if (o1 instanceof MarkerAnnotation) {
											MarkerAnnotation a1 = (MarkerAnnotation)o1;
											Position p1 = model.getPosition(a1);
											if(p1.offset>=p.offset && p1.offset + p1.length <= p.offset + p.length) {
												IMarker marker = a1.getMarker();
												markerlist.add(marker);
											}
										}	
									}
								}
								if(markerlist.size() > 0) {
									IMarker[] markers = new IMarker[markerlist.size()];
									markerlist.toArray(markers);
									return markers;
								}
							}
						} catch(BadLocationException x) {
							
						}
						//IMarker marker = a.getMarker();
						/*int line = marker.getAttribute(IMarker.LINE_NUMBER, 0);
						if(line == aLine+1)
							return marker;*/
					}	
				}
			}
			return null;
		}
		
		public void setMCMASEditor(MCMASEditor editor) {
			this.editor = editor;
		}

}
