/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.mcmas.ui.editors;


import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.text.TextAttribute;
import org.eclipse.jface.text.rules.*;

/**
 * A Java code scanner.
 */
public class IsplCodeScanner extends RuleBasedScanner {

	private static String[] fgKeywords= { "Action", "Actions", "Agent", "end", "Environment", "Evaluation", 
										  "Evolution", "Fairness", "Formulae", "Groups", "InitStates", "Obsvars", 
										  "Protocol", "RedStates", "Vars", "Other", "if", "GreenStates" };
	
	private static String[] fgOperators= { "and", "or", "AG", "EG", "AX", "EX", "X", "F", "G", "AF", "EF", "A", 
										   "E" , "U", "K", "GK", "GCK", "O", "KH", "DKH", "DK" };

	private static String[] fgTypes= { "boolean" }; 

	private static String[] fgConstants= { "false", "true" }; 

	/**
	 * Creates an Ispl code scanner with the given color provider.
	 * 
	 * @param provider the color provider
	 */
	public IsplCodeScanner(IsplColorProvider provider) {
		IToken keyword= new Token(new TextAttribute(provider.getColor(IsplColorProvider.KEYWORD)));
		IToken operator= new Token(new TextAttribute(provider.getColor(IsplColorProvider.OPERATOR)));
		IToken type= new Token(new TextAttribute(provider.getColor(IsplColorProvider.TYPE)));
		IToken comment= new Token(new TextAttribute(provider.getColor(IsplColorProvider.SINGLE_LINE_COMMENT)));
		IToken other= new Token(new TextAttribute(provider.getColor(IsplColorProvider.DEFAULT)));

		List<IRule> rules= new ArrayList<IRule>();

		// Add rule for single line comments.
		rules.add(new EndOfLineRule("--", comment)); //$NON-NLS-1$


		// Add generic whitespace rule.
		rules.add(new WhitespaceRule(new IsplWhitespaceDetector()));

		// Add word rule for keywords, types, and constants.
		WordRule wordRule= new WordRule(new IsplWordDetector(), other);
		for (int i= 0; i < fgKeywords.length; i++)
			wordRule.addWord(fgKeywords[i], keyword);
		for (int i= 0; i < fgTypes.length; i++)
			wordRule.addWord(fgTypes[i], type);
		for (int i= 0; i < fgConstants.length; i++)
			wordRule.addWord(fgConstants[i], type);
		for (int i= 0; i < fgOperators.length; i++)
			wordRule.addWord(fgOperators[i], operator);
		rules.add(wordRule);

		IRule[] result= new IRule[rules.size()];
		rules.toArray(result);
		setRules(result);
	}
}
