package org.mcmas.ui.editors;


import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.text.*;
import org.eclipse.jface.text.rules.FastPartitioner;
import org.eclipse.jface.text.rules.IToken;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.ui.texteditor.IDocumentProvider;
import org.eclipse.ui.texteditor.ITextEditor;
import org.eclipse.ui.views.contentoutline.ContentOutlinePage;
import org.mcmas.ui.syntax.Util;

/**
 * A content outline page which always represents the content of the
 * connected editor in 10 segments.
 */
public class IsplContentOutlinePage extends ContentOutlinePage {

	protected static IsplPartitionScanner scanner; 
	protected static IsplAgentScanner AgentScanner; 
	protected static IsplWordScanner WordScanner; 
	protected static IsplStateScanner StateScanner; 
	protected static IsplGroupScanner GroupScanner; 
	protected static IsplItemScanner ItemScanner; 
	protected static FindReplaceDocumentAdapter adapter;
	protected ISelection previous;
	protected boolean moveCursor = true;
	protected ContentProvider contentprovider;
	protected MCMASEditor editor;
	
	ISelectionChangedListener editorListener = new ISelectionChangedListener() {

		public void selectionChanged(SelectionChangedEvent event) {
			System.out.println("In editorListener.selectionChanged");
			ISelection s = event.getSelection();
			//if (s instanceof TextSelection) {
				if(previous==null || !previous.equals(s)) {
					getTreeViewer().setSelection(s);
					previous = s;
				}
			//}

		}
	};

	public ISelectionChangedListener getISelectionChangedListener() {
		return editorListener;
	}
	
	/**
	 * Divides the editor's document into ten segments and provides elements for
	 * them.
	 */
	protected class ContentProvider implements ITreeContentProvider {

		protected final static String SEGMENTS= "__ISPL_segments"; //$NON-NLS-1$
		protected IPositionUpdater fPositionUpdater= new DefaultPositionUpdater(SEGMENTS);
		protected List<Segment> fContent= new ArrayList<Segment>();

		private void getVariables(IDocument document, ITypedRegion field, Segment vars, ArrayList<Segment> states, int count) {
			try{
				IDocumentPartitioner partitioner2= new FastPartitioner(StateScanner, IsplStateScanner.ISPL_PARTITION_TYPES);
				partitioner2.connect(document);
				String str = document.get(field.getOffset()+count, field.getLength()-count*2);
				int firstcolumn = str.indexOf(':');
				if(firstcolumn<0)
					return;
				str = document.get(field.getOffset()+count+firstcolumn+1, 
						field.getLength()-count*2-firstcolumn-1);
				int firstcolumn1 = str.indexOf(':');
				if(firstcolumn1<0)
					return;
				
				ITypedRegion[] variables =  partitioner2.computePartitioning(field.getOffset()+count+firstcolumn1-firstcolumn-1, 
						field.getLength()-count*2-firstcolumn1-firstcolumn-1);
				//for(int k=0; k<variables.length; k++)
				//	str = document.get(variables[k].getOffset(), variables[k].getLength());
				for(int k=0; k<variables.length; k+=2) {
					if(k==0)
						WordScanner.setRange(document, field.getOffset()+count+firstcolumn+1, firstcolumn1);
					else
						WordScanner.setRange(document, variables[k-1].getOffset(), variables[k-1].getLength());
					IToken t;
					do {
						t = WordScanner.nextToken();
					} while (!(t.getData()!=null && t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD)==0) && !t.isEOF());
					if(WordScanner.getTokenLength()>0 && t.getData()!=null &&
							t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD)==0) {
						int offset = WordScanner.getTokenOffset();
						String varname = document.get(offset, WordScanner.getTokenLength());
						if(varname.compareTo("end")==0)
							break;
						String typedetail = document.get(variables[k].getOffset()+1, variables[k].getLength()-2);
						if(typedetail.indexOf("boolean")>=0)
							varname += " : boolean";
						else if(typedetail.indexOf("{")>=0 && typedetail.indexOf("}")>=0)
							varname += " : enumeration";
						else if(typedetail.indexOf("..")>=0)
							varname += " : bounded integer";
						//Position p2= new Position(WordScanner.getTokenOffset(), WordScanner.getTokenLength());
						Position p2= new Position(offset, variables[k].getOffset()+variables[k].getLength()-1-offset);
						document.addPosition(SEGMENTS, p2);
						Segment state = new Segment(MessageFormat.format(varname, 
								new Object[] { new Integer(variables[k].getOffset()) }), p2, vars, null);
						states.add(state);
					}
				}
			} catch (BadPositionCategoryException x) {
			} catch (BadLocationException x) {
			}

		}
		
		private void getGroups(IDocument document, ITypedRegion field, Segment vars, ArrayList<Segment> groups) {
			try{
				IDocumentPartitioner partitioner2= new FastPartitioner(GroupScanner, IsplGroupScanner.ISPL_PARTITION_TYPES);
				partitioner2.connect(document);
				String str = document.get(field.getOffset()+6, field.getLength()-12);
				int firstcolumn = str.indexOf('=');
				if(firstcolumn<0)
					return;
				
				ITypedRegion[] variables =  partitioner2.computePartitioning(field.getOffset()+6+firstcolumn, 
						field.getLength()-12-firstcolumn);
				str = document.get(field.getOffset()+6+firstcolumn-1, 
						field.getLength()-12-firstcolumn-1);
				//for(int k=0; k<variables.length; k++)
				//	str = document.get(variables[k].getOffset(), variables[k].getLength());
				for(int k=0; k<variables.length; k+=2) {
					if(k==0)
						WordScanner.setRange(document, field.getOffset()+6, firstcolumn);
					else
						WordScanner.setRange(document, variables[k-1].getOffset(), variables[k-1].getLength());
					IToken t;
					do {
						t = WordScanner.nextToken();
					} while (!(t.getData()!=null && t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD)==0) && !t.isEOF());
					if(WordScanner.getTokenLength()>0 && t.getData()!=null &&
							t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD)==0) {
						int offset = WordScanner.getTokenOffset();
						String varname = document.get(offset, WordScanner.getTokenLength());
						if(varname.compareTo("end")==0)
							break;
						Position p2= new Position(offset, variables[k].getOffset()+variables[k].getLength()-1-offset);
						document.addPosition(SEGMENTS, p2);
						Segment group = new Segment(MessageFormat.format(varname, 
								new Object[] { new Integer(variables[k].getOffset()) }), p2, vars, null);
						groups.add(group);
					}
				}
			} catch (BadPositionCategoryException x) {
			} catch (BadLocationException x) {
			}

		}
		
		private ITypedRegion getBigRegion(IRegion start, IRegion end, String type) {
			return new TypedRegion(start.getOffset(), end.getOffset()+end.getLength()-start.getOffset()+1, type);
		}
		
		private int getRegions(IDocument document, String strType, int start, ArrayList<IRegion> agents, 
				int Offset, int Length) {
			int len = Length + Offset - start;
			if(len>0) {
				/*IRegion r = adapter.find(start, str, true, true, true, false);
				while(r!=null && r.getOffset()<Offset+Length){
					agents.add(r);
					start = r.getOffset()+r.getLength();
					r = adapter.find(start, str, true, true, true, false);
				}*/
				WordScanner.setRange(document, start, len);
				IToken t = WordScanner.nextToken();
				while(!t.isEOF()) {
					if(t.getData()!=null && t.getData().toString().compareTo(strType)==0) {
						TypedRegion r = new TypedRegion(WordScanner.getTokenOffset(), WordScanner.getTokenLength(), strType);
						agents.add(r);
						start = WordScanner.getTokenOffset();
					}
					t = WordScanner.nextToken();
				}
			}
			return start;
		}
		private ITypedRegion[] partitionISPL(IDocument document, int Offset, int Length) {
			adapter = new FindReplaceDocumentAdapter(document);
			ArrayList<ITypedRegion> regions = new ArrayList<ITypedRegion>();
			int start = 0;
			ArrayList<IRegion> agents = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_AGENT, start, agents, Offset, Length);
			for(int i=0; i<agents.size(); i+=2)
				if(i+1<agents.size())
					regions.add(getBigRegion(agents.get(i), agents.get(i+1), IsplWordScanner.ISPL_AGENT));
			
			ArrayList<IRegion>  evaluation = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_EVALUATION, start,evaluation, Offset, Length);
			if(evaluation.size()>=2)
				regions.add(getBigRegion(evaluation.get(0), evaluation.get(evaluation.size()-1), 
						IsplWordScanner.ISPL_EVALUATION));
			
			ArrayList<IRegion>  initstates = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_INITSTATES, start, initstates, Offset, Length);
			if(initstates.size()>=2)
				regions.add(getBigRegion(initstates.get(0), initstates.get(initstates.size()-1), 
						IsplWordScanner.ISPL_INITSTATES));
			
			ArrayList<IRegion>  groups = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_GROUPS, start, groups, Offset, Length);
			if(groups.size()>=2)
				regions.add(getBigRegion(groups.get(0), groups.get(groups.size()-1), 
						IsplWordScanner.ISPL_GROUPS));

			ArrayList<IRegion>  fairness = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_FAIRNESS, start, fairness, Offset, Length);
			if(fairness.size()>=2)
				regions.add(getBigRegion(fairness.get(0), fairness.get(fairness.size()-1), 
						IsplWordScanner.ISPL_FAIRNESS));

			ArrayList<IRegion>  formulae = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_FORMULAE, start, formulae, Offset, Length);
			if(formulae.size()>=2)
				regions.add(getBigRegion(formulae.get(0), formulae.get(formulae.size()-1), 
						IsplWordScanner.ISPL_FORMULAE));

			ITypedRegion[] result = new ITypedRegion[regions.size()];
			regions.toArray(result);
			return result;
		}
	
		private ITypedRegion[] partitionAgent(IDocument document, int Offset, int Length) {
			adapter = new FindReplaceDocumentAdapter(document);
			ArrayList<ITypedRegion> regions = new ArrayList<ITypedRegion>();
			int start = Offset;
			ArrayList<IRegion> obsvars = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_OBSVARS, start, obsvars, Offset, Length);
			if(obsvars.size()>=2)
				regions.add(getBigRegion(obsvars.get(0), obsvars.get(obsvars.size()-1), 
						IsplWordScanner.ISPL_OBSVARS));
			
			ArrayList<IRegion>  vars = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_VARS, start,vars, Offset, Length);
			if(vars.size()>=2)
				regions.add(getBigRegion(vars.get(0), vars.get(vars.size()-1), 
						IsplWordScanner.ISPL_VARS));
			
			ArrayList<IRegion>  redstates = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_REDSTATES, start, redstates, Offset, Length);
			if(redstates.size()>=2)
				regions.add(getBigRegion(redstates.get(0), redstates.get(redstates.size()-1), 
						IsplWordScanner.ISPL_REDSTATES));
			
			ArrayList<IRegion>  actions = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_ACTIONS, start, actions, Offset, Length);
			if(actions.size()>=1) {
				IRegion r = actions.get(0);
				boolean flag = false;
				try{
					IRegion eq = adapter.find(r.getOffset(), "=", true, false, false, false);
					IRegion openBracket = adapter.find(eq.getOffset(), "{", true, false, false, false);
					IRegion closeBracket = adapter.find(openBracket.getOffset(), "}", true, false, false, false);
					IRegion semicolumn = adapter.find(eq.getOffset(), ";", true, false, false, false);
					boolean beq = eq != null && eq.getOffset()+eq.getLength() <= Offset+Length;
					boolean bopenBracket = openBracket != null && openBracket.getOffset()+openBracket.getLength() <= Offset+Length;
					boolean bcloseBracket = closeBracket != null && closeBracket.getOffset()+closeBracket.getLength() <= Offset+Length;
					boolean bsemicolumn = semicolumn != null && eq.getOffset()+semicolumn.getLength() <= Offset+Length;
					if((beq || bopenBracket) && (bcloseBracket || bsemicolumn)) {
						int start1 = beq ? eq.getOffset()+1 : openBracket.getOffset()+1;
						int length = bsemicolumn ? semicolumn.getOffset()-start1 : closeBracket.getOffset()-start1;
						WordScanner.setRange(document, start1, length);
						IToken t = WordScanner.nextToken();
						while (!t.isEOF()) {
							if(t.isOther() && t.getData() != null){
								if(t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD) != 0) {
									int endAction = Util.getOffsetBackward(document, WordScanner.getTokenOffset());
									TypedRegion end = new TypedRegion(endAction, 1, IsplWordScanner.ISPL_ACTIONS);
									regions.add(getBigRegion(r, end, IsplWordScanner.ISPL_ACTIONS));
									flag = true;
									break;
								}
							}
							t = WordScanner.nextToken();
						}
						if(!flag) {
							TypedRegion end = new TypedRegion(start1+length, 0, IsplWordScanner.ISPL_ACTIONS);
							regions.add(getBigRegion(r, end, IsplWordScanner.ISPL_ACTIONS));
							String s = document.get(regions.get(regions.size()-1).getOffset(), regions.get(regions.size()-1).getLength());
							flag = true;
						}						
					} 
				} catch(BadLocationException e) {
					
				}
				if(!flag)
					regions.add(getBigRegion(actions.get(0), actions.get(0), 
							IsplWordScanner.ISPL_ACTIONS));
			}

			ArrayList<IRegion>  protocol = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_PROTOCOL, start, protocol, Offset, Length);
			if(protocol.size()>=2)
				regions.add(getBigRegion(protocol.get(0), protocol.get(protocol.size()-1), 
						IsplWordScanner.ISPL_PROTOCOL));

			ArrayList<IRegion>  evolution = new ArrayList<IRegion>();
			start = getRegions(document, IsplWordScanner.ISPL_EVOLUTION, start, evolution, Offset, Length);
			if(evolution.size()>=2)
				regions.add(getBigRegion(evolution.get(0), evolution.get(evolution.size()-1), 
						IsplWordScanner.ISPL_EVOLUTION));

			ITypedRegion[] result = new ITypedRegion[regions.size()];
			regions.toArray(result);
			return result;
		}

		private void getPropositions(IDocument document, int Offset, int Length, 
				ArrayList<Segment> propositions, Segment vars) {
			adapter = new FindReplaceDocumentAdapter(document);
			ArrayList<TypedRegion> delimit = new ArrayList<TypedRegion>();
			try{
				/*IRegion r = adapter.find(start, "if", true, true, true, false);
				while(r!=null && r.getOffset()<Offset+Length){
					delimit.add(r);
					start = r.getOffset()+r.getLength();
					r = adapter.find(start, ";", true, false, false, false);
					if(r!=null && r.getOffset()<Offset+Length) {
						delimit.add(r);
						start = r.getOffset()+r.getLength();
						r = adapter.find(start, "if", true, true, true, false);
					}
				}*/
				WordScanner.setRange(document, Offset, Length);
				IToken t = WordScanner.nextToken();
				while(!t.isEOF()) {
					if(t.getData()!=null && t.getData().toString().compareTo(IsplWordScanner.ISPL_IF)==0) {
						TypedRegion r = new TypedRegion(WordScanner.getTokenOffset(), WordScanner.getTokenLength(), 
								IsplWordScanner.ISPL_IF);
						delimit.add(r);
					}
					t = WordScanner.nextToken();
				}

				for(int i=0; i<delimit.size(); i++) {
					TypedRegion r = delimit.get(i);
					if(i==0)
						WordScanner.setRange(document, Offset, r.getOffset()-Offset);
					else {
						IRegion r1 = adapter.find(delimit.get(i-1).getOffset()+2, ";", true, false, false, false);
						if(r1 == null)
							break;
						WordScanner.setRange(document, r1.getOffset()+1, 
								r.getOffset()-r1.getOffset()-1);
					}
						
					do {
						t = WordScanner.nextToken();
					} while (!(t.getData()!=null && t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD)==0) && !t.isEOF());
					if(WordScanner.getTokenLength()>0 && t.getData()!=null &&
							t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD)==0) {
						int offset = WordScanner.getTokenOffset();
						String name = document.get(offset, WordScanner.getTokenLength());
						IRegion r1 = adapter.find(delimit.get(i).getOffset()+2, ";", true, false, false, false);
						int length = r1 == null ? WordScanner.getTokenLength() : r1.getOffset() - offset;
						Position p2= new Position(offset, length);
						document.addPosition(SEGMENTS, p2);
						Segment proposition = new Segment(MessageFormat.format(name, 
								new Object[] { WordScanner.getTokenOffset() }), p2, vars, null);
						propositions.add(proposition);
					}
				}
			} catch (BadPositionCategoryException x) {
			} catch(BadLocationException x) {
			}
		}

		private void getItems(IDocument document, int Offset, int Length, 
				ArrayList<Segment> items, Segment vars, String prefix) {
			adapter = new FindReplaceDocumentAdapter(document);
			int start = Offset;
			ArrayList<IRegion> delimit = new ArrayList<IRegion>();
			try{
				IRegion r = adapter.find(start, ";", true, false, false, false);
				while(r!=null && r.getOffset()<Offset+Length){
					delimit.add(r);
					start = r.getOffset()+r.getLength();
					r = adapter.find(start, ";", true, false, false, false);
				}
				int count = 0;
				for(int i=0; i<delimit.size(); i++) {
					r = delimit.get(i);
					if(i==0)
						WordScanner.setRange(document, Offset, r.getOffset()-Offset);
					else
						WordScanner.setRange(document, delimit.get(i-1).getOffset()+1, 
								r.getOffset()-delimit.get(i-1).getOffset()-1);
					IToken t;
					do {
						t = WordScanner.nextToken();
					} while (t.isUndefined() || t.isWhitespace());
					if(!t.isEOF()) {
						int offset = WordScanner.getTokenOffset();
						int length = r.getOffset()-offset;
						Position p2= new Position(offset, length);
						document.addPosition(SEGMENTS, p2);
						Segment item = new Segment(MessageFormat.format(prefix + " " + count, 
								new Object[] { WordScanner.getTokenOffset() }), p2, vars, null);
						items.add(item);
						count ++;
					}
				}
			} catch (BadPositionCategoryException x) {
			} catch(BadLocationException x) {
			}
		}

		protected void parse(IDocument document) {
			//FindReplaceDocumentAdapter adapter = new FindReplaceDocumentAdapter(document);
			//IDocumentPartitioner partitioner= new FastPartitioner(scanner, IsplPartitionScanner.ISPL_PARTITION_TYPES);
			//partitioner.connect(document);
			//IDocumentPartitioner partitioner1= new FastPartitioner(AgentScanner, IsplAgentScanner.ISPL_PARTITION_TYPES);
			//partitioner1.connect(document);

			//ITypedRegion[] regions =  partitioner.computePartitioning(0, document.getLength());
			ITypedRegion[] regions = partitionISPL(document, 0, document.getLength());
			for(int i=0; i<regions.length; i++) {
				try{
					String type = regions[i].getType();
					if(type.compareTo(IsplWordScanner.ISPL_AGENT)==0) {
						//String sss = document.get(regions[i].getOffset(), regions[i].getLength());
						Position p= new Position(regions[i].getOffset(), regions[i].getLength());
						document.addPosition(SEGMENTS, p);
						Segment seg = new Segment(MessageFormat.format("Agent", 
								new Object[] { new Integer(regions[i].getOffset()) }), p, null, null);
						
						//ITypedRegion[] fields =  partitioner1.computePartitioning(regions[i].getOffset()+5, regions[i].getLength()-5);
						ITypedRegion[] fields = partitionAgent(document, regions[i].getOffset()+5, regions[i].getLength()-10); 
						if(fields.length>0 && fields[0].getOffset()-regions[i].getOffset()-5 > 0) {
							WordScanner.setRange(document, regions[i].getOffset()+5, 
									fields[0].getOffset()-regions[i].getOffset()-5);
							//sss = document.get(regions[i].getOffset()+5, fields[0].getOffset()-regions[i].getOffset()-5);
							IToken t;
							do {
								t = WordScanner.nextToken();
							} while (!(t.getData()!=null && t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD)==0) && !t.isEOF());
							if(WordScanner.getTokenLength()>0 && t.getData()!=null && t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD)==0)
								seg.setName("Agent " + document.get(WordScanner.getTokenOffset(), WordScanner.getTokenLength()));
						}
						ArrayList<Segment> children = new ArrayList<Segment>();
						for(int j=0; j<fields.length; j++) {
							String type1 = fields[j].getType();
							if(type1.compareTo(IsplWordScanner.ISPL_OBSVARS)==0) {
								Position p1= new Position(fields[j].getOffset(), fields[j].getLength());
								document.addPosition(SEGMENTS, p1);
								Segment vars = new Segment(MessageFormat.format("Obsvars", 
										new Object[] { new Integer(fields[j].getOffset()) }), p1, seg, null);
								children.add(vars);
								ArrayList<Segment> states = new ArrayList<Segment>();
								getVariables(document, fields[j], vars, states, 7);
								if(states.size()>0)
									vars.setChildren(states);
							} else if(type1.compareTo(IsplWordScanner.ISPL_VARS)==0) {
								Position p1= new Position(fields[j].getOffset(), fields[j].getLength());
								document.addPosition(SEGMENTS, p1);
								Segment vars = new Segment(MessageFormat.format("Vars", 
										new Object[] { new Integer(fields[j].getOffset()) }), p1, seg, null);
								children.add(vars);
								ArrayList<Segment> states = new ArrayList<Segment>();
								getVariables(document, fields[j], vars, states, 4);
								if(states.size()>0)
									vars.setChildren(states);
							} else if(type1.compareTo(IsplWordScanner.ISPL_REDSTATES)==0) {
								Position p1= new Position(fields[j].getOffset(), fields[j].getLength());
								document.addPosition(SEGMENTS, p1);
								children.add(new Segment(MessageFormat.format("RedStates", 
										new Object[] { new Integer(fields[j].getOffset()) }), p1, seg, null));
							} else if(type1.compareTo(IsplWordScanner.ISPL_ACTIONS)==0) {
								Position p1= new Position(fields[j].getOffset(), fields[j].getLength());
								document.addPosition(SEGMENTS, p1);
								children.add(new Segment(MessageFormat.format("Actions", 
										new Object[] { new Integer(fields[j].getOffset()) }), p1, seg, null));
							} else if(type1.compareTo(IsplWordScanner.ISPL_PROTOCOL)==0) {
								Position p1= new Position(fields[j].getOffset(), fields[j].getLength());
								document.addPosition(SEGMENTS, p1);
								Segment fs = new Segment(MessageFormat.format("Protocol", 
										new Object[] { new Integer(fields[j].getOffset()) }), p1, seg, null);
								children.add(fs);
								ArrayList<Segment> items = new ArrayList<Segment>();
								getItems(document, fields[j].getOffset()+9, fields[j].getLength()-17, 
										items, fs, "item");
								if(items.size()>0)
									fs.setChildren(items);
							} else if(type1.compareTo(IsplWordScanner.ISPL_EVOLUTION)==0) {
								Position p1= new Position(fields[j].getOffset(), fields[j].getLength());
								document.addPosition(SEGMENTS, p1);
								Segment fs = new Segment(MessageFormat.format("Evolution", 
										new Object[] { new Integer(fields[j].getOffset()) }), p1, seg, null);
								children.add(fs);
								ArrayList<Segment> items = new ArrayList<Segment>();
								getItems(document, fields[j].getOffset()+10, fields[j].getLength()-19, 
										items, fs, "item");
								if(items.size()>0)
									fs.setChildren(items);
							}
						}
						if(children.size()>0)
							seg.setChildren(children);
						fContent.add(seg);
					}
					else if(type.compareTo(IsplWordScanner.ISPL_EVALUATION)==0) {
						Position p= new Position(regions[i].getOffset(), regions[i].getLength());
						document.addPosition(SEGMENTS, p);
						Segment ps = new Segment(MessageFormat.format("Evaluation", 
								new Object[] { new Integer(regions[i].getOffset()) }), p, null, null);
						fContent.add(ps);
						ArrayList<Segment> propositions = new ArrayList<Segment>();
						getPropositions(document, regions[i].getOffset()+10, regions[i].getLength()-20, 
								propositions, ps);
						if(propositions.size()>0)
							ps.setChildren(propositions);
					}
					else if(type.compareTo(IsplWordScanner.ISPL_INITSTATES)==0) {
						Position p= new Position(regions[i].getOffset(), regions[i].getLength());
						document.addPosition(SEGMENTS, p);
						fContent.add(new Segment(MessageFormat.format("InitStates", 
								new Object[] { new Integer(regions[i].getOffset()) }), p, null, null));
					}
					else if(type.compareTo(IsplWordScanner.ISPL_GROUPS)==0) {
						Position p= new Position(regions[i].getOffset(), regions[i].getLength());
						document.addPosition(SEGMENTS, p);
						Segment gs = new Segment(MessageFormat.format("Groups", 
								new Object[] { new Integer(regions[i].getOffset()) }), p, null, null);
						fContent.add(gs);
						ArrayList<Segment> groups = new ArrayList<Segment>();
						getGroups(document, regions[i], gs, groups);
						if(groups.size()>0)
							gs.setChildren(groups);
						
					}
					else if(type.compareTo(IsplWordScanner.ISPL_FAIRNESS)==0) {
						Position p= new Position(regions[i].getOffset(), regions[i].getLength());
						document.addPosition(SEGMENTS, p);
						Segment fs = new Segment(MessageFormat.format("Fairness", 
								new Object[] { new Integer(regions[i].getOffset()) }), p, null, null);
						fContent.add(fs);
						ArrayList<Segment> formulae = new ArrayList<Segment>();
						getItems(document, regions[i].getOffset()+8, regions[i].getLength()-16, 
								formulae, fs, "fairness");
						if(formulae.size()>0)
							fs.setChildren(formulae);
					}
					else if(type.compareTo(IsplWordScanner.ISPL_FORMULAE)==0) {
						Position p= new Position(regions[i].getOffset(), regions[i].getLength());
						document.addPosition(SEGMENTS, p);
						Segment fs = new Segment(MessageFormat.format("Formulae", 
								new Object[] { new Integer(regions[i].getOffset()) }), p, null, null);
						fContent.add(fs);
						ArrayList<Segment> formulae = new ArrayList<Segment>();
						getItems(document, regions[i].getOffset()+8, regions[i].getLength()-16, 
								formulae, fs, "formula");
						if(formulae.size()>0)
							fs.setChildren(formulae);
					}
				} catch (BadPositionCategoryException x) {
				} catch (BadLocationException x) {
				}
			}
			/*int lines= document.getNumberOfLines();
			int increment= Math.max(Math.round(lines / 10), 10);

			for (int line= 0; line < lines; line += increment) {

				int length= increment;
				if (line + increment > lines)
					length= lines - line;

				try {

					int offset= document.getLineOffset(line);
					int end= document.getLineOffset(line + length);
					length= end - offset;
					Position p= new Position(offset, length);
					document.addPosition(SEGMENTS, p);
					fContent.add(new Segment(MessageFormat.format("OutlinePage.segment.title_pattern", 
							new Object[] { new Integer(offset) }), p)); 

				} catch (BadPositionCategoryException x) {
				} catch (BadLocationException x) {
				}
			}*/
		}
		
		public void searchSegment(int offset, ArrayList<Segment> segs, ArrayList<Segment> paths) {
			for(int i=0; i<segs.size(); i++) {
				Segment seg = segs.get(i);
				if(seg.position.offset<=offset && seg.position.length+seg.position.offset >= offset) {
					paths.add(seg);
					if(seg.children != null)
						searchSegment(offset, seg.children, paths);
				}
			}
		}

		public ArrayList<Position> getProjectionPositions() {
			ArrayList<Position> positions = new ArrayList<Position>();
			if(fContent != null) {
				for(int i=0; i<fContent.size(); i++) {
					Segment seg = fContent.get(i);
					positions.add(seg.position);
					if(seg.name.indexOf("Agent") == 0 && seg.children != null)
						for(int j=0; j<seg.children.size(); j++)
							positions.add(seg.children.get(j).position);
				}
			}
			return positions;
		}
		
		public Segment[] getSelectedSegment(int offset) {
			if(fContent != null) {
				ArrayList<Segment> paths = new ArrayList<Segment>();
				searchSegment(offset, (ArrayList<Segment>)fContent, paths);
				if(paths.size() > 0) {
					Segment[] segs = new Segment[paths.size()];
					paths.toArray(segs);
					return segs;
				} else 
					return null;
			}
			return null;
		}
		
		public Segment getAgentSegmentbyName(String name) {
			if(fContent != null) {
				for(int i=0; i<fContent.size(); i++) {
					Segment seg = fContent.get(i);
					int spaceoffset = seg.name.indexOf(' ');
					if(spaceoffset == -1)
						return null;
					if(seg.name.substring(spaceoffset+1).compareTo(name)==0)
						return seg;
				}
			}
			return null;
		}
		public Segment getSegmentbyName(String name) {
			if(fContent != null) {
				for(int i=0; i<fContent.size(); i++) {
					Segment seg = fContent.get(i);
					if(seg.name.compareTo(name)==0)
						return seg;
				}
			}
			return null;
		}
		/*
		 * @see IContentProvider#inputChanged(Viewer, Object, Object)
		 */
		public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
			if (oldInput != null) {
				IDocument document= fDocumentProvider.getDocument(oldInput);
				if (document != null) {
					try {
						document.removePositionCategory(SEGMENTS);
					} catch (BadPositionCategoryException x) {
					}
					document.removePositionUpdater(fPositionUpdater);
				}
			}

			fContent.clear();

			if (newInput != null) {
				IDocument document= fDocumentProvider.getDocument(newInput);
				if (document != null) {
					document.addPositionCategory(SEGMENTS);
					document.addPositionUpdater(fPositionUpdater);

					parse(document);
				}
			}
		}

		/*
		 * @see IContentProvider#dispose
		 */
		public void dispose() {
			if (fContent != null) {
				for(int i=0; i<fContent.size(); i++) {
					Segment seg = fContent.get(i);
					ArrayList<Segment> children = seg.getChildren();
					if(children!=null) {
						for(int j=0; j<children.size(); j++) {
							Segment child = children.get(j);
							if(child.getChildren() != null)
								child.getChildren().clear();
						}
						children.clear();
					}
				}
				fContent.clear();
				fContent= null;
			}
		}

		/*
		 * @see IContentProvider#isDeleted(Object)
		 */
		public boolean isDeleted(Object element) {
			return false;
		}

		/*
		 * @see IStructuredContentProvider#getElements(Object)
		 */
		public Object[] getElements(Object element) {
			/*ArrayList<Segment> elements = new ArrayList<Segment>();
			if (fContent != null) {
				for(int i=0; i<fContent.size(); i++) {
					Segment seg = fContent.get(i);
					elements.add(seg);
					ArrayList<Segment> children = seg.getChildren();
					if(children!=null) {
						for(int j=0; j<children.size(); j++) {
							Segment child = children.get(j);
							elements.add(child);
							if(child.getChildren() != null)
								elements.addAll(child.children);
						}
					}
				}
			}
			return elements.toArray();*/
			return fContent.toArray();
		}

		/*
		 * @see ITreeContentProvider#hasChildren(Object)
		 */
		public boolean hasChildren(Object element) {
			if(element == fInput)
				return true;
			else if((element instanceof Segment) &&((Segment)element).hasChildren())
				return true;
			return false;
		}

		/*
		 * @see ITreeContentProvider#getParent(Object)
		 */
		public Object getParent(Object element) {
			if (element instanceof Segment) {
				Segment parent = ((Segment)element).getParent();
				if(parent == null)
					return fInput;
				else
					return parent;
			}
			return null;
		}

		/*
		 * @see ITreeContentProvider#getChildren(Object)
		 */
		public Object[] getChildren(Object element) {
			if (element == fInput)
				return fContent.toArray();
			else if(element instanceof Segment) {
				ArrayList<Segment> children = ((Segment)element).getChildren();
				if(children != null && children.size()>0)
					return children.toArray();
			}
			return new Object[0];
		}
	}

	protected Object fInput;
	protected IDocumentProvider fDocumentProvider;
	protected ITextEditor fTextEditor;

	/**
	 * Creates a content outline page using the given provider and the given editor.
	 * 
	 * @param provider the document provider
	 * @param editor the editor
	 */
	public IsplContentOutlinePage(IDocumentProvider provider, ITextEditor editor) {
		super();
		fDocumentProvider= provider;
		fTextEditor= editor;
		if(scanner == null)
			scanner = new IsplPartitionScanner();
		if(AgentScanner == null)
			AgentScanner = new IsplAgentScanner();
		if(WordScanner == null)
			WordScanner = new IsplWordScanner();
		if(StateScanner == null)
			StateScanner = new IsplStateScanner();
		if(GroupScanner == null)
			GroupScanner = new IsplGroupScanner();
		if(ItemScanner == null)
			ItemScanner = new IsplItemScanner();
	}
	
	public IsplPartitionScanner getIsplPartitionScanner() {
		return scanner;
	}

	public IsplAgentScanner getIsplAgentScanner() {
		return AgentScanner;
	}

	public IsplWordScanner getIsplWordScanner() {
		return WordScanner;
	}
	
	public IsplStateScanner getIsplStateScanner() {
		return StateScanner;
	}

	/* (non-Ispldoc)
	 * Method declared on ContentOutlinePage
	 */
	public void createControl(Composite parent) {

		super.createControl(parent);

		TreeViewer viewer= getTreeViewer();
		contentprovider = new ContentProvider();
		viewer.setContentProvider(contentprovider);
		viewer.setLabelProvider(new LabelProvider());
		viewer.addSelectionChangedListener(this);

		if (fInput != null) {
			viewer.setInput(fInput);
			viewer.expandAll();
			if(fTextEditor instanceof MCMASEditor) {
				((MCMASEditor)fTextEditor).updateFoldingStructure();
			}
		}
	}
	
	private boolean compareTwoSelections(ISelection s1, ISelection s2) {
		if (s1 instanceof TreeSelection && s2 instanceof TreeSelection) {
			TreePath[] p1 = ((TreeSelection) s1).getPaths();
			TreePath[] p2 = ((TreeSelection) s2).getPaths();
			if (p1 != null && p2 != null && p1.length == p2.length) {
				for (int i = 0; i < p1.length; i++) {
					if (p1[i].getSegmentCount() != p2[i].getSegmentCount())
						return false;
					Segment seg1 = (Segment) p1[i].getLastSegment();
					Segment seg2 = (Segment) p2[i].getLastSegment();
					if (seg1.position.offset != seg1.position.offset
							|| seg1.position.length != seg2.position.length)
						return false;
				}
				return true;
			}
		}
		return false;
	}
	
	/* (non-Ispldoc)
	 * Method declared on ContentOutlinePage
	 */
	public void selectionChanged(SelectionChangedEvent event) {
		//System.out.println("in selectionChanged");

		super.selectionChanged(event);

		ISelection selection= event.getSelection();
		/*if(compareTwoSelections(previous, selection))
			return;
		previous = selection;*/
		if(!moveCursor) {
			moveCursor = true;
			return;
		}
		if (selection.isEmpty())
			fTextEditor.resetHighlightRange();
		else {
			Segment segment= (Segment) ((IStructuredSelection) selection).getFirstElement();
			int start= segment.position.getOffset();
			int length= segment.position.getLength();
			try {
				fTextEditor.setHighlightRange(start, length, true);
			} catch (IllegalArgumentException x) {
				fTextEditor.resetHighlightRange();
			}
		}
	}
	
	/**
	 * Sets the input of the outline page
	 * 
	 * @param input the input of this outline page
	 */
	public void setInput(Object input) {
		fInput= input;
		update();
	}
	
	/**
	 * Updates the outline page.
	 */
	public void update() {
		TreeViewer viewer= getTreeViewer();

		if (viewer != null) {
			Control control= viewer.getControl();
			if (control != null && !control.isDisposed()) {
				control.setRedraw(false);
				viewer.setInput(fInput);
				viewer.expandAll();
				control.setRedraw(true);
			}
		}
	}
	
	public void setSelectedItem(int offset) {
		Segment[] paths= contentprovider.getSelectedSegment(offset);
		if(paths != null) {
			moveCursor = false;
			setSelection(new TreeSelection(new TreePath(paths)));
		}
	}
	
	public Segment[] getTreePath(int offset) {
		return contentprovider.getSelectedSegment(offset);
	}
	
	public ArrayList<Position> getProjectionPositions() {
		return contentprovider.getProjectionPositions();
	}
	public Segment getSegmentbyName(String name) {
		return contentprovider.getSegmentbyName(name);
	}
	public Segment getAgentSegmentbyName(String name) {
		return contentprovider.getAgentSegmentbyName(name);
	}
}
