package org.mcmas.ui.editors;

import java.util.LinkedList;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.DefaultIndentLineAutoEditStrategy;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.TextUtilities;
import org.eclipse.jface.text.formatter.ContextBasedFormattingStrategy;
import org.eclipse.jface.text.formatter.FormattingContextProperties;
import org.eclipse.jface.text.formatter.IFormattingContext;
import org.eclipse.jface.text.rules.IToken;
import org.eclipse.jface.text.rules.Token;
import org.mcmas.ui.syntax.Util;

public class IsplFormattingStrategy extends ContextBasedFormattingStrategy {
	/** Documents to be formatted by this strategy */
	private final LinkedList fDocuments = new LinkedList();

	/** access to the preferences store * */
	private int indent = 0;

	public IsplFormattingStrategy() {
	}

	public IsplFormattingStrategy(int indent) {
		this.indent = indent;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.jface.text.formatter.IFormattingStrategyExtension#format()
	 */
	public void format() {
		super.format();
		final IDocument document = (IDocument) fDocuments.removeFirst();
		if (document != null) {
			// TODO allow formatting of regions, not just the entire document; bug 75611
			//String documentText = document.get();
			/*XmlDocumentFormatter formatter = new XmlDocumentFormatter();
			if (indent != -1) {
			    formatter.setInitialIndent(indent);
			}
			formatter.setDefaultLineDelimiter(TextUtilities.getDefaultLineDelimiter(document));
			String  formattedText = formatter.format(documentText, this.prefs);
			if (formattedText != null && !formattedText.equals(documentText)) {
			    document.set(formattedText);
			}*/
			//String delimiter = TextUtilities.getDefaultLineDelimiter(document);
			String newcontent = format(document);
			document.set(newcontent);
		}
	}

	/*
	 * @see org.eclipse.jface.text.formatter.ContextBasedFormattingStrategy#formatterStarts(org.eclipse.jface.text.formatter.IFormattingContext)
	 */
	public void formatterStarts(final IFormattingContext context) {
		super.formatterStarts(context);

		fDocuments.addLast(context.getProperty(FormattingContextProperties.CONTEXT_MEDIUM));
	}

	/*
	 * @see org.eclipse.jface.text.formatter.ContextBasedFormattingStrategy#formatterStops()
	 */
	public void formatterStops() {
		super.formatterStops();

		fDocuments.clear();
	}
	
	private String format(IDocument document) {
		StringBuffer newcontent= new StringBuffer("");
		String delimiter = TextUtilities.getDefaultLineDelimiter(document);
		IsplWordScanner WordScanner = new IsplWordScanner();
		
		try{
			int lines = document.getNumberOfLines();
			boolean pendingend = false;
			boolean expectColumn = false;
			boolean expectID = false;
			boolean linebreak = true;
			for(int i=0; i<lines; i++) {
				int offset = document.getLineOffset(i);
				int length = document.getLineLength(i);
				String strLine = document.get(offset, length);
				strLine = strLine.trim();
				if(Util.isEmpty(strLine))
					continue;
				else if(strLine.startsWith("--")) {
					newcontent.append(indentstr() + strLine + delimiter);
				} else {
					int whiteend= Util.findEndOfWhiteSpace(document, offset, offset+length);
					WordScanner.setRange(document, whiteend, strLine.length());
					boolean isFirstToken = true;
					IToken t = WordScanner.nextToken();
					while (!t.isEOF()) {
						if(t.isUndefined()) {
							String strToken = document.get(WordScanner.getTokenOffset(), WordScanner.getTokenLength());
							if(isFirstToken) {
								newcontent.append(indentstr());
								isFirstToken = false;
							}
							newcontent.append(strToken);
						} else if(t.isWhitespace()) {
							if(isFirstToken) {
								newcontent.append(indentstr());
								isFirstToken = false;
							}
							else
								newcontent.append(' ');
						} else if(t.isOther()) {
							if(t.getData()!=null) {
								String strType = t.getData().toString(); 
								if(pendingend) {
									pendingend = false;
									boolean addspace = Character.isWhitespace(newcontent.charAt(newcontent.length()-1)) ? false : true;
									if(strType.compareTo(IsplWordScanner.ISPL_AGENT) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("Agent\n\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_OBSVARS) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("Obsvars\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_VARS) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("Vars\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_REDSTATES) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("RedStates\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_PROTOCOL) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("Protocol\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_EVOLUTION) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("Evolution\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_EVALUATION) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("Evaluation\n\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_INITSTATES) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("InitStates\n\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_GROUPS) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("Groups\n\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_FAIRNESS) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("Fairness\n\n");
										isFirstToken = true;
									} else if(strType.compareTo(IsplWordScanner.ISPL_FORMULAE) == 0) {
										if(addspace)
											newcontent.append(' ');
										newcontent.append("Formulae\n");
										isFirstToken = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_END) == 0) {
										decreaseIndent();
										newcontent.append("\n" + indentstr() + "end");
										pendingend = true;
										expectColumn = false;
										expectID = false;
										isFirstToken = false;
										linebreak = false;
									} else {
										String strToken = document.get(WordScanner.getTokenOffset(), 
																	   WordScanner.getTokenLength());
										newcontent.append("\n" + indentstr() + strToken);
										pendingend = true;
										expectColumn = false;
										expectID = false;
										isFirstToken = false;
										linebreak = true;
									}
								} else { //                                                    pendingend == false
									if(strType.compareTo(IsplWordScanner.ISPL_AGENT) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append("Agent");
										setIndent(1);
										isFirstToken = false;
										pendingend = false;
										expectColumn = false;
										expectID = true;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_OBSVARS) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append(indentstr(1));
										newcontent.append("Obsvars");
										isFirstToken = false;
										setIndent(2);
										pendingend = false;
										expectColumn = true;
										expectID = false;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_VARS) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append(indentstr(1));
										newcontent.append("Vars");
										isFirstToken = false;
										setIndent(2);
										pendingend = false;
										expectColumn = true;
										expectID = false;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_REDSTATES) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append(indentstr(1));
										newcontent.append("RedStates");
										isFirstToken = false;
										setIndent(2);
										pendingend = false;
										expectColumn = true;
										expectID = false;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_PROTOCOL) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append(indentstr(1));
										newcontent.append("Protocol");
										isFirstToken = false;
										setIndent(2);
										pendingend = false;
										expectColumn = true;
										expectID = false;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_EVOLUTION) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append(indentstr(1));
										newcontent.append("Evolution");
										isFirstToken = false;
										setIndent(2);
										pendingend = false;
										expectColumn = true;
										expectID = false;
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_EVALUATION) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append("Evaluation\n");
										isFirstToken = true;
										setIndent(1);
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_INITSTATES) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append("InitStates\n");
										isFirstToken = true;
										setIndent(1);
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_GROUPS) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append("Groups\n");
										isFirstToken = true;
										setIndent(1);
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_FAIRNESS) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append("Fairness\n");
										isFirstToken = true;
										setIndent(1);
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_FORMULAE) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										newcontent.append("Formulae\n");
										isFirstToken = true;
										setIndent(1);
										linebreak = false;
									} else if(strType.compareTo(IsplWordScanner.ISPL_END) == 0) {
										if(newcontent.length() > 0 && newcontent.charAt(newcontent.length()-1) != '\n')
											newcontent.append('\n');
										decreaseIndent();
										newcontent.append(indentstr() + "end");
										isFirstToken = false;
										pendingend = true;
										expectColumn = false;
										expectID = false;
										linebreak = false;
									} else { //                                 ID or other keyword
										String strToken = document.get(WordScanner.getTokenOffset(), WordScanner.getTokenLength());
										if(expectID) {
											if(!Character.isWhitespace(newcontent.charAt(newcontent.length()-1)))
												newcontent.append(' ');
											newcontent.append(strToken + "\n");													
											linebreak = false;
										} else {
											if(isFirstToken)
												newcontent.append(indentstr());
											newcontent.append(strToken);
											linebreak = true;
											increaseTwoIndent();
										}
										pendingend = false;
										expectColumn = false;
										expectID = false;
										isFirstToken = false;
									}
								}
							} else {
								String strToken = document.get(WordScanner.getTokenOffset(), WordScanner.getTokenLength());
								if(expectColumn) {
									if(strToken.compareTo(":") == 0) {
										while(newcontent.length() > 0) {
											if(Character.isWhitespace(newcontent.length()-1))
												newcontent.deleteCharAt(newcontent.length()-1);
											else
												break;
										}
										newcontent.append(":\n");
										isFirstToken = true;
										linebreak = false;
									} else {
										newcontent.append("\n" + indentstr() + strToken);
										isFirstToken = false;
										if(strToken.compareTo(";") == 0) {
											decreaseTwoIndent();
											newcontent.append('\n');
											linebreak = false;
										}
									}
								} else {
									if(isFirstToken) {
										newcontent.append(indentstr());
										isFirstToken = false;
									}
									newcontent.append(strToken);
									if(strToken.compareTo(";") == 0) {
										decreaseTwoIndent();
										newcontent.append('\n');
										linebreak = false;
									}
								}
								
								pendingend = false;
								expectColumn = false;
								expectID = false;									
							}
						}
						t = WordScanner.nextToken();	
					}
					if(linebreak) {
						newcontent.append('\n');
					}

				}
			}
		} catch(BadLocationException e) {
			
		}
		
		return newcontent.toString();
	}

	private String indentstr() {
		String str = "";
		for(int i=0; i<indent; i++)
			str += "\t";
		return str;
	}

	private String indentstr(int count) {
		String str = "";
		for(int i=0; i<count; i++)
			str += "\t";
		return str;
	}
	
	private void decreaseIndent() {
		indent = indent > 0 ? indent - 1 : 0; 
	}

	private void increaseTwoIndent() {
		indent = indent < 3 ? indent + 2 : indent; 
	}

	private void decreaseTwoIndent() {
		indent = indent > 2 ? indent - 2 : indent; 
	}

	private void setIndent(int i) {
		indent = i; 
	}
}
