package org.mcmas.ui.editors;


import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.text.rules.*;

/**
 * This scanner recognizes agents, evalualtion, initstates, groups, fariness and formulae.
 */
public class IsplPartitionScanner extends RuleBasedPartitionScanner {

	public final static String ISPL_AGENT= "__ispl_agent"; 
	public final static String ISPL_EVALUATION= "__ispl_evaluation";
	public final static String ISPL_INITSTATES= "__ispl_initstates"; 
	public final static String ISPL_GROUPS= "__ispl_groups";
	public final static String ISPL_FAIRNESS= "__ispl_fairness"; 
	public final static String ISPL_FORMULAE= "__ispl_formulae";
	
	public final static String[] ISPL_PARTITION_TYPES= new String[] { ISPL_AGENT, ISPL_EVALUATION, ISPL_INITSTATES,
		ISPL_GROUPS, ISPL_FAIRNESS, ISPL_FORMULAE };


	/**
	 * Creates the partitioner and sets up the appropriate rules.
	 */
	public IsplPartitionScanner() {
		super();

		IToken IsplAgent= new Token(ISPL_AGENT);
		IToken IsplEvaluation= new Token(ISPL_EVALUATION);
		IToken IsplInitstates= new Token(ISPL_INITSTATES);
		IToken IsplGroups= new Token(ISPL_GROUPS);
		IToken IsplFariness= new Token(ISPL_FAIRNESS);
		IToken IsplFormulae= new Token(ISPL_FORMULAE);

		List<IPredicateRule> rules= new ArrayList<IPredicateRule>();

		// Add rule for single line comments.
		rules.add(new EndOfLineRule("--", Token.UNDEFINED)); //$NON-NLS-1$


		// Add rules for multi-line agents, evalualtion, initstates, groups, fairness and formulae.
		rules.add(new MultiLineRule("Agent", "Agent", IsplAgent, (char) 0, true)); 
		rules.add(new MultiLineRule("Evaluation", "Evaluation", IsplEvaluation, (char) 0, true)); 
		rules.add(new MultiLineRule("InitStates", "InitStates", IsplInitstates, (char) 0, true)); 
		rules.add(new MultiLineRule("Groups", "Groups", IsplGroups, (char) 0, true)); 
		rules.add(new MultiLineRule("Fairness", "Fairness", IsplFariness, (char) 0, true)); 
		rules.add(new MultiLineRule("Formulae", "Formulae", IsplFormulae, (char) 0, true)); 

		IPredicateRule[] result= new IPredicateRule[rules.size()];
		rules.toArray(result);
		setPredicateRules(result);
	}
}
