package org.mcmas.ui.editors;

import org.eclipse.jface.text.DefaultIndentLineAutoEditStrategy;
import org.eclipse.jface.text.IAutoEditStrategy;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ITextHover;
//import org.eclipse.jface.text.TextAttribute;
import org.eclipse.jface.text.contentassist.ContentAssistant;
import org.eclipse.jface.text.contentassist.IContentAssistant;
import org.eclipse.jface.text.formatter.IContentFormatter;
import org.eclipse.jface.text.formatter.MultiPassContentFormatter;
import org.eclipse.jface.text.presentation.IPresentationReconciler;
import org.eclipse.jface.text.presentation.PresentationReconciler;
import org.eclipse.jface.text.reconciler.IReconciler;
import org.eclipse.jface.text.reconciler.MonoReconciler;
import org.eclipse.jface.text.rules.DefaultDamagerRepairer;
import org.eclipse.jface.text.source.IAnnotationHover;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.SourceViewerConfiguration;
//import org.mcmas.ui.editors.IsplSourceViewerConfiguration.SingleTokenScanner;
import org.eclipse.swt.graphics.RGB;

public class IsplSourceViewerConfig extends SourceViewerConfiguration {
	private static IsplCodeScanner scanner;
	private MCMASEditor fEditor;
	
	public IsplSourceViewerConfig(MCMASEditor anEditor) {
		this.fEditor = anEditor;
		if(scanner==null)
			scanner = new IsplCodeScanner(new IsplColorProvider());
	}
	
	public IPresentationReconciler getPresentationReconciler(ISourceViewer sourceViewer) {
		
		PresentationReconciler reconciler= new PresentationReconciler();
		reconciler.setDocumentPartitioning(getConfiguredDocumentPartitioning(sourceViewer));
		
		DefaultDamagerRepairer dr= new DefaultDamagerRepairer(scanner);
		reconciler.setDamager(dr, IDocument.DEFAULT_CONTENT_TYPE);
		reconciler.setRepairer(dr, IDocument.DEFAULT_CONTENT_TYPE);
		
		return reconciler;
	}

    public IAnnotationHover getAnnotationHover(ISourceViewer aSourceViewer) {
    	IsplAnnotationHover h = new IsplAnnotationHover();
    	h.setMCMASEditor(fEditor);
       return h;
    }
    
    public ITextHover getTextHover(ISourceViewer sourceViewer, String contentType) {
    	IsplTextHover h = new IsplTextHover();
    	h.setMCMASEditor(fEditor);
		return h;
	}
    
    public IReconciler getReconciler(ISourceViewer aSourceViewer) {
        return new MonoReconciler(fEditor.getReconcilingStrategy(), false);
    }
    
	public int getTabWidth(ISourceViewer sourceViewer) {
		return 4;
	}

	public String[] getIndentPrefixes(ISourceViewer sourceViewer, String contentType) {
		return new String[] { "\t", "    " }; //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	public IAutoEditStrategy[] getAutoEditStrategies(ISourceViewer sourceViewer, String contentType) {
		IAutoEditStrategy strategy= (IDocument.DEFAULT_CONTENT_TYPE.equals(contentType) ? new IsplAutoIndentStrategy() : new DefaultIndentLineAutoEditStrategy());
		return new IAutoEditStrategy[] { strategy };
	}
	
	public IContentFormatter getContentFormatter(ISourceViewer sourceViewer) {
		MultiPassContentFormatter formatter= 
		      new MultiPassContentFormatter(
		      getConfiguredDocumentPartitioning(sourceViewer), 
		      IDocument.DEFAULT_CONTENT_TYPE);
		   formatter.setMasterStrategy(
		      new IsplFormattingStrategy());
		   
		   return formatter;
	}

	public IContentAssistant getContentAssistant(ISourceViewer sourceViewer) {

		ContentAssistant assistant= new ContentAssistant();
		//assistant.setDocumentPartitioning(getConfiguredDocumentPartitioning(sourceViewer));
		assistant.setContentAssistProcessor(new IsplCompletionProcessor(fEditor), IDocument.DEFAULT_CONTENT_TYPE);

		assistant.enableAutoActivation(true);
		assistant.setAutoActivationDelay(100);
		assistant.setProposalPopupOrientation(IContentAssistant.PROPOSAL_OVERLAY);
		//assistant.setContextInformationPopupOrientation(IContentAssistant.CONTEXT_INFO_ABOVE);
		//assistant.setContextInformationPopupBackground(IsplEditorPlugin.getDefault().getIsplColorProvider().getColor(new RGB(150, 150, 0)));

		return assistant;
	}

}
