package org.mcmas.ui.editors;


import java.util.ArrayList;

import org.eclipse.jface.text.*;
import org.eclipse.swt.graphics.Point;
import org.mcmas.ui.syntax.Util;

/**
 * Example implementation for an <code>ITextHover</code> which hovers over ISPL code.
 */
public class IsplTextHover implements ITextHover {

	private MCMASEditor editor;
	/* (non-Ispldoc)
	 * Method declared on ITextHover
	 */
	public String getHoverInfo(ITextViewer textViewer, IRegion hoverRegion) {
		if (hoverRegion != null) {
			try {
				IDocument document = textViewer.getDocument();
				if (hoverRegion.getLength() > -1) {
					int offset = hoverRegion.getOffset();
					int length = hoverRegion.getLength();
					if(length == 0) {
						Position p = Util.getCurrentWord(document, offset);
						if(p == null)
							return "";
						else { 
							offset = p.offset;
							length = p.length;
						}
					}
					String str = document.get(offset, length);
					if(Util.isEmpty(str) || Util.isKeyword(str))
						return str;
					Segment[] path = editor.getTreePath(offset);
					if(path != null && path.length >= 2) {
						if(path[0].name.startsWith("Agent")) {
							if(path.length==3 || (path.length==2 && path[1].name.startsWith("RedStates"))) {
								char c1 = offset==0 ? ' ' : document.getChar(offset-1);
								char c2 = offset+length>=document.getLength() ? ' ' : document.getChar(offset+length);
								if(!Character.isLetterOrDigit(c1) && !Character.isLetterOrDigit(c2)) {
									if(path[1].name.startsWith("RedStates") ||
											path[1].name.startsWith("Protocol") ||
											path[1].name.startsWith("Evolution")) {
										int dotoffset = Util.getDotOffset(document, offset);
										if(dotoffset == -1) { // local variable
											for(int i=0; i<path[0].children.size(); i++) {
												Segment seg = path[0].children.get(i);
												if(seg.name.startsWith("Obsvars") || seg.name.startsWith("Vars")) {
													Segment varseg = Util.searchVariable(seg.children, str);
													if(varseg != null) {
														return document.get(varseg.position.offset, varseg.position.length);
													} 
												}
											}
										} else {
											Position pos = Util.getWordPositionBackward(document, dotoffset);
											if(pos.length == 0)
												return str;
											String prefix = document.get(pos.offset, pos.length);
											if(prefix.compareTo("Environment")!=0)
												return str;
											Segment envseg = editor.getAgentSegmentbyName("Environment");
											if(envseg == null || envseg.children == null)
												return str;
											for(int i=0; i<envseg.children.size(); i++) {
												Segment seg = envseg.children.get(i);
												if(seg.name.startsWith("Obsvars")) {
													Segment varseg = Util.searchVariable(seg.children, str);
													if(varseg != null) {
														return document.get(varseg.position.offset, varseg.position.length);
													} else
														return str;
												}
											}
										}
									} 
								}
							}
						} else if(path[0].name.startsWith("Evaluation")) {
							char c1 = offset==0 ? ' ' : document.getChar(offset-1);
							char c2 = offset+length>=document.getLength() ? ' ' : document.getChar(offset+length);
							if(!Character.isLetterOrDigit(c1) && !Character.isLetterOrDigit(c2)) {
								int dotoffset = Util.getDotOffset(document, offset);
								if(dotoffset >= 0) {
									Position pos = Util.getWordPositionBackward(document, dotoffset);
									if(pos.length == 0)
										return str;
									String prefix = document.get(pos.offset, pos.length);
									Segment envseg = editor.getAgentSegmentbyName(prefix);
									if(envseg == null || envseg.children == null)
										return str;
									for(int i=0; i<envseg.children.size(); i++) {
										Segment seg = envseg.children.get(i);
										if(seg.name.startsWith("Obsvars") || seg.name.startsWith("Vars")) {
											Segment varseg = Util.searchVariable(seg.children, str);
											if(varseg != null) {
												return document.get(varseg.position.offset, varseg.position.length);
											} else
												return str;
										}
									}
								}
							}
						} else if(path[0].name.startsWith("Fairness") || path[0].name.startsWith("Formulae")) {
							char c1 = offset==0 ? ' ' : document.getChar(offset-1);
							char c2 = offset+length>=document.getLength() ? ' ' : document.getChar(offset+length);
							if(!Character.isLetterOrDigit(c1) && !Character.isLetterOrDigit(c2)) {
								Segment envseg = editor.getSegmentbyName("Evaluation");
								if(envseg == null || envseg.children == null)
									return str;
								for(int i=0; i<envseg.children.size(); i++) {
									Segment seg = envseg.children.get(i);
									if(seg.name.compareTo(str) == 0)
										return document.get(seg.position.offset, seg.position.length);
								}
							}
						} 
					} else if(path != null && path.length == 1 && path[0].name.startsWith("InitStates")) {
						char c1 = offset==0 ? ' ' : document.getChar(offset-1);
						char c2 = offset+length>=document.getLength() ? ' ' : document.getChar(offset+length);
						if(!Character.isLetterOrDigit(c1) && !Character.isLetterOrDigit(c2)) {
							int dotoffset = Util.getDotOffset(document, offset);
							if(dotoffset >= 0) {
								Position pos = Util.getWordPositionBackward(document, dotoffset);
								if(pos.length == 0)
									return str;
								String prefix = document.get(pos.offset, pos.length);
								Segment envseg = editor.getAgentSegmentbyName(prefix);
								if(envseg == null || envseg.children == null)
									return str;
								for(int i=0; i<envseg.children.size(); i++) {
									Segment seg = envseg.children.get(i);
									if(seg.name.startsWith("Obsvars") || seg.name.startsWith("Vars")) {
										Segment varseg = Util.searchVariable(seg.children, str);
										if(varseg != null) {
											return document.get(varseg.position.offset, varseg.position.length);
										} else
											return str;
									}
								}
							}
						}

					}
					return str;
				}
			} catch (BadLocationException x) {
			}
		}
		return /*IsplEditorMessages.getString*/("IsplTextHover.emptySelection"); //$NON-NLS-1$
	}
	
	
	/* (non-Ispldoc)
	 * Method declared on ITextHover
	 */
	public IRegion getHoverRegion(ITextViewer textViewer, int offset) {
		Point selection= textViewer.getSelectedRange();
		if (selection.x <= offset && offset < selection.x + selection.y)
			return new Region(selection.x, selection.y);
		return new Region(offset, 0);
	}
	
	public void setMCMASEditor(MCMASEditor editor) {
		this.editor = editor;
	}
}
