package org.mcmas.ui.editors;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.URI;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.TreeSet;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.TypedRegion;
import org.eclipse.jface.text.rules.IToken;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.IVerticalRuler;
import org.eclipse.jface.text.source.SourceViewer;
import org.eclipse.jface.text.source.projection.ProjectionAnnotation;
import org.eclipse.jface.text.source.projection.ProjectionAnnotationModel;
import org.eclipse.jface.text.source.projection.ProjectionSupport;
import org.eclipse.jface.text.source.projection.ProjectionViewer;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.editors.text.TextEditor;
import org.eclipse.ui.ide.FileStoreEditorInput;
import org.eclipse.ui.texteditor.ITextEditorActionDefinitionIds;
import org.eclipse.ui.texteditor.MarkerAnnotation;
import org.eclipse.ui.texteditor.TextOperationAction;
import org.eclipse.ui.views.contentoutline.IContentOutlinePage;
//import org.antlr.runtime.*;
import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.*;
import org.mcmas.ui.Activator;
import org.mcmas.ui.preferences.PreferenceConstants;
import org.mcmas.ui.syntax.*;

public class MCMASEditor extends TextEditor {
	private IsplContentOutlinePage fOutlinePage;
	
	private static IsplWordScanner WordScanner; 
	
	public String errorMessage = "";
	public InterpretedSystem is;
	
	private ISPLLexer lexer;
	private CommonTokenStream tokens;
	private ISPLParser parser;
	public ArrayList<RawSegment> ispltree;
	private IsplReconcilingStrategy fReconcilingStrategy;
    private ProjectionSupport projectionSupport;
	private Annotation[] oldAnnotations;
	public ProjectionAnnotationModel annotationModel;
	private boolean isInteractiveMode = false;
	private boolean isInteractiveMode1 = false;

	private Process proc;
	private BufferedReader br;
	private BufferedWriter bw;
	private MultiPageEditor parentEditor;
	private CounterExample r;
	private EnabledTransitions et;
	private int step;
	private ArrayList<Position> symbolicStack;
	private int symbolicOffset;
	//private ArrayList<FlowChart> flowchartWindow;
	
	private IsplColorProvider colorProvider;
	
	public MCMASEditor() {
		super();
		fReconcilingStrategy = new IsplReconcilingStrategy(this);
		setSourceViewerConfiguration(new IsplSourceViewerConfig(this));
		//setDocumentProvider(new XMLDocumentProvider());
		if(WordScanner == null)
			WordScanner = new IsplWordScanner();
		ispltree = new ArrayList<RawSegment>();
		colorProvider = new IsplColorProvider();
		//flowchartWindow = new ArrayList<FlowChart>();
	}
	
	public void setMultiPageEditor(MultiPageEditor parentEditor) {
		this.parentEditor = parentEditor;
	}
	
	public MultiPageEditor getMultiPageEditor() {
		return parentEditor;
	}
	
	public Object getAdapter(Class required) {
		if (IContentOutlinePage.class.equals(required)) {
			if (fOutlinePage == null) {
				fOutlinePage= new IsplContentOutlinePage(getDocumentProvider(), this);
				if (getEditorInput() != null) {
					fOutlinePage.setInput(getEditorInput());
					//updateFoldingStructure();
				}
				//getSelectionProvider().addSelectionChangedListener(fOutlinePage.getISelectionChangedListener());
			}
			return fOutlinePage;
		}
		
		/*if (projectionSupport != null) {
			Object adapter= projectionSupport.getAdapter(getSourceViewer(), required);
			if (adapter != null)
				return adapter;
		}*/
		
		return super.getAdapter(required);
	}

	public void dispose() {
		if(fOutlinePage!=null)
			getSelectionProvider().removeSelectionChangedListener(fOutlinePage.getISelectionChangedListener());
		disposeIsplTree();
		IEditorInput editorInput = getEditorInput();
		try{
			if (editorInput instanceof IFileEditorInput) { 
				IFile file = ((IFileEditorInput)editorInput).getFile(); 
				file.deleteMarkers(IMarker.PROBLEM, true, IResource.DEPTH_INFINITE);
			}
			else if(editorInput instanceof FileStoreEditorInput) {
				URI uri = ((FileStoreEditorInput)editorInput).getURI();
				IFile[] files = ResourcesPlugin.getWorkspace().getRoot().findFilesForLocationURI(uri);
				if(files != null && files.length>0)
					for(int j=0; j<files.length; j++) {
						files[j].deleteMarkers(IMarker.PROBLEM, true, IResource.DEPTH_INFINITE);
					}
			}
		} catch(CoreException e) {
			
		}
	}
	/** The <code>JavaEditor</code> implementation of this 
	 * <code>AbstractTextEditor</code> method performs any extra 
	 * revert behavior required by the java editor.
	 */
	public void doRevertToSaved() {
		super.doRevertToSaved();
		checkSyntax();
		if (fOutlinePage != null)
			fOutlinePage.update();
	}
	
	/** The <code>JavaEditor</code> implementation of this 
	 * <code>AbstractTextEditor</code> method performs any extra 
	 * save behavior required by the java editor.
	 * 
	 * @param monitor the progress monitor
	 */
	public void doSave(IProgressMonitor monitor) {
		super.doSave(monitor);
		checkSyntax();
		if (fOutlinePage != null)
			fOutlinePage.update();
	}
	
	/** The <code>JavaEditor</code> implementation of this 
	 * <code>AbstractTextEditor</code> method performs any extra 
	 * save as behavior required by the java editor.
	 */
	public void doSaveAs() {
		super.doSaveAs();
		checkSyntax();
		if (fOutlinePage != null)
			fOutlinePage.update();
	}
	

	public void updateOutlinePage() {
		if (fOutlinePage != null) {
			fOutlinePage.update();
		}
	}

	public IsplReconcilingStrategy getReconcilingStrategy() {
		return fReconcilingStrategy;	
	}

	protected void handleCursorPositionChanged() {
		super.handleCursorPositionChanged();
		ISelection selection= getEditorSite().getSelectionProvider().getSelection();
		if (selection instanceof ITextSelection) {
		    int start= ((ITextSelection)selection).getOffset();
			IDocument document = getDocumentProvider().getDocument(getEditorInput());
		    //System.out.print("Offset = " + start + ", Length = " + length + " : ");
		    try {
			    char c = document.getChar(start);
			    while(Character.isWhitespace(c) && c != '\n') {
			    	start ++;
			    	c = document.getChar(start);
			    }
			    if(fOutlinePage != null)
			    	fOutlinePage.setSelectedItem(start);
		    } catch(BadLocationException x) {
				
			}
		}
		/*String pos = getCursorPosition();
		System.out.print(pos);
		int delimit = pos.indexOf(' ');
		int line = Integer.parseInt(pos.substring(0, delimit)) - 1;
		delimit = pos.lastIndexOf(' ');
		int column = Integer.parseInt(pos.substring(delimit+1, pos.length())) - 1;
		IDocument document = getDocumentProvider().getDocument(getEditorInput());
		try {
			int offset = document.getLineOffset(line) + column;
			System.out.println(" -- line = "+ line + ", column = " + column + ", offset = " + offset);
			fOutlinePage.setSelectedItem(offset);
		} catch(BadLocationException x) {
			
		}*/
	}
	
	public void createPartControl(Composite parent)	{
	    super.createPartControl(parent);
	    ProjectionViewer viewer =(ProjectionViewer)getSourceViewer();

	    projectionSupport = new ProjectionSupport(viewer,getAnnotationAccess(),getSharedColors());
	    projectionSupport.addSummarizableAnnotationType("org.eclipse.ui.workbench.texteditor.error");
	    projectionSupport.install();

	    //turn projection mode on
	    viewer.doOperation(ProjectionViewer.TOGGLE);

	    annotationModel = viewer.getProjectionAnnotationModel();
	    /*projectionSupport.setHoverControlCreator(new IInformationControlCreator() {
	    	   public IInformationControl createInformationControl(Shell shell) {
	    	     return new CustomSourceInformationControl(shell, IDocument.DEFAULT_CONTENT_TYPE);
	    	   }
	    	});*/
	}
	
    protected ISourceViewer createSourceViewer(Composite parent, IVerticalRuler ruler, int styles) {
        ISourceViewer viewer = new ProjectionViewer(parent, ruler, getOverviewRuler(), isOverviewRulerVisible(), styles);

    	// ensure decoration support has been created and configured.
    	getSourceViewerDecorationSupport(viewer);
    	
    	return viewer;
    }
    
    public void updateFoldingStructure() {
    	if (fOutlinePage != null) {
	    	ArrayList<Position> positions = fOutlinePage.getProjectionPositions();
			Annotation[] annotations = new Annotation[positions.size()];

			// this will hold the new annotations along
			// with their corresponding positions
			HashMap<ProjectionAnnotation, Position> newAnnotations = new HashMap<ProjectionAnnotation, Position>();
			for (int i = 0; i < positions.size(); i++) {
				ProjectionAnnotation annotation = new ProjectionAnnotation();
				newAnnotations.put(annotation, positions.get(i));
				annotations[i] = annotation;
			}

			annotationModel.modifyAnnotations(oldAnnotations, newAnnotations, null);
			oldAnnotations = annotations;
    	}
    }


	public void addAgent() {
		IDocument document = getDocumentProvider().getDocument(getEditorInput());
		WordScanner.setRange(document, 0, document.getLength());
		IToken t = WordScanner.nextToken();
		TypedRegion r = null;
		while(!t.isEOF()) {
			if(t.getData()!=null && t.getData().toString().compareTo(IsplWordScanner.ISPL_AGENT)==0)
				r = new TypedRegion(WordScanner.getTokenOffset(), WordScanner.getTokenLength(), 
						IsplWordScanner.ISPL_AGENT);
			t = WordScanner.nextToken();
		}
		String basicFrame = "Agent\n";
		basicFrame += "    Vars:\n";
		basicFrame += "    end Vars\n";
		basicFrame += "    RedStates:\n";
		basicFrame += "    end RedStates\n";
		basicFrame += "    Actions = {};\n";
		basicFrame += "    Protocol:\n";
		basicFrame += "    end Protocol\n";
		basicFrame += "    Evolution:\n";
		basicFrame += "    end Evolution\n";
		basicFrame += "end Agent";
		try {
			if(r != null) 
				document.replace(r.getOffset()+r.getLength(), 0, "\n\n" + basicFrame);
			else
				document.replace(0, 0, basicFrame+"\n\n");
		} catch(BadLocationException x) {
			x.printStackTrace();
		}
	}

	public void disposeIsplTree() {
		if (ispltree != null) {
			for(int i=0; i<ispltree.size(); i++) {
				RawSegment seg = ispltree.get(i);
				if(seg==null)
					continue;
				ArrayList<RawSegment> children = seg.getChildren();
				if(children!=null) {
					for(int j=0; j<children.size(); j++) {
						RawSegment child = children.get(j);
						if(child.getChildren() != null)
							child.getChildren().clear();
					}
					children.clear();
				}
			}
			ispltree.clear();
			ispltree= null;
		}
	}
	
	public synchronized boolean checkSyntax() {
		disposeIsplTree();
		ispltree = new ArrayList<RawSegment>();
		errorMessage = "";
		IEditorInput editorInput = getEditorInput();
		IDocument document = getDocumentProvider().getDocument(editorInput);
		try {
			ANTLRStringStream cs = new ANTLRStringStream(document.get());
			lexer = new ISPLLexer(cs);
			lexer.setMCMASEditor(this);
			tokens = new CommonTokenStream();
			tokens.setTokenSource(lexer);
			parser = new ISPLParser(tokens);
			parser.setMCMASEditor(this);
			ISPLParser.is_return r = parser.is();
			
			if(Util.isEmpty(errorMessage)) {
				CommonTree t = (CommonTree)r.getTree();
				CommonTreeNodeStream nodes = new CommonTreeNodeStream(t);
				ISPLTREE walker = new ISPLTREE(nodes);
				walker.setMCMASEditor(this);
				walker.is();
			}
		} catch(RecognitionException e) {
			//e.printStackTrace(System.out);
		} catch(Exception e) {
			//e.printStackTrace();
			//System.out.println(errorMessage);
        }
		
		if (Util.isEmpty(errorMessage))
			errorMessage = Util.global_consistency_check(is.is_agents, is.agents, is.is_evaluation, 
														 is.is_istates, is.is_groups, is.is_formulae, 
														 is.is_fairness, document);
		
		try {
			if (!Util.isEmpty(errorMessage)) {
				ArrayList<ErrorInfo> info = getErrorInfo(document);
				if (info != null) {
					if (editorInput instanceof IFileEditorInput) {
						IFile file = ((IFileEditorInput) editorInput).getFile();
						file.deleteMarkers(IMarker.PROBLEM, true,
								IResource.DEPTH_INFINITE);
						for (int i = 0; i < info.size(); i++) {
							ErrorInfo m = info.get(i);
							/*System.out.println("Line = " + m.line
									+ ", Column = " + m.pos + ", Offset = "
									+ m.offset + ": " + m.message);*/
							addProblemMarker(file, m.message, m.line + 1,
									m.length, m.offset);
						}
					} else if (editorInput instanceof FileStoreEditorInput) {
						URI uri = ((FileStoreEditorInput) editorInput).getURI();
						IFile[] files = ResourcesPlugin.getWorkspace()
								.getRoot().findFilesForLocationURI(uri);
						if (files != null && files.length > 0)
							// IFile file =
							// workspace.getRoot().getFilesForLocation(path);
							for (int j = 0; j < files.length; j++) {
								files[j].deleteMarkers(IMarker.PROBLEM, true,
										IResource.DEPTH_INFINITE);
								for (int i = 0; i < info.size(); i++) {
									ErrorInfo m = info.get(i);
									System.out.println("Line = " + m.line
											+ ", Column = " + m.pos
											+ ", Offset = " + m.offset
											+ ", Length = " + m.length + ": "
											+ m.message);
									addProblemMarker(files[j], m.message,
											m.line + 1, m.length, m.offset);
								}
							}
					}
				}
			} else {
				if (editorInput instanceof IFileEditorInput) {
					IFile file = ((IFileEditorInput) editorInput).getFile();
					file.deleteMarkers(IMarker.PROBLEM, true,
							IResource.DEPTH_INFINITE);
				} else if (editorInput instanceof FileStoreEditorInput) {
					URI uri = ((FileStoreEditorInput) editorInput).getURI();
					IFile[] files = ResourcesPlugin.getWorkspace().getRoot()
							.findFilesForLocationURI(uri);
					if (files != null && files.length > 0)
						for (int j = 0; j < files.length; j++) {
							files[j].deleteMarkers(IMarker.PROBLEM, true,
									IResource.DEPTH_INFINITE);
						}
				}
			}
		} catch (CoreException e) {

		}
		
		/*ArrayList<Segment> fContent = new ArrayList<Segment>();
		for(int i=0; i<ispltree.size(); i++) {
			RawSegment rawseg = ispltree.get(i);
			if(rawseg!=null) {
				Segment seg = rawseg.getSegment(document, null);
				fContent.add(seg);
			}
		}*/
		if(!Util.isEmpty(errorMessage))
			return false;
		else
			return true;
	}
	
	public ArrayList<ErrorInfo> getErrorInfo(IDocument document) {
		ArrayList<ErrorInfo> info = null;
		if(errorMessage != null && errorMessage.compareTo("") != 0) {
			String[] errors = errorMessage.split("\n");
			info = new ArrayList<ErrorInfo>();
			for(int i=0; i<errors.length; i++) {
				try{
					int start = errors[i].indexOf(' ');
					int column1 = errors[i].indexOf(':', start+1);
					int column2 = errors[i].indexOf(':', column1+1);
					int end = errors[i].indexOf(' ', column2+1);
					int line = Integer.parseInt(errors[i].substring(start+1, column1))-1;
					int pos = Integer.parseInt(errors[i].substring(column1+1, column2));
					int length = Integer.parseInt(errors[i].substring(column2+1, end));
					int offset = document.getLineOffset(line)+pos;
					String message = errors[i].substring(end+1);
					ErrorInfo m = new ErrorInfo(line, pos, offset, length, message);
					/*if(i==0)
						info.add(m);
					else
						for(int j=0; j<info.size(); j++) 
							if(offset<info.get(j).offset) {
								info.add(j, m);
								break;
							} else if(j==info.size()-1) {
								info.add(m);
								break;
							}*/
					info.add(m);
				} catch(BadLocationException e) {
					
				}
			}
		}
		return info;
	}
	
	public void addProblemMarker(IFile file, String message, int line, int length, int offset) {
		try {
			IMarker marker = file.createMarker(IMarker.PROBLEM);
			marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_ERROR);
			marker.setAttribute(IMarker.LINE_NUMBER, line);
			marker.setAttribute(IMarker.MESSAGE, message);
			Position pos = new Position(offset);
			marker.setAttribute(IMarker.CHAR_START, offset);
			marker.setAttribute(IMarker.CHAR_END, offset+length);
			marker.setAttribute(IMarker.PRIORITY, IMarker.PRIORITY_HIGH);
			MarkerAnnotation ma = new MarkerAnnotation(marker);
			getSourceViewer().getAnnotationModel().addAnnotation(ma, pos);
		} catch (Exception e) {
			
		}
	}
	
	public Segment[] getTreePath(int offset) {
		return fOutlinePage.getTreePath(offset);
	}
	
	public Segment getSegmentbyName(String name) {
		return fOutlinePage.getSegmentbyName(name);
	}
	public Segment getAgentSegmentbyName(String name) {
		return fOutlinePage.getAgentSegmentbyName(name);
	}
	
	public void formatText() {
		SourceViewer viewer =(SourceViewer)getSourceViewer();
		viewer.doOperation(ISourceViewer.FORMAT);
	}
	
	protected void createActions() {
		super.createActions();
		
		IAction a= new TextOperationAction(MCMASEditorMessages.getResourceBundle(), "ContentAssistProposal.", this, ISourceViewer.CONTENTASSIST_PROPOSALS); //$NON-NLS-1$
		a.setActionDefinitionId(ITextEditorActionDefinitionIds.CONTENT_ASSIST_PROPOSALS);
		setAction("ContentAssistProposal", a); //$NON-NLS-1$
	}
	
	public void interactiveExecution() {
		if(isInteractiveMode) {
			boolean confirmed = MessageDialog.openQuestion(null, "Warning", "Explicit interactive execution is running! Restart it?");
			if(!confirmed)
				return;
		} else if(isInteractiveMode1) {
			boolean confirmed = MessageDialog.openQuestion(null, "Warning", "Symbolic interactive execution is running! Restart it?");
			if(!confirmed)
				return;
			else {
				isInteractiveMode1 = false;
				if(proc != null)
					proc.destroy();
			}
		}
		getSourceViewer().setEditable(false);
		isInteractiveMode = true;
		parentEditor.setText("");
		ArrayList<GlobalState> gstates = is.generateInitialGlobalStates();
		if(gstates != null && gstates.size() > 0) {
			String strStates = is.displayCandidateStates();
			parentEditor.setText2(strStates);
			parentEditor.setNumberText("");
			parentEditor.setChosenFlag(true);
			parentEditor.enableButtons(true);
			parentEditor.setPromptText("Choose an initial state");
			if(gstates.size() == 1) 
				chooseState(1);
			parentEditor.enableBacktrackButtons(false);				
			/*Hashtable<String, TreeSet<String>> enabledActions = new Hashtable<String, TreeSet<String>>();
			is.getEnabledActions(initstate, enabledActions);
			ArrayList<Hashtable<String, String>> trans = is.getPossibleTransitions(enabledActions);
			if(trans != null && trans.size() > 0) {
				Hashtable<String, String> actions = trans.get(0);
				ArrayList<GlobalState> successors = is.getSuccessors(initstate, actions);
				if(successors != null && successors.size() > 0) {
					GlobalState secondstate = successors.get(0);
				}
			}
			System.out.println("Possible transitions are generated successfully!");*/
		} else {
			MessageDialog.openInformation(null, "Warning", "There is no initial state in the model!");
		}
	}
	
	public void interactiveExecution1() {
		if(isInteractiveMode1) {
			boolean confirmed = MessageDialog.openQuestion(null, "Warning", "Symbolic nteractive execution is running! Restart it?");
			if(!confirmed)
				return;
			if(proc != null)
				proc.destroy();
		} else if(isInteractiveMode) {
			boolean confirmed = MessageDialog.openQuestion(null, "Warning", "Explicit interactive execution is running! Restart it?");
			if(!confirmed)
				return;
			else
				isInteractiveMode = false;
		}
		getSourceViewer().setEditable(false);
		isInteractiveMode1 = true;
		parentEditor.setText("");
		symbolicStack = new ArrayList<Position>();
		symbolicOffset = 0;
		
		IPreferenceStore store = Activator.getDefault().getPreferenceStore();
		String MCMASPath = store.getString(PreferenceConstants.MCMAS_PATH);
		String cygwinPath = store.getString(PreferenceConstants.CYGWIN_PATH);
		String path = parentEditor.getFilePath();
		String title = parentEditor.getFileTitle();
		int osType = parentEditor.getOsType();
		//String mcmasPath = osType == 0 ? MCMASPath + "\\mcmas -s -q " + path + title :
		//	MCMASPath + "/mcmas -s -q " + path + title;
		String slash = osType == 0 ? "\\" : "/";
		String[] mcmasCommand = {MCMASPath + slash + "mcmas", "-s", "-q", path + title};
		Runtime rt = Runtime.getRuntime();
		try{
			if(osType > 0)
		    	proc = rt.exec(mcmasCommand);
		    else {
		    	proc = rt.exec(mcmasCommand, new String[]{"Path="+cygwinPath});
		    }
			InputStreamReader isr = new InputStreamReader(proc.getInputStream());
			InputStreamReader isr1 = new InputStreamReader(proc.getErrorStream());
			br = new BufferedReader(isr);
			OutputStreamWriter isw = new OutputStreamWriter(proc.getOutputStream());
			bw = new BufferedWriter(isw);
			BufferedReader br1 = new BufferedReader(isr1);
			String line = null;
			String output = "";
			while(!br.ready()) {
				if(br1.ready())
					while(br1.ready()) 
						line = br1.readLine();
			}
			while((line = br.readLine()) != null) {
				if(line.startsWith("Done."))
					break;
				output += line + "\n";
			}

				
			if(!output.startsWith("No initial state. Check your model!")) {
				r = getGlobalStates(output);
				String strStates = r.toString1();
				parentEditor.setText2(strStates);
				parentEditor.setNumberText("");
				parentEditor.setChosenFlag(true);
				parentEditor.enableButtons(true);
				parentEditor.setPromptText("Choose an initial state");
				step = 0;
				if(r.numberOfStates() == 1) 
					chooseState1(1);
				parentEditor.enableBacktrackButtons(false);				
				/*Hashtable<String, TreeSet<String>> enabledActions = new Hashtable<String, TreeSet<String>>();
				is.getEnabledActions(initstate, enabledActions);
				ArrayList<Hashtable<String, String>> trans = is.getPossibleTransitions(enabledActions);
				if(trans != null && trans.size() > 0) {
					Hashtable<String, String> actions = trans.get(0);
					ArrayList<GlobalState> successors = is.getSuccessors(initstate, actions);
					if(successors != null && successors.size() > 0) {
						GlobalState secondstate = successors.get(0);
					}
				}
				System.out.println("Possible transitions are generated successfully!");*/
			} else {
				MessageDialog.openInformation(null, "Warning", "There is no initial state in the model!");
			}
		} catch(IOException e) {
			e.printStackTrace();
		} 
	}
	
	public void quitInteractiveMode() {
		if(isInteractiveMode)
			is.quitInteractiveMode();
		else if(isInteractiveMode1 && proc != null) {
			try{
				bw.write("-1\n");
				bw.flush();
			} catch(IOException e) {
				
			}
			proc.destroy();
		}
			
		isInteractiveMode = false;
		isInteractiveMode1 = false;
		
		getSourceViewer().setEditable(true);
	}
	
	public void chooseState(int number) {
		boolean result = is.chooseState(number);
		if(!result) {
			MessageDialog.openError(null, "Error", "State No is invalid!");
		} else {
			String strState = is.displayCurrentState();
			parentEditor.appendText(strState);
			ArrayList<Hashtable<String, String>> trans = is.generateEnabledActions();
			if(trans != null && trans.size() > 0) {
				String strActions = is.displayCurrentActions();
				parentEditor.setText2(strActions);
				parentEditor.setChosenFlag(false);
				parentEditor.setNumberText("");
				parentEditor.enableBacktrackButtons(true);
				parentEditor.enableChooseButtons(true);
				parentEditor.setPromptText("Choose a transition");
			} else {
				parentEditor.enableChooseButtons(false);
				MessageDialog.openInformation(null, "Warning", "There is no enabled transition in the current state!");
			}
		}
	}

	public void chooseState1(int number) {
		if(number < 1 || number > r.numberOfStates()) {
			MessageDialog.openError(null, "Error", "State No is invalid!");
		} else {
			String strState = r.getState(number, step++);
			parentEditor.appendText(strState);
			Position p = new Position(symbolicOffset, strState.length());
			symbolicStack.add(p);
			symbolicOffset += strState.length();
			try{
				bw.write(Integer.toString(number)+"\n");
				bw.flush();
				String line = null;
				String output = "";
				while((line = br.readLine()) != null) {
					if(line.startsWith("Done."))
						break;
					output += line + "\n";
				}
 
				if(output.startsWith("There is no enabled transition. Please backtrack or quit.")) {
					parentEditor.enableChooseButtons(false);
					MessageDialog.openInformation(null, "Warning", output);
				} else if(output.startsWith("There is no enabled transition in the initial state.")) {
					parentEditor.enableButtons(false);
					MessageDialog.openInformation(null, "Warning", output+" Interactive execution stops.");
					proc.destroy();
					quitInteractiveMode();
				} else {
					et = getEnabledTransitions(output);
					String strActions = et.toString();
					parentEditor.setText2(strActions);
					parentEditor.setChosenFlag(false);
					parentEditor.setNumberText("");
					parentEditor.enableBacktrackButtons(true);
					parentEditor.enableChooseButtons(true);
					parentEditor.setPromptText("Choose a transition");
				} 
			} catch(IOException e) {
				e.printStackTrace();
			} 
		}
	}

	public void chooseAction(int number) {
		boolean result = is.chooseAction(number);
		if(!result) {
			MessageDialog.openError(null, "Error", "Action No is invalid!");
		} else {
			String strAction = is.displayChosenAction();
			parentEditor.appendText(strAction);
			ArrayList<GlobalState> successors = is.generateSuccessorStates();
			if(successors != null && successors.size() > 0) {
				String strStates = is.displayCandidateStates();
				parentEditor.setText2(strStates);
				parentEditor.setChosenFlag(true);
				parentEditor.setNumberText("");
				parentEditor.enableBacktrackButtons(true);
				parentEditor.enableChooseButtons(true);
				parentEditor.setPromptText("Choose a successor state");
				if(successors.size() == 1)
					chooseState(1);
			} else {
				MessageDialog.openInformation(null, "Warning", "The current state does not have any successor states!");
			}
		}
	}

	public void chooseAction1(int number) {
		if(number < 1 || number > et.transitions.length) {
			MessageDialog.openError(null, "Error", "Action No is invalid!");
		} else {
			String strAction = et.getTransition(number);
			parentEditor.appendText(strAction);
			symbolicOffset += strAction.length();
			try{
				bw.write(Integer.toString(number)+"\n");
				bw.flush();
				String line = null;
				String output = "";
				while((line = br.readLine()) != null) {
					if(line.startsWith("Done."))
						break;
					output += line + "\n";
				}

	
				if(!output.startsWith("Maximum stack depth is reached. Please backtrack or quit.")) {
					r = getGlobalStates(output);
					String strStates = r.toString1();
					parentEditor.setText2(strStates);
					parentEditor.setChosenFlag(true);
					parentEditor.setNumberText("");
					parentEditor.enableBacktrackButtons(true);
					parentEditor.enableChooseButtons(true);
					parentEditor.setPromptText("Choose a successor state");
					//step++;
					if(r.numberOfStates() == 1)
						chooseState1(1);
				} else {
					MessageDialog.openInformation(null, "Warning", output);
				}
			} catch(IOException e) {
				e.printStackTrace();
			} 

		}
	}

	public void backtrack() {
		if(isInteractiveMode) {
			is.backtrack();
			Position p = is.getTopTextState();
			int start = p == null ? 0 : p.offset + p.length;
			parentEditor.deleteText(start);
			if(is.isStackEmpty()) {
				ArrayList<GlobalState> gstates = is.generateInitialGlobalStates();
				String strStates = is.displayCandidateStates();
				parentEditor.setText2(strStates);
				parentEditor.setNumberText("");
				parentEditor.setChosenFlag(true);
				parentEditor.enableButtons(true);
				parentEditor.setPromptText("Choose an initial state");
				if(gstates.size() == 1) 
					chooseState(1);
				parentEditor.enableBacktrackButtons(false);
			} else {
				//String strState = is.displayCurrentState();
				//parentEditor.appendText(strState);
				ArrayList<Hashtable<String, String>> trans = is.generateEnabledActions();
				String strActions = is.displayCurrentActions();
				parentEditor.setText2(strActions);
				parentEditor.setChosenFlag(false);
				parentEditor.setNumberText("");
				parentEditor.enableBacktrackButtons(true);
				parentEditor.setPromptText("Choose a transition");
			}
		} else {
			Position p = symbolicStack.get(symbolicStack.size()-1);
			if(p.offset+p.length == symbolicOffset) {
				if(step > 0)
					step--;
				symbolicStack.remove(symbolicStack.size()-1);
				if(!symbolicStack.isEmpty()) {
					p = symbolicStack.get(symbolicStack.size()-1);
					symbolicOffset = p.length+p.offset;
				} else
					symbolicOffset = 0;
			} else
				symbolicOffset = p.length+p.offset;
			parentEditor.deleteText(symbolicOffset);
			
			try{
				bw.write("0\n");
				bw.flush();
				String line = null;
				String output = "";
				while((line = br.readLine()) != null) {
					if(line.startsWith("Done."))
						break;
					output += line + "\n";
				}
				if(symbolicStack.isEmpty()) {
					r = getGlobalStates(output);
					String strStates = r.toString1();
					parentEditor.setText2(strStates);
					parentEditor.setNumberText("");
					parentEditor.setChosenFlag(true);
					parentEditor.enableButtons(true);
					parentEditor.setPromptText("Choose an initial state");
					if(r.numberOfStates() == 1) 
						chooseState(1);
					parentEditor.enableBacktrackButtons(false);
				} else {
					et = getEnabledTransitions(output);
					String strActions = et.toString();
					parentEditor.setText2(strActions);
					parentEditor.setChosenFlag(false);
					parentEditor.setNumberText("");
					parentEditor.enableBacktrackButtons(true);
					parentEditor.setPromptText("Choose a transition");
				}
			} catch(Exception e) {
				e.printStackTrace();
			}
		}
	}
	

	public Color getColor(RGB rgb) {
		return colorProvider.getColor(rgb);
	}
	
	private CounterExample getGlobalStates(String content) {
		try{
			ANTLRStringStream cs = new ANTLRStringStream(content);
			counterexampleLexer lexer1 = new counterexampleLexer(cs);
		    tokens = new CommonTokenStream();
		    tokens.setTokenSource(lexer1);
		    counterexampleParser parser = new counterexampleParser(tokens);
		    CounterExample r = parser.is();
		    return r;
		} catch(Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	private EnabledTransitions getEnabledTransitions(String content) {
		try{
			ANTLRStringStream cs = new ANTLRStringStream(content);
			enabledtransitionsLexer lexer1 = new enabledtransitionsLexer(cs);
		    tokens = new CommonTokenStream();
		    tokens.setTokenSource(lexer1);
		    enabledtransitionsParser parser = new enabledtransitionsParser(tokens);
		    EnabledTransitions r = parser.is();
		    return r;
		} catch(Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public boolean isExplicitInteractiveExecution() {
		return isInteractiveMode;
	}
	
	public void incrementStep() {
		step ++;
	}
	
	public boolean saveFile() {
		if(this.isDirty()) {
			boolean save = MessageDialog.openConfirm(null, "File changed", "The content has been changed. Do you want to save it to proceed?");
			if(save)
				this.doSave(null);
			else
				return false;
		}
		return true;
	}
	
	public int[] exploration(int option) {
		is.storeIniStates();
		
		int[] results = option==0 ? is.exploration() : is.por_exploration();
		
		return results;
	}
}
