package org.mcmas.ui.editors;


import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.ScrollPane;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.lang.reflect.InvocationTargetException;
import java.net.URI;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.StringTokenizer;

import javax.swing.JScrollPane;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeNodeStream;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.FindReplaceDocumentAdapter;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.TypedRegion;
import org.eclipse.jface.text.rules.IToken;
import org.eclipse.swt.SWT;
import org.eclipse.swt.awt.SWT_AWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.custom.StyleRange;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowData;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.FontDialog;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.*;
//import org.eclipse.ui.editors.text.TextEditor;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.part.MultiPageEditorPart;
import org.eclipse.ui.ide.FileStoreEditorInput;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.forms.widgets.Form;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.ScrolledForm;
import org.mcmas.ui.Activator;
import org.mcmas.ui.preferences.PreferenceConstants;
import org.mcmas.ui.syntax.CounterExample;
import org.mcmas.ui.syntax.ISPLLexer;
import org.mcmas.ui.syntax.ISPLParser;
import org.mcmas.ui.syntax.ISPLTREE;
import org.mcmas.ui.syntax.Util;
import org.mcmas.ui.syntax.counterexampleLexer;
import org.mcmas.ui.syntax.counterexampleParser;

/**
 * An example showing how to create a multi-page editor.
 * This example has 3 pages:
 * <ul>
 * <li>page 0 contains a nested text editor.
 * <li>page 1 allows you to change the font used in page 2
 * <li>page 2 shows the words in page 0 in sorted order
 * </ul>
 */
public class MultiPageEditor extends MultiPageEditorPart implements IResourceChangeListener{

	/** The text editor used in page 0. */
	private MCMASEditor editor;
	
	/** The font chosen in page 1. */
	private Font font;
	private StyledText text1;
	private SashForm sashForm; 
	private StyledText text;
	private StyledText text2;
	private Text numberText;
	private Button Choose;
	private Button Backtrack;
	private Button Quit;
	private Label label;
	private boolean chosenFlag;
	private Label prompt;
	
	/** The text widget used in page 2. */
	private ScrolledForm form;
	private FormToolkit toolkit;
	private Composite formParent;

	private int osType;
	private counterexampleLexer lexer;
	private counterexampleParser parser;
	private CommonTokenStream tokens;
	
	private boolean isModelChecking = false;
	private String sPath;
	private String title;
	private String MCMASPath;
	private String dotPath;
	private String cygwinPath;
	//private Process proc;
	//private StreamGobbler outputGobbler;
	
	private ArrayList<Label> formLabels = new ArrayList<Label>();
	private ArrayList<Text> formTexts = new ArrayList<Text>();
	private ArrayList<Button> formButtons = new ArrayList<Button>();
	private StyledText t1;
	private BddButtonSelectionAdapter bddadapter;
	private Button bddinfo;
	private ArrayList<SelectionAdapter> cwadapter = new ArrayList<SelectionAdapter>();

	public String output;
	public Dialog cancelDialog;
	
	private class CES {
		String name;
		int count;
		CounterExample counterexample;
		CounterExample projected;
		boolean orphan;
		Composite c;
		
		public CES(String name, int count, CounterExample counterexample, Composite c) {
			this.name = name;
			this.count =count;
			this.counterexample = counterexample;
			this.projected = counterexample; 
			this.c = c;
			orphan = false;
		}
		
		public void setProjected(String[] names) {
			projected = counterexample.project(names);
		}
		
		public CounterExample getProjected() {
			return projected;
		}
		
		public void resetProjected() {
			projected = counterexample;
		}
		
		public boolean isOrphan() {
			return orphan;
		}
		
		public void setOrphan() {
			if(!orphan) {
				orphan = true;
				addCloseButton(name, count);
			}
		}
		
		private void addCloseButton(final String name, final int count) {
			Button close = new Button(c, SWT.PUSH);
			close.setForeground(editor.getColor(IsplColorProvider.OPERATOR));
			close.setText("Close");
			close.addSelectionListener(new SelectionAdapter() {
				String cesname = name;
				int cescount = count;
				public void widgetSelected(SelectionEvent e) {
					boolean result = MessageDialog.openConfirm(null, "Confirm", "Do you want to close this page?");
					if(result) {
						closePage(cesname, cescount);
			    		//disposeCounterExamplePage(pageIndex);
					}
				}
			});
			c.layout();
		}
	}

	final private Hashtable<Integer, CES> cesPages = new Hashtable<Integer, CES>();
	final private ArrayList<Integer> cesPageIndex = new ArrayList<Integer>();
	/**
	 * Creates a multi-page editor example.
	 */
	public MultiPageEditor() {
		super();
		ResourcesPlugin.getWorkspace().addResourceChangeListener(this);
		String sOsName = System.getProperty("os.name");
		if (sOsName.startsWith("Windows") || sOsName.startsWith("windows"))
            osType = 0;
        else if (sOsName.startsWith("Linux") || sOsName.startsWith("linux"))
            osType = 1;
        else
        	osType = 2;
	}
	/**
	 * Creates page 0 of the multi-page editor,
	 * which contains a text editor.
	 */
	void createPage0() {
		try {
			editor = new MCMASEditor();
			editor.setMultiPageEditor(this);
			int index = addPage(editor, getEditorInput());
			setPageText(index, editor.getTitle());
		} catch (PartInitException e) {
			ErrorDialog.openError(
				getSite().getShell(),
				"Error creating nested text editor",
				null,
				e.getStatus());
		}
	}
	/**
	 * Creates page 1 of the multi-page editor,
	 * which allows you to change the font used in page 2.
	 */
	void createPage1() {

		/*Composite composite = new Composite(getContainer(), SWT.NONE);
		GridLayout layout = new GridLayout();
		composite.setLayout(layout);
		layout.numColumns = 2;

		Button fontButton = new Button(composite, SWT.NONE);
		GridData gd = new GridData(GridData.BEGINNING);
		gd.horizontalSpan = 2;
		fontButton.setLayoutData(gd);
		fontButton.setText("Change Font...");
		
		fontButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent event) {
				setFont();
			}
		});

		int index = addPage(composite);*/
		Composite composite = new Composite(getContainer(), SWT.NONE);
		FillLayout layout = new FillLayout();
		composite.setLayout(layout);
		composite.setLayoutData(new GridData(GridData.FILL_BOTH));
	    text = new StyledText(composite, SWT.H_SCROLL | SWT.V_SCROLL);
		text.setEditable(false);
		
		sashForm = new SashForm(composite, SWT.VERTICAL);
		sashForm.SASH_WIDTH = 5;

		prompt = new Label(sashForm, SWT.NULL);
		prompt.setText("");
		FontData fontData = prompt.getFont().getFontData()[0];
		final int fontHeight = fontData.getHeight();
		fontData.setHeight(fontHeight*2);
		fontData.setStyle(SWT.BOLD);
		prompt.addControlListener(new ControlAdapter(){
			//Label tmpprompt = prompt;
			int normalHeight = fontHeight;
			public void controlResized(ControlEvent e) {
				Point rect = prompt.getSize();
		    	FontData fontData = prompt.getFont().getFontData()[0];
		    	int currentHeight = fontData.getHeight();
		    	GC g = new GC(prompt);
			    if(rect.y < normalHeight * 2 + 10 /*&& currentHeight != normalHeight*/) {
					fontData.setHeight(normalHeight);
					//fontData.setStyle(SWT.BOLD);
					if (font != null)
						font.dispose();
					font = new Font(prompt.getDisplay(), fontData);
					//prompt.setFont(font);
			    } else if(rect.y >= normalHeight * 4 + 10 /*&& currentHeight != normalHeight * 4*/){
					fontData.setHeight(normalHeight*4);
					//fontData.setStyle(SWT.BOLD);
					if (font != null)
						font.dispose();
					font = new Font(prompt.getDisplay(), fontData);
					//prompt.setFont(font);
			    } else if(rect.y >= normalHeight * 2 + 10 /*&& currentHeight != normalHeight * 2*/){
					fontData.setHeight(normalHeight*2);
					//fontData.setStyle(SWT.BOLD);
					if (font != null)
						font.dispose();
					font = new Font(prompt.getDisplay(), fontData);
					//prompt.setFont(font);
			    } 
			    g.setFont(font);
			    String str = prompt.getText();
			    Point stringSize = g.stringExtent(str);
			    while(stringSize.x > rect.x && fontData.getHeight() > normalHeight) {
			    	fontData.setHeight(fontData.getHeight()/2);
			    	if (font != null)
						font.dispose();
					font = new Font(prompt.getDisplay(), fontData);
					g.setFont(font);
				    stringSize = g.stringExtent(str);
			    }
			    prompt.setFont(font);
			    g.dispose();
			}
		});
		
		if (font != null)
			font.dispose();
		font = new Font(prompt.getDisplay(), fontData);
		prompt.setFont(font);
		
		text2 = new StyledText(sashForm, SWT.H_SCROLL | SWT.V_SCROLL);
		text2.setEditable(false);	
		
		Composite buttons = new Composite(sashForm, SWT.NONE);
		GridLayout layout1 = new GridLayout();
		buttons.setLayout(layout1);
		layout1.numColumns = 5;
		layout1.verticalSpacing = 9;
		
		label = new Label(buttons, SWT.NULL);
		label.setText("No.:");
		numberText = new Text(buttons, SWT.BORDER | SWT.SINGLE);
		GridData gd = new GridData(GridData.FILL_HORIZONTAL);
		numberText.setLayoutData(gd);
		Choose = new Button(buttons, SWT.PUSH);
		Choose.setText("Choose");
		Choose.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	        chooseNumber();
	      }
	    });

	    Backtrack = new Button(buttons, SWT.PUSH);
	    Backtrack.setText("Backtrack");
	    Backtrack.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	        backtrackState();
	      }
	    });

	    Quit = new Button(buttons, SWT.PUSH);
	    Quit.setText("Quit");
	    Quit.addSelectionListener(new SelectionAdapter() {
	      public void widgetSelected(SelectionEvent event) {
	        quitInteractiveMode();
	      }
	    });

	    int[] weights = {5, 85, 10};
	    sashForm.setWeights(weights);

	    enableButtons(false);
	    
	    int index = addPage(composite);
		setPageText(index, "Interactive Mode");
	}
	/**
	 * Creates page 2 of the multi-page editor,
	 * which shows the sorted text.
	 */
	void createPage2() {
		formParent = new Composite(getContainer(), SWT.NONE);
		FillLayout layout = new FillLayout();
		layout.type = SWT.VERTICAL;
		formParent.setLayout(layout);
		//text1 = new StyledText(composite, SWT.H_SCROLL | SWT.V_SCROLL);
		//text1.setEditable(false);
		toolkit = new FormToolkit(formParent.getDisplay());
		form = toolkit.createScrolledForm(formParent);
		form.setText("Verification result");
		GridLayout layout1 = new GridLayout();
		form.getBody().setLayout(layout1);
		layout1.numColumns = 4;
		layout1.verticalSpacing = 9;
		t1 = new StyledText(formParent, SWT.H_SCROLL | SWT.V_SCROLL);
		t1.setEditable(false);
		
		int index = addPage(formParent);
		setPageText(index, "Model Checking");
	}

	int createCounterExamplePage(String dotPath, File dotfile, File textfile, String strFormula) {
		Composite newroot = new Composite(getContainer(), SWT.NONE);
		GridLayout filllayout = new GridLayout();
		filllayout.numColumns = 1;
		newroot.setLayout(filllayout);
		
		/*FormToolkit toolkit1 = new FormToolkit(newroot.getDisplay());
		Form newform = toolkit1.createForm(newroot);
		newform.setText(strFormula);
		newform.getBody().setLayout(new GridLayout());*/
		Label formula = new Label(newroot, SWT.READ_ONLY);
		//formula.setText(strFormula);
		formula.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		
		//Composite root = new Composite(getContainer(), SWT.NONE);
		Composite root = new Composite(newroot, SWT.NONE);
		FillLayout layout1 = new FillLayout();
		//GridLayout layout1 = new GridLayout();
		root.setLayout(layout1);
		root.setLayoutData(new GridData(GridData.FILL_BOTH));
		
		final SashForm sashForm1 = new SashForm(root, SWT.HORIZONTAL);
		sashForm1.SASH_WIDTH = 5;
		
		//Composite composite = new Composite(root, SWT.EMBEDDED);
		Composite composite = new Composite(sashForm1, SWT.EMBEDDED);
		FillLayout layout = new FillLayout();
		composite.setLayout(layout);
		Frame flowChart = SWT_AWT.new_Frame(composite);
		BorderLayout borderLayout1 = new BorderLayout();
		flowChart.setLayout(borderLayout1);
		final FlowCanvas jPFlow = new FlowCanvas(flowChart);
		jPFlow.setFile(dotfile);
        jPFlow.readDotFile(dotPath, osType);
        Dimension d = jPFlow.getCanvasSize();
        jPFlow.setMinimumSize(d);
        jPFlow.setPreferredSize(d);

        JScrollPane jSPFLowChart = new JScrollPane();
        flowChart.add(jSPFLowChart, java.awt.BorderLayout.CENTER);
        jSPFLowChart.getViewport().add(jPFlow);
        jPFlow.setBackground(Color.white);

		/*SashForm sashForm1 = new SashForm(root, SWT.HORIZONTAL);
		sashForm1.SASH_WIDTH = 5;*/
		final StyledText info = new StyledText(sashForm1, SWT.H_SCROLL | SWT.V_SCROLL | SWT.BORDER);
	    BufferedReader bis = null;
	    String content = "";
	    CounterExample r = null;
	    try {
		      bis = new BufferedReader(new InputStreamReader(new FileInputStream(textfile)));
		      String s = bis.readLine();
		      while ( s != null) {
		    	  content += s + "\n";
		    	  s = bis.readLine();
		      }
		      bis.close();

		      ANTLRStringStream cs = new ANTLRStringStream(content);
		      lexer = new counterexampleLexer(cs);
		      tokens = new CommonTokenStream();
		      tokens.setTokenSource(lexer);
		      parser = new counterexampleParser(tokens);
		      r = parser.is();
		      setCounterExampleText(r, info);
		      jPFlow.setCounterExampleInfo(info, r.getStatePositions());
	    } catch (FileNotFoundException e) {
	      e.printStackTrace();
	    } catch (IOException e) {
	      e.printStackTrace();
	    } catch(RecognitionException e) {
			e.printStackTrace(System.out);
		} catch(Exception e) {
			e.printStackTrace();
        }
		
		final String[] allAgents = r.getAllAgents();
		
		Composite c = new Composite(sashForm1, SWT.NONE);
		GridLayout layout2 = new GridLayout();
		c.setLayout(layout2);
		layout2.numColumns = 2;
		layout2.verticalSpacing = 9;
		Group group = new Group(c, SWT.NONE);
		GridData gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		gd.horizontalSpan = 2;
		group.setLayoutData(gd);
		RowLayout rl = new RowLayout();
		rl.type = SWT.VERTICAL;
		group.setLayout(rl);

		final Button[] checkButtons = new Button[allAgents.length];
		for(int i=0; i<allAgents.length; i++) {
			checkButtons[i] = new Button(group, SWT.CHECK);
			checkButtons[i].setText(allAgents[i]);
			checkButtons[i].setSelection(true);
		}
		
		Button selectAll = new Button(c, SWT.PUSH);
		selectAll.setText("Select all");
		selectAll.addSelectionListener(new SelectionAdapter() {
			Button[] buttons = checkButtons;
			public void widgetSelected(SelectionEvent e) {
				for(int i=0; i<buttons.length; i++)
					buttons[i].setSelection(true);
			}
		});

		Button unselectAll = new Button(c, SWT.PUSH);
		unselectAll.setText("Unselect all");
		unselectAll.addSelectionListener(new SelectionAdapter() {
			Button[] buttons = checkButtons;
			public void widgetSelected(SelectionEvent e) {
				for(int i=0; i<buttons.length; i++)
					buttons[i].setSelection(false);
			}
		});
		
		sashForm1.addControlListener(new ControlAdapter() {
			public void controlResized(ControlEvent e) {
				Shell shell = PlatformUI.getWorkbench().getDisplay().getActiveShell();
				Rectangle clientArea = shell.getClientArea();
				if(clientArea.width >= 1680)
					sashForm1.setWeights(new int[]{70,15,15});
				else if(clientArea.width >= 1280)
					sashForm1.setWeights(new int[]{60,20,20});
				else 
					sashForm1.setWeights(new int[]{50,25,25});
			}
		});
		
 
		//sashForm1.setWeights(new int[]{70,17,13});

		//final int index = addPage(root);
		final int index = addPage(newroot);

		String name1 = dotfile.getName();
		final String name = name1.substring(0, name1.lastIndexOf(".dot")).replace("formula", "formula ");
		final int count = getPageNameCount(name);
		CES ces = new CES(name, count, r, c);
		addCounterExamplePage(index, ces);
		
		Button apply = new Button(c, SWT.PUSH);
		apply.setText("Apply");
		apply.addSelectionListener(new SelectionAdapter() {
			Button[] buttons = checkButtons;
			public void widgetSelected(SelectionEvent e) {
				ArrayList<String> names = new ArrayList<String>();
				for(int i=0; i<buttons.length; i++)
					if(buttons[i].getSelection())
						names.add(buttons[i].getText());
				if(names.isEmpty()) {
					MessageDialog.openError(null, "Error", "At least one agent must be chosen!");
					return;
				}
				CES ces = getCounterExamplePageData(name, count);
				if(ces != null) {
					String[] nameArray = new String[names.size()];
					names.toArray(nameArray);
					ces.setProjected(nameArray);
					setCounterExampleText(ces.projected, info);
				    jPFlow.setCounterExampleInfo(info, ces.getProjected().getStatePositions());
				}
			}
		});
		
		setPageText(index, (count == 0 ? name : name + " (" + count + ")"));
		setActivePage(index);
		
		/*int sashOffset = 200;
		Point sashSize = sashForm1.getSize();
		Point sashLocation = sashForm1.getLocation();
		Point compositeSize = composite.getSize();
		sashForm1.setSize(sashSize.x-sashOffset, sashSize.y);
		sashForm1.setLocation(sashLocation.x+sashOffset, sashLocation.y);
		composite.setSize(compositeSize.x+sashOffset, compositeSize.y);*/
			
		formula.setText(strFormula);
		return index;
	}

	private int getPageNameCount(String name) {
		for(int i=cesPageIndex.size()-1; i>=0; i--) {
			CES ces = cesPages.get(cesPageIndex.get(i));
			if(ces.name.compareTo(name) == 0)
				return ces.count + 1;
		}
		return 0;
	}
	
	private void addCounterExamplePage(int index, CES ces) {
		cesPages.put(new Integer(index), ces);
		cesPageIndex.add(new Integer(index));
	}
	
	public CES getCES(int index) {
		return cesPages.get(new Integer(index));
	}
	
	public boolean hasOpenCEPage() {
		return cesPages.isEmpty();
	}
	
	private CES getCounterExamplePageData(String name, int index) {
		for(int i=0; i<cesPageIndex.size(); i++) {
			Integer key = cesPageIndex.get(i);
			CES ces = cesPages.get(key);
			if(ces.name.compareTo(name)==0 && ces.count==index) {
				return ces;
			}
		}
		return null;
	}
	
	private void markOrphan() {
		for(int i=0; i<cesPageIndex.size(); i++) {
			Integer index = cesPageIndex.get(i);
			CES ces = cesPages.get(index);
			if(!ces.isOrphan()) {
				ces.setOrphan();
			}
		}
	}
	
	private void setCounterExampleText(CounterExample r, StyledText info) {
		String content = r.toString();
		info.setText(content);
		for (int i = 0; i < r.numberOfStates(); i++) {
			int line = r.getFirstLine(i);
			info.setLineBackground(line, 1, editor
					.getColor(IsplColorProvider.CYAN));
		}
		for (int i = 0; i < r.numberOfAgents(); i++) {
			Position p = r.getAgentPosition(i);
			StyleRange styleRange = new StyleRange();
			styleRange.start = p.offset;
			styleRange.length = p.length;
			styleRange.fontStyle = SWT.BOLD;
			styleRange.foreground = editor.getColor(IsplColorProvider.KEYWORD);
			info.setStyleRange(styleRange);
			p = r.getNamePosition(i);
			styleRange = new StyleRange();
			styleRange.start = p.offset;
			styleRange.length = p.length;
			styleRange.fontStyle = SWT.BOLD;
			styleRange.foreground = editor.getColor(IsplColorProvider.OPERATOR);
			info.setStyleRange(styleRange);
		}
	}
	
	/**
	 * Creates the pages of the multi-page editor.
	 */
	protected void createPages() {
		createPage0();
		createPage1();
		createPage2();
	}
	/**
	 * The <code>MultiPageEditorPart</code> implementation of this 
	 * <code>IWorkbenchPart</code> method disposes all nested editors.
	 * Subclasses may extend.
	 */
	public void dispose() {
		if(editor != null)
			editor.dispose();
		if(font != null)
			font.dispose();
		if(text1 != null)
			text1.dispose();
		if(sashForm != null)
			sashForm.dispose(); 
		if(text != null)
			text.dispose();
		if(text2 != null)
			text2.dispose();
		if(numberText != null)
			numberText.dispose();
		if(Choose != null)
			Choose.dispose();
		if(Backtrack != null)
			Backtrack.dispose();
		if(Quit != null)
			Quit.dispose();
		if(label != null)
			label.dispose();
		if(prompt != null)
			prompt.dispose();
		if(form != null)
			form.dispose();
		if(toolkit != null)
			toolkit.dispose();
		if(formParent != null)
			formParent.dispose();
		
		ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
		super.dispose();
	}
	/**
	 * Saves the multi-page editor's document.
	 */
	public void doSave(IProgressMonitor monitor) {
		getEditor(0).doSave(monitor);
	}
	/**
	 * Saves the multi-page editor's document as another file.
	 * Also updates the text for page 0's tab, and updates this multi-page editor's input
	 * to correspond to the nested editor's.
	 */
	public void doSaveAs() {
		IEditorPart editor = getEditor(0);
		editor.doSaveAs();
		setPageText(0, editor.getTitle());
		setInput(editor.getEditorInput());
	}
	/* (non-Javadoc)
	 * Method declared on IEditorPart
	 */
	public void gotoMarker(IMarker marker) {
		setActivePage(0);
		IDE.gotoMarker(getEditor(0), marker);
	}
	/**
	 * The <code>MultiPageEditorExample</code> implementation of this method
	 * checks that the input is an instance of <code>IFileEditorInput</code>.
	 */
	public void init(IEditorSite site, IEditorInput editorInput)
		throws PartInitException {
		if (!(editorInput instanceof IFileEditorInput) && !(editorInput instanceof FileStoreEditorInput))
			throw new PartInitException("Invalid Input: Must be IFileEditorInput");
			
		super.init(site, editorInput);
		setPartName(editorInput.getName());
	}
	/* (non-Javadoc)
	 * Method declared on IEditorPart.
	 */
	public boolean isSaveAsAllowed() {
		return true;
	}
	/**
	 * Calculates the contents of page 2 when the it is activated.
	 */
	protected void pageChange(int newPageIndex) {
		super.pageChange(newPageIndex);
		/* if (newPageIndex == 2) {
			sortWords();
		} */
	}
	/**
	 * Closes all project files on project close.
	 */
	public void resourceChanged(final IResourceChangeEvent event){
		if(event.getType() == IResourceChangeEvent.PRE_CLOSE){
			Display.getDefault().asyncExec(new Runnable(){
				public void run(){
					IWorkbenchPage[] pages = getSite().getWorkbenchWindow().getPages();
					for (int i = 0; i<pages.length; i++){
						if(((FileEditorInput)editor.getEditorInput()).getFile().getProject().equals(event.getResource())){
							IEditorPart editorPart = pages[i].findEditor(editor.getEditorInput());
							pages[i].closeEditor(editorPart,true);
						}
					}
				}            
			});
		}
	}
	/**
	 * Sets the font related data to be applied to the text in page 2.
	 */
	/*void setFont() {
		FontDialog fontDialog = new FontDialog(getSite().getShell());
		fontDialog.setFontList(text.getFont().getFontData());
		FontData fontData = fontDialog.open();
		if (fontData != null) {
			if (font != null)
				font.dispose();
			font = new Font(text.getDisplay(), fontData);
			text1.setFont(font);
		}
	}*/
	/**
	 * Sorts the words in page 0, and shows them in page 2.
	 */
	/*void sortWords() {

		String editorText =
			editor.getDocumentProvider().getDocument(editor.getEditorInput()).get();

		StringTokenizer tokenizer =
			new StringTokenizer(editorText, " \t\n\r\f!@#\u0024%^&*()-_=+`~[]{};:'\",.<>/?|\\");
		ArrayList editorWords = new ArrayList();
		while (tokenizer.hasMoreTokens()) {
			editorWords.add(tokenizer.nextToken());
		}

		Collections.sort(editorWords, Collator.getInstance());
		StringWriter displayText = new StringWriter();
		for (int i = 0; i < editorWords.size(); i++) {
			displayText.write(((String) editorWords.get(i)));
			displayText.write(System.getProperty("line.separator"));
		}
		text1.setText(displayText.toString());
	}*/

	private void quitInteractiveMode() {
		editor.quitInteractiveMode();
		enableButtons(false);
	}
	
	private void chooseNumber() {
		String strNumber = numberText.getText();
		try{
			int number = Integer.parseInt(strNumber);
			if(chosenFlag)
				if(editor.isExplicitInteractiveExecution())
					editor.chooseState(number);
				else {
					//editor.incrementStep();
					editor.chooseState1(number);
				}
			else
				if(editor.isExplicitInteractiveExecution())
					editor.chooseAction(number);
				else
					editor.chooseAction1(number);
			numberText.setFocus();
		} catch(NumberFormatException e) {
			MessageDialog.openError(null, "Error", "Please enter an valid number!");
		}
	}
	
	private void backtrackState() {
		editor.backtrack();
		numberText.setFocus();
	}
	
	public void enableButtons(boolean status){
		label.setEnabled(status);
		numberText.setEnabled(status);
		Choose.setEnabled(status);
		Backtrack.setEnabled(status);
		Quit.setEnabled(status);
		numberText.setFocus();
	}
	
	public void enterInteractiveMode() {
		setActivePage(1);
	}

	public void enterModelChecking() {
		setActivePage(2);
	}

	public void appendText(String str) {
		int pos = text.getCaretOffset();
		text.append(str);
		int lastpos = pos + str.length();
		text.setCaretOffset(lastpos);
		text.setSelection(lastpos);
	}

	public void setText(String str) {
		text.setText(str);
	}
	
	public void deleteText(int start) {
		int length = text.getCharCount()-start;
		text.replaceTextRange(start, length, "");
	}


	public void setText2(String str) {
		text2.setText(str);
	}
	
	public void setNumberText(String str) {
		numberText.setText(str);
	}
	
	public void setChosenFlag(boolean flag) {
		chosenFlag = flag;
	}
	
	public void enableChooseButtons(boolean status) {
		label.setEnabled(status);
		numberText.setEnabled(status);
		Choose.setEnabled(status);
	}

	public void enableBacktrackButtons(boolean status) {
		Backtrack.setEnabled(status);
	}
	
	public MCMASEditor getMCMASEditor() {
		return editor;
	}
	
	public void setPromptText(String str) {
		prompt.setText(str);
	}
	
	public void displayResultsInForm(String[] results, final String dotPath, final String path, final String title) {
		//markOrphan();
		//if(form != null && !form.isDisposed())
		//	form.dispose();
		/*form = toolkit.createScrolledForm(formParent);
		form.setText("Verification result");
		GridLayout layout1 = new GridLayout();
		form.getBody().setLayout(layout1);
		layout1.numColumns = 3;
		layout1.verticalSpacing = 9;*/
		String fullpath = path+"."+ title + (osType==0 ? "\\" : "/");
		final String dPath = (Util.isEmpty(dotPath) || dotPath.endsWith("/") || dotPath.endsWith("\\") ? 
				dotPath : (osType == 0 ? dotPath+"\\" : dotPath+"/"));
		
		String log = "";
		//int maxTextSize = 0;
		for(int i=0; i<results.length; i++) {
			if(results[i] != null && !Util.isEmpty(results[i])) {
				if(results[i].compareTo("FALSE")==0 || results[i].compareTo("TRUE")==0) {
					Label l1 = toolkit.createLabel(form.getBody(), "Formula " + (i+1) + ": ");
					final String strFormula = editor.is.is_formulae.get(i).toString();
					Text t1 = toolkit.createText(form.getBody(), "", SWT.READ_ONLY);
					t1.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
					Label l2 = toolkit.createLabel(form.getBody(), results[i]);
					formLabels.add(l1);
					formTexts.add(t1);
					formLabels.add(l2);
					final File f = new File(fullpath + "formula" + (i+1) + ".dot");
					final File f1 = new File(fullpath + "formula" + (i+1) + ".info");
					if(f.exists()) {
						final Button b3 = toolkit.createButton(form.getBody(), "show counterexample/witness", SWT.PUSH);
						formButtons.add(b3);
						SelectionAdapter sa = new SelectionAdapter() {
							int count = 0;
							int index;
							String name;
						      public void widgetSelected(SelectionEvent event) {
						    	  if(count == 0) {
						    		  b3.setText("close counterexample/witness");
						    		  count = 1;
						    		  int t = createCounterExamplePage(dPath, f, f1, strFormula);
						    		  CES ces = getCES(t);
						    		  name = ces.name;
						    		  index = ces.count;
						    	  } else {
						    		  b3.setText("show counterexample/witness");
						    		  count = 0;
						    		  closePage(name, index);
						    		  //disposeCounterExamplePage(name, index);
						    	  }
						      }
						    };
						b3.addSelectionListener(sa);
						cwadapter.add(sa);
					} else {
						Label l3 = toolkit.createLabel(form.getBody(), "");
						formLabels.add(l3);
					}
					if(i % 2 == 0) {
						l1.setBackground(editor.getColor(IsplColorProvider.YELLOW));
						l2.setBackground(editor.getColor(IsplColorProvider.YELLOW));
						//l3.setBackground(editor.getColor(IsplColorProvider.YELLOW));
					} else {
						l1.setBackground(editor.getColor(IsplColorProvider.CYAN));
						l2.setBackground(editor.getColor(IsplColorProvider.CYAN));
						//l3.setBackground(editor.getColor(IsplColorProvider.CYAN));
					}
				} else {
					log += results[i] + "\n";
				}
			}
		}
		
		if(!Util.isEmpty(log)) {
			//StyledText t1 = new StyledText(formParent, SWT.H_SCROLL | SWT.V_SCROLL);
			//t1.setEditable(false);
			Label l1 = toolkit.createLabel(form.getBody(), " ");
			Label l2 = toolkit.createLabel(form.getBody(), " ");
			Label l3 = toolkit.createLabel(form.getBody(), " ");
			formLabels.add(l1);
			formLabels.add(l2);
			formLabels.add(l3);
			bddinfo = toolkit.createButton(form.getBody(), "show BDD information", SWT.PUSH);
			bddadapter = new BddButtonSelectionAdapter(log, bddinfo, t1);
			bddinfo.addSelectionListener(bddadapter);
		}
		form.pack();	
		formParent.layout();
		
		
		for(int i=0; i<formTexts.size(); i++) 
			formTexts.get(i).setText(editor.is.is_formulae.get(i).toString());
	}
	
	public void closePage(String name, int index) {
		int i;
		Integer key;
		for(i=0; i<cesPageIndex.size(); i++) {
			key = cesPageIndex.get(i);
			CES ces = cesPages.get(key);
			if(ces.name.compareTo(name)==0 && ces.count==index) {
				removePage(key.intValue());
				cesPages.remove(key);
				cesPageIndex.remove(i);
				break;
			}
		}
		//ArrayList<Integer> temp = new ArrayList<Integer>();
		for(int j=i; j<cesPageIndex.size(); j++) {
			Integer k = cesPageIndex.remove(j);
			CES ces = cesPages.remove(k);
			Integer newk = new Integer(k-1);
			cesPages.put(newk, ces);
			cesPageIndex.add(j, newk);
		}
	}
	
	public void setPathAndTitle() {
		if(sPath == null || Util.isEmpty(sPath)) {
			IEditorInput editorInput = editor.getEditorInput();
			if(editorInput instanceof IFileEditorInput) {
				IFileEditorInput fInput = (IFileEditorInput)editorInput;
				String path = fInput.getFile().getLocation().toOSString();
				title = fInput.getFile().getName();
				int index = path.lastIndexOf(title);
				sPath = path.substring(0, index);
				//System.out.println("Full path = " + sPath);
			} else {
				URI uri = ((FileStoreEditorInput)editorInput).getURI();
				IFile[] files = ResourcesPlugin.getWorkspace().getRoot().findFilesForLocationURI(uri);
				if (files != null && files.length > 0) {
					String path = files[0].getLocation().toOSString();
					title = files[0].getName();
					int index = path.lastIndexOf(title);
					sPath = path.substring(0, index);
				}
			}
		}
	}
	
	public String getFilePath() {
		if(sPath == null || Util.isEmpty(sPath)) 
			setPathAndTitle();
		return sPath;
	}
	public String getFileTitle() {
		if(title == null || Util.isEmpty(title)) 
			setPathAndTitle();
		return title;
	}

	public void modelChecking() {
		/*if(isModelChecking) {
			boolean confirmed = MessageDialog.openQuestion(null, "Warning", "Model checking process is running! Restart it?");
			if(!confirmed)
				return; 
		}*/
		markOrphan();
		//if(form != null && !form.isDisposed())
		//	form.dispose();
		
		for(int i=0; i<formLabels.size(); i++) {
			Label l = formLabels.get(i);
			l.dispose();
		}
		formLabels.clear();
		for(int i=0; i<formTexts.size(); i++) {
			Text l = formTexts.get(i);
			l.dispose();
		}
		formTexts.clear();

		for(int i=0; i<formButtons.size(); i++) {
			Button b = formButtons.get(i);
			SelectionAdapter sa = cwadapter.get(i);
			b.removeSelectionListener(sa);
			b.dispose();
		}
		formButtons.clear();
		cwadapter.clear();
		t1.setText("");
		if(bddinfo != null && !bddinfo.isDisposed()) {
			bddinfo.removeSelectionListener(bddadapter);
			bddinfo.dispose();
		}

		//getSourceViewer().setEditable(false);
		setPathAndTitle();
		
		IPreferenceStore store = Activator.getDefault().getPreferenceStore();
		MCMASPath = store.getString(PreferenceConstants.MCMAS_PATH);
		dotPath = store.getString(PreferenceConstants.DOT_PATH);
		cygwinPath = store.getString(PreferenceConstants.CYGWIN_PATH);
		
		output = "";
		String filepath = sPath + title;
		String cexpath = sPath + "." + title;
		
		File f = new File(cexpath);
		boolean dirExist = true; 
		if(!f.exists())
			dirExist = f.mkdir();
		
		String slash = osType == 0 ? "\\" : "/";
		String[] mcmasCommand = {MCMASPath + slash + "mcmas", "-q", "-bc", "1", "-c", "-bdd_stats", "-cex_prefix",
				cexpath+slash, filepath};
		if(dirExist) {
			// delete old counterexample files
			File[] cexfiles = f.listFiles();
			if(cexfiles != null)
				for(int i=0; i<cexfiles.length; i++)
					cexfiles[i].delete();
		}
		else {
			System.out.println("Cannot create the directory " + cexpath + ". Counterexample generation is disabled!");
			mcmasCommand = new String[]{MCMASPath + slash + "mcmas", "-q", "-bc", "1", "-c", "-bdd_stats", filepath};
		}
		
		try {
			ExecuteMCMAS et = new ExecuteMCMAS(this, mcmasCommand, osType, "Path="+cygwinPath);
			new ProgressMonitorDialog(getSite().getShell()).run(true, true,	et);
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			System.out.println("Executing MCMAS was interrupted by the user.");
		}
	    if(!Util.isEmpty(output)) {
	       	String[] results = output.split("\n");
	        displayResultsInForm(results, dotPath, sPath, title);
	    }
	}
	
	public int getOsType() {
		return osType;
	}
}

class BddButtonSelectionAdapter extends SelectionAdapter {
	private int count;
	private String info;
	private Button b3;
	private StyledText t1;
	public BddButtonSelectionAdapter(String info, Button b3, StyledText t1) {
		super();
		count = 0;
		this.info = info;
		this.b3 = b3;
		this.t1 = t1;
	}
	
	public void widgetSelected(SelectionEvent event) {
  	  if(count == 0) {
  		  b3.setText("close BDD information");
  		  count = 1;
  		  t1.setText(info);
  	  } else {
  		  b3.setText("show BDD information");
  		  count = 0;
  		  t1.setText("");
  	  }
    }
}

