package org.mcmas.ui.editors;

import java.util.ResourceBundle;

import org.eclipse.jface.action.*;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.editors.text.TextEditorActionContributor;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.ide.IDEActionFactory;
import org.eclipse.ui.part.MultiPageEditorActionBarContributor;
import org.eclipse.ui.part.MultiPageEditorPart;
import org.eclipse.ui.part.MultiPageEditorSite;
import org.eclipse.ui.texteditor.ITextEditor;
import org.eclipse.ui.texteditor.ITextEditorActionConstants;
import org.eclipse.ui.texteditor.ITextEditorActionDefinitionIds;
import org.eclipse.ui.texteditor.RetargetTextEditorAction;


interface IMcmasActionConstants {
	public final String group1 = "group1";
	public final String group2 = "group2";
}

/**
 * Manages the installation/deinstallation of global actions for multi-page editors.
 * Responsible for the redirection of global actions to the active editor.
 * Multi-page contributor replaces the contributors for the individual editors in the multi-page editor.
 */
public class MultiPageEditorContributor extends MultiPageEditorActionBarContributor {
	private IEditorPart activeEditorPart;
	private Action addAgent;
	private Action formatText;
	private Action syntaxChecking;
	private Action interactiveMode;
	private Action interactiveMode1;
	private Action modelChecking;
	//private Action exploration;
	//private Action exploration1;
	
	private TextEditorActionContributor textEditorContributor = null;
	private RetargetTextEditorAction fContentAssistProposal;
	private MultiPageEditor activeMultiPageEditor;
	
	/**
	 * Creates a multi-page contributor.
	 */
	public MultiPageEditorContributor() {
		super();
		createActions();
	}
	/**
	 * Returns the action registed with the given text editor.
	 * @return IAction or null if editor is null.
	 */
	protected IAction getAction(ITextEditor editor, String actionID) {
		return (editor == null ? null : editor.getAction(actionID));
	}
	
	public void contributeToStatusLine(IStatusLineManager statusLineManager) {
		super.contributeToStatusLine(statusLineManager);
		textEditorContributor = new TextEditorActionContributor();
		textEditorContributor.init(getActionBars());
		textEditorContributor.contributeToStatusLine(statusLineManager);
	}


	public void setActiveEditor(IEditorPart part) {
		if(part != null) {
			if(part instanceof MultiPageEditorPart) {
				activeMultiPageEditor = (MultiPageEditor)part;
				//System.out.println("get active multipageeditor");
			}
		}
	}
	
	/* (non-JavaDoc)
	 * Method declared in AbstractMultiPageEditorActionBarContributor.
	 */

	public void setActivePage(IEditorPart part) {
		if (activeEditorPart == part)
			return;

		activeEditorPart = part;

		IActionBars actionBars = getActionBars();
		if (actionBars != null) {

			ITextEditor editor = (part instanceof ITextEditor) ? (ITextEditor) part : null;

			actionBars.setGlobalActionHandler(
				ActionFactory.DELETE.getId(),
				getAction(editor, ITextEditorActionConstants.DELETE));
			actionBars.setGlobalActionHandler(
				ActionFactory.UNDO.getId(),
				getAction(editor, ITextEditorActionConstants.UNDO));
			actionBars.setGlobalActionHandler(
				ActionFactory.REDO.getId(),
				getAction(editor, ITextEditorActionConstants.REDO));
			actionBars.setGlobalActionHandler(
				ActionFactory.CUT.getId(),
				getAction(editor, ITextEditorActionConstants.CUT));
			actionBars.setGlobalActionHandler(
				ActionFactory.COPY.getId(),
				getAction(editor, ITextEditorActionConstants.COPY));
			actionBars.setGlobalActionHandler(
				ActionFactory.PASTE.getId(),
				getAction(editor, ITextEditorActionConstants.PASTE));
			actionBars.setGlobalActionHandler(
				ActionFactory.SELECT_ALL.getId(),
				getAction(editor, ITextEditorActionConstants.SELECT_ALL));
			actionBars.setGlobalActionHandler(
				ActionFactory.FIND.getId(),
				getAction(editor, ITextEditorActionConstants.FIND));
			actionBars.setGlobalActionHandler(
				IDEActionFactory.BOOKMARK.getId(),
				getAction(editor, IDEActionFactory.BOOKMARK.getId()));
			actionBars.updateActionBars();
			
			if (part instanceof ITextEditor) {
				// textEditorContributor.init(actionBars);
				textEditorContributor.setActiveEditor(part);
				fContentAssistProposal.setAction(getAction(editor, "ContentAssistProposal")); //$NON-NLS-1$
			}

		}
	}
	private void createActions() {
		addAgent = new Action() {
			public void run() {
				MCMASEditor editor = (activeEditorPart instanceof MCMASEditor) ? (MCMASEditor) activeEditorPart : null;
				editor.addAgent();
			
				//MessageDialog.openInformation(null, "MCMAS", "An agent has been added");
			}
		};
		addAgent.setText("Add agent");
		addAgent.setToolTipText("Add an agent");
		//addAgent.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
		//		getImageDescriptor(IDE.SharedImages.IMG_OBJS_TASK_TSK));
		
		fContentAssistProposal = new RetargetTextEditorAction(MCMASEditorMessages.getResourceBundle(), "ContentAssistProposal."); //$NON-NLS-1$
		fContentAssistProposal.setActionDefinitionId(ITextEditorActionDefinitionIds.CONTENT_ASSIST_PROPOSALS);
		
		formatText = new Action() {
			public void run() {
				MCMASEditor editor = (activeEditorPart instanceof MCMASEditor) ? (MCMASEditor) activeEditorPart : null;
				editor.formatText();
			
				//MessageDialog.openInformation(null, "MCMAS", "An agent has been added");
			}
		};
		formatText.setText("Format source");
		formatText.setToolTipText("Format whole ISPL source code");
		//formatText.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
		//		getImageDescriptor(IDE.SharedImages.IMG_OBJS_TASK_TSK));

		syntaxChecking = new Action() {
			public void run() {
				MCMASEditor editor = (activeEditorPart instanceof MCMASEditor) ? (MCMASEditor) activeEditorPart : null;
				editor.checkSyntax();
				editor.updateOutlinePage();
				MessageDialog.openInformation(null, "MCMAS", "Syntax Checking Executed");
			}
		};
		syntaxChecking.setText("Check syntax");
		syntaxChecking.setToolTipText("Check the syntax globally");
		//syntaxChecking.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
		//		getImageDescriptor(IDE.SharedImages.IMG_OBJS_TASK_TSK));
		
		interactiveMode = new Action() {
			public void run() {
				if(activeMultiPageEditor != null) {
					MCMASEditor ispleditor = activeMultiPageEditor.getMCMASEditor();
					boolean result = ispleditor.checkSyntax();
					ispleditor.updateOutlinePage();
					if(result) {
						activeMultiPageEditor.enterInteractiveMode();
						ispleditor.interactiveExecution();
					} else
						MessageDialog.openInformation(null, "MCMAS", 
								"There are syntex errors in the ISPL code. Please correct them before running interactive execution.");	
				}
			}
		};
		interactiveMode.setText("Explicit interactive mode");
		interactiveMode.setToolTipText("Execute the model interactively in an explicity way");
		//interactiveMode.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
		//		getImageDescriptor(IDE.SharedImages.IMG_OBJS_TASK_TSK));
		
		interactiveMode1 = new Action() {
			public void run() {
				if(activeMultiPageEditor != null) {
					MCMASEditor ispleditor = activeMultiPageEditor.getMCMASEditor();
					if(!ispleditor.saveFile())
						return;
					boolean result = ispleditor.checkSyntax();
					ispleditor.updateOutlinePage();
					if(result) {
						activeMultiPageEditor.enterInteractiveMode();
						ispleditor.interactiveExecution1();
					} else
						MessageDialog.openInformation(null, "MCMAS", 
								"There are syntex errors in the ISPL code. Please correct them before running interactive execution.");	
				}
			}
		};
		interactiveMode1.setText("Symbolic interactive mode");
		interactiveMode1.setToolTipText("Execute the model interactively in a symbolic way");
		
		modelChecking = new Action() {
			public void run() {
				if(activeMultiPageEditor != null) {
					MCMASEditor ispleditor = activeMultiPageEditor.getMCMASEditor();
					if(!ispleditor.saveFile())
						return;
					boolean result = ispleditor.checkSyntax();
					ispleditor.updateOutlinePage();
					if(result) {
						activeMultiPageEditor.enterModelChecking();
						activeMultiPageEditor.modelChecking();
					} else
						MessageDialog.openInformation(null, "MCMAS", 
								"There are syntex errors in the ISPL code. Please correct them before running interactive execution.");	
				}
			}
		};
		modelChecking.setText("Launch verification");
		modelChecking.setToolTipText("Verifying the model");
		//modelChecking.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
		//		getImageDescriptor(IDE.SharedImages.IMG_OBJS_TASK_TSK));
		
		/*exploration = new Action() {
			public void run() {
				MCMASEditor editor = (activeEditorPart instanceof MCMASEditor) ? (MCMASEditor) activeEditorPart : null;
				if(editor != null) {
					boolean result = editor.checkSyntax();
					if(result) {
						int[] n = editor.exploration(0);
						MessageDialog.openInformation(null, "MCMAS", 
								"Size of full state space = " + n[0] + ", Time = " + n[1]+"."+n[2]);
					}
				}
			}
		};
		exploration.setText("Exploration");
		exploration.setToolTipText("Generate full state space");

		exploration1 = new Action() {
			public void run() {
				MCMASEditor editor = (activeEditorPart instanceof MCMASEditor) ? (MCMASEditor) activeEditorPart : null;
				if(editor != null) {
					boolean result = editor.checkSyntax();
					if(result) {
						int[] n = editor.exploration(1);
						MessageDialog.openInformation(null, "MCMAS", 
								"Size of reduced state space = " + n[0] + ", Time = " + n[1]+"."+n[2]);
					}
				}
			}
		};
		exploration1.setText("Exploration1");
		exploration1.setToolTipText("Generate reduce state space");*/

		
	}
	public void contributeToMenu(IMenuManager manager) {
		IMenuManager menu = new MenuManager("&MCMAS");
		manager.prependToGroup(IWorkbenchActionConstants.MB_ADDITIONS, menu);
		menu.add(new Separator(IMcmasActionConstants.group1));
		menu.add(new Separator(IMcmasActionConstants.group2));
		
		menu.appendToGroup(IMcmasActionConstants.group1, addAgent);
		menu.appendToGroup(IMcmasActionConstants.group1, fContentAssistProposal);
		menu.appendToGroup(IMcmasActionConstants.group1, formatText);
		menu.appendToGroup(IMcmasActionConstants.group2, syntaxChecking);
		menu.appendToGroup(IMcmasActionConstants.group2, interactiveMode);
		menu.appendToGroup(IMcmasActionConstants.group2, interactiveMode1);
		menu.appendToGroup(IMcmasActionConstants.group2, modelChecking);
		
		//menu.add(addAgent);
		//menu.add(syntaxChecking);
		//menu.add(interactiveMode);
		//menu.add(modelChecking);
	}
	public void contributeToToolBar(IToolBarManager manager) {
		manager.add(new Separator(IMcmasActionConstants.group1));
		manager.add(new Separator(IMcmasActionConstants.group2));

		manager.appendToGroup(IMcmasActionConstants.group1, addAgent);
		manager.appendToGroup(IMcmasActionConstants.group1, fContentAssistProposal);
		manager.appendToGroup(IMcmasActionConstants.group1, formatText);
		manager.appendToGroup(IMcmasActionConstants.group2, syntaxChecking);
		manager.appendToGroup(IMcmasActionConstants.group2, interactiveMode);
		manager.appendToGroup(IMcmasActionConstants.group2, interactiveMode1);
		manager.appendToGroup(IMcmasActionConstants.group2, modelChecking);
		/*manager.appendToGroup(IMcmasActionConstants.group2, exploration);
		manager.appendToGroup(IMcmasActionConstants.group2, exploration1);*/

		//manager.add(syntaxChecking);
		//manager.add(interactiveMode);
		//manager.add(modelChecking);
	}
}
