package org.mcmas.ui.syntax;

import java.util.Hashtable;

import org.eclipse.jface.text.Position;

public class CESystem {
	private int index;
	private CEAgent[] agents;
	private int length;
	private Hashtable<String, CEAgent> agenttable;
	
	public CESystem(int index, CEAgent[] agents) {
		this.index = index;
		this.agents = agents;
		length = 1;
		agenttable = new Hashtable<String, CEAgent>();
		for(int i=0; i<agents.length; i++) {
			agenttable.put(agents[i].getName(), agents[i]);
			length += agents[i].numberOfLines();
		}
	}
	
	public String prefix() {
		return "-------- State " + index + " --------\n";
	}
	public String toString() {
		String str = prefix();
		for(int i=0; i<agents.length; i++)
			str += agents[i].toString();
		
		return str;
	}

	public String toString1() {
		String str = "----------------------------- State No. " + index + " -----------------------------\n";
		for(int i=0; i<agents.length; i++)
			str += agents[i].toString();
		
		return str;
	}
	public String toString2(int step) {
		String str = step==0 ? InterpretedSystem.initprefix :
			InterpretedSystem.stateprefix1 + step + InterpretedSystem.stateprefix2;
		for(int i=0; i<agents.length; i++)
			str += agents[i].toString();
		
		return str;
	}

	public int numberOfLines() {
		return length;
	}
	
	public int numberOfAgents() {
		return agents.length;
	}

	public CESystem project(String[] names) {
		CEAgent[] subset = new CEAgent[names.length];
		int count = 0;
		for(int i=0; i<names.length; i++) {
			subset[count] = agenttable.get(names[i]);
			count++;
		}
		return new CESystem(index, subset);
	}
	
	public Position[] getAgentPositions() {
		Position[] positions = new Position[agents.length];
		int count = prefix().length();
		for(int i=0; i<agents.length; i++) {
			positions[i] = new Position(count, 5);
			count += agents[i].toString().length();
		}
		
		return positions;
	}

	public Position[] getNamePositions() {
		Position[] positions = new Position[agents.length];
		int count = prefix().length();
		for(int i=0; i<agents.length; i++) {
			positions[i] = new Position(count+6, agents[i].getName().length());
			count += agents[i].toString().length();
		}
		
		return positions;
	}
	
	public String[] getAllAgents() {
		String[] names = new String[agents.length];
		for(int i=0; i<agents.length; i++)
			names[i] = agents[i].getName();
		return names;
	}
	
	public int getIndex() {
		return index;
	}
}
