package org.mcmas.ui.syntax;

import java.util.Hashtable;

import org.eclipse.jface.text.Position;

public class CounterExample {
	private CESystem[] states;
	private Position[] positions;
	private Position[] agentpositions;
	private Position[] namepositions;
	private int[] firstlines;
	private Hashtable<Integer, Position> indexedPosition;
	
	public CounterExample(CESystem[] states) {
		this.states = states;
		positions = new Position[states.length];
		int numOfAgents = states[0].numberOfAgents();
		agentpositions = new Position[states.length*numOfAgents];
		namepositions = new Position[states.length*numOfAgents];
		firstlines = new int[states.length];
		int start = 0;
		int line = 0;
		for(int i=0; i<states.length; i++) {
			String s = states[i].toString();
			positions[i] = new Position(start, s.length());
			firstlines[i] = line;
			Position[] aps = states[i].getAgentPositions();
			Position[] nps = states[i].getNamePositions();
			for(int j=0; j<numOfAgents; j++) {
				agentpositions[i*numOfAgents+j] = new Position(aps[j].offset+start, aps[j].length);
				namepositions[i*numOfAgents+j] = new Position(nps[j].offset+start, nps[j].length);
			}
			line += states[i].numberOfLines();
			start += s.length();
		}
	}
	
	public String toString() {
		String str = "";
		for(int i=0; i<states.length; i++)
			str += states[i].toString();
		return str;
	}
	
	public String toString1() {
		String str = "";
		for(int i=0; i<states.length; i++)
			str += states[i].toString1();
		return str;
	}
	
	public CounterExample project(String[] names) {
		CESystem[] subset = new CESystem[states.length];
		
		for(int i=0; i<states.length; i++)
			subset[i] = states[i].project(names);
		return new CounterExample(subset);
	}
	
	public Position getStatePosition(int index) {
		if(index >= 0 && index < positions.length) 
			return positions[index];
		else
			return null;
	}
	
	public Position getAgentPosition(int index) {
		if(index >= 0 && index < agentpositions.length) 
			return agentpositions[index];
		else
			return null;
	}
	
	public int numberOfAgents() {
		return agentpositions.length;
	}

	public Position getNamePosition(int index) {
		if(index >= 0 && index < namepositions.length) 
			return namepositions[index];
		else
			return null;
	}
	
	public int numberOfNames() {
		return namepositions.length;
	}

	public int getFirstLine(int index) {
		if(index >= 0 && index < firstlines.length) 
			return firstlines[index];
		else
			return -1;
	}
	
	public int numberOfStates() {
		return states.length;
	}
	
	public String[] getAllAgents() {
		return states[0].getAllAgents();
	}
	
	public Hashtable<Integer, Position> getStatePositions() {
		if(indexedPosition == null)
			indexedPosition = new Hashtable<Integer, Position>();
		for(int i=0; i<states.length; i++)
			indexedPosition.put(new Integer(states[i].getIndex()), positions[i]);
		return indexedPosition;
	}
	
	public String getState(int index, int step) {
		if(index >=1 && index <=states.length)
			return states[index-1].toString2(step);
		return "";
	}
}
