package org.mcmas.ui.syntax;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

public class GlobalState {
	public Hashtable<String, LocalState> localStates;
	
	public GlobalState(Hashtable<String, LocalState> localStates) {
		this.localStates = localStates;
	}
	
	public GlobalState() {
		this.localStates = new Hashtable<String, LocalState>();
	}
	
	public Hashtable<String, LocalState> getGlobalState() {
		return localStates;
	}
	
	public boolean addOneLocalVariable(String agentName, OneVariable var) {
		LocalState lstate = localStates.get(agentName);
		if(lstate == null) {
			lstate = new LocalState();
			localStates.put(agentName, lstate);
		}
		return lstate.addVariable(var);
	}
	
	public GlobalState Clone() {
		Hashtable<String, LocalState> states = new Hashtable<String, LocalState>();
		
		for(Enumeration<String> i = localStates.keys(); i.hasMoreElements();) {
			String agent = i.nextElement();
			LocalState lstate = localStates.get(agent);
			states.put(agent, lstate.Clone());
		}
		
		return new GlobalState(states);
	}
	
	public boolean isVariableAssigned(String name, String agentName) {
		LocalState lstate = localStates.get(agentName);
		if(lstate == null)
			return false;
		return lstate.isVariableAssigned(name);
	}
	
	public String toString() {
		String str = "";
		
		for(Enumeration<String> i=localStates.keys(); i.hasMoreElements();) {
			String agentName = i.nextElement();
			LocalState lstate = localStates.get(agentName);
			str += agentName + "\n" + lstate.toString();
		}
		
		return str;
	}

	public String toString(ArrayList<basic_agent> agents) {
		String str = "";
		
		for(int i=0; i<agents.size(); i++) {
			String agentName = agents.get(i).get_name();
			LocalState lstate = localStates.get(agentName);
			if(lstate != null)
				str += agentName + "\n" + lstate.toString();
		}
		
		return str;
	}

	public String toString(ArrayList<basic_agent> agents, ArrayList<ArrayList<String>> localvariables) {
		String str = "";
		
		for(int i=0; i<agents.size(); i++) {
			String agentName = agents.get(i).get_name();
			LocalState lstate = localStates.get(agentName);
			if(lstate != null)
				str += /*agentName + "\n" + */lstate.toString(localvariables.get(i));
		}
		
		return str;
	}
	
	public void toIntArray(ArrayList<basic_agent> agents, ArrayList<ArrayList<String>> localvariables, int[] source) {
		for(int i=0; i<agents.size(); i++) {
			String agentName = agents.get(i).get_name();
			LocalState lstate = localStates.get(agentName);
			if(lstate != null)
				lstate.toIntArray(source, localvariables.get(i));
		}
	}

	public expression getValue(String agentName, String varName) {
		LocalState lstate = localStates.get(agentName);
		if(lstate == null)
			return null;
		return lstate.getValue(varName);
	}
	
	public void replaceValue(String agentName, String varName, expression value) {
		LocalState lstate = localStates.get(agentName);
		if(lstate == null)
			return;
		lstate.replaceValue(varName, value);
	}
	
	public boolean equalTo(GlobalState gstate) {
		if(gstate != null && gstate.getGlobalState() != null && localStates != null) {
			Hashtable<String, LocalState> vars = gstate.getGlobalState();
			if(localStates.size() == vars.size()) {
				for(Enumeration<String> i=localStates.keys(); i.hasMoreElements();) {
					String vname = i.nextElement();
					LocalState value = localStates.get(vname);
					LocalState value1 = vars.get(vname);
					if(value1 == null || !value.equalTo(value1))
						return false;
				}
				return true;
			}
		}
		
		return false;
	}
	
}
