package org.mcmas.ui.syntax;

import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Stack;
import java.util.TreeSet;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.text.Position;

public class InterpretedSystem {
	public Hashtable<String, basic_agent> is_agents;
	public ArrayList<basic_agent> agents;
	public Hashtable<String, bool_expression> is_evaluation;
	public bool_expression is_istates;
	public Hashtable<String, TreeSet<groupitem>> is_groups;
	public ArrayList<modal_formula> is_formulae;
	public ArrayList<fairness_expression> is_fairness;
	private ArrayList<ArrayList<String>> localvariables;
	private ArrayList<GlobalState> iniStates;
	
	private ArrayList<GlobalState> interactiveStack;
	private ArrayList<Position> textStack;
	private int offset = 0;
	private ArrayList<GlobalState> currentStates;
	private ArrayList<Hashtable<String, String>> currentActions;
	private Hashtable<String, String> chosenAction;
	private boolean stateActionFlag;
	//Hashtable<String, Integer> options;

	final static public String initprefix = "----------------------------- Initial State -----------------------------\n";
	final static public String tranprefix = "----------------------------- Transition -----------------------------\n";
	final static public String stateprefix1 = "----------------------------- Step ";
	final static public String stateprefix2 = " -----------------------------\n";
	
	public InterpretedSystem(Hashtable<String, basic_agent> is_agents,
			ArrayList<basic_agent> agents,
			Hashtable<String, bool_expression> is_evaluation,
			bool_expression is_istates,
			Hashtable<String, TreeSet<groupitem>> is_groups,
			ArrayList<modal_formula> is_formulae,
			ArrayList<fairness_expression> is_fairness) {
		this.is_agents = is_agents;
		this.agents = agents;
		this.is_evaluation = is_evaluation;
		this.is_istates = is_istates;
		this.is_groups = is_groups;
		this.is_formulae = is_formulae;
		this.is_fairness = is_fairness;
		interactiveStack = new ArrayList<GlobalState>();
		textStack = new ArrayList<Position>();
		
		localvariables = new ArrayList<ArrayList<String>>();
		for(int i=0; i<agents.size(); i++) {
			basic_agent agent = agents.get(i);
			ArrayList<String> lvariables = new ArrayList<String>();
			if(agent.obsvars != null) 
				for(Enumeration<String> j=agent.obsvars.keys(); j.hasMoreElements(); ) 
					lvariables.add(j.nextElement());
			if(agent.vars != null) 
				for(Enumeration<String> j=agent.vars.keys(); j.hasMoreElements(); ) 
					lvariables.add(j.nextElement());
			localvariables.add(lvariables);
		}
	}
	
	public ArrayList<GlobalState> generateInitialGlobalStates() {
		ArrayList<GlobalState> globalStates = new ArrayList<GlobalState>();
		GlobalState gstate = new GlobalState();
		globalStates.add(gstate);
		
		bool_expression initialStates = is_istates.push_negation_inside();
		processBoolExpression(initialStates, globalStates);
		addMissingVariables(globalStates);
		
		currentStates = globalStates;
		stateActionFlag = true;
		
		return globalStates;
	}
	
	/**
	 * Parse the given boolean expression to generate the initial global states
	 * @param stateExpression: the boolean expression needs to be parsed
	 * @param globalStates: the set of initial global states generated
	 */
	private void processBoolExpression(bool_expression stateExpression, ArrayList<GlobalState> globalStates) {
		int op = stateExpression.get_op();
		if (op == 0) {// logic_expression
			addLogicExpressionToGlobalStates((logic_expression)stateExpression.get_operand(0), globalStates);
		} else if (op == 3) {// not operaror
			bool_expression expr = (bool_expression)stateExpression.get_operand(0);
			logic_expression lexpr = (logic_expression)expr.get_operand(0);
			ArrayList<logic_expression> lexprs = lexpr.getNegation();
			ArrayList<GlobalState> gstates = null;
			ArrayList<GlobalState> allStates = new ArrayList<GlobalState>();
			for(int j=lexprs.size()-1; j>=0; j--) {
				logic_expression exp = lexprs.get(j);
				if(j>0)
					gstates = duplicateGlobalStates(globalStates);
				else
					gstates = globalStates;
				addLogicExpressionToGlobalStates(exp, gstates);
				for(int i=0; i<gstates.size(); i++) {
					GlobalState gstate = gstates.get(i);
					boolean flag = false;
					for(int k=0; k<allStates.size(); k++) {
						GlobalState gstate1 = allStates.get(k);
						if(gstate.equalTo(gstate1)) {
							flag = true;
							break;
						}
					}
					if(!flag)
						allStates.add(gstate);
				}
				//allStates.addAll(gstates);
			}
			globalStates.clear();
			globalStates.addAll(allStates);
		}
		else if (op == 1) { // and
			bool_expression expr1 = (bool_expression)stateExpression.get_operand(0);
			bool_expression expr2 = (bool_expression)stateExpression.get_operand(1);
			processBoolExpression(expr1, globalStates);
			processBoolExpression(expr2, globalStates);
		} else { // or
			bool_expression expr1 = (bool_expression)stateExpression.get_operand(0);
			bool_expression expr2 = (bool_expression)stateExpression.get_operand(1);
			ArrayList<GlobalState> newGlobalStates = duplicateGlobalStates(globalStates);
			processBoolExpression(expr1, globalStates);
			processBoolExpression(expr2, newGlobalStates);
			for(int i=0; i<newGlobalStates.size(); i++) {
				GlobalState gstate = newGlobalStates.get(i);
				boolean flag = false;
				for(int j=0; j<globalStates.size(); j++) {
					GlobalState gstate1 = globalStates.get(j);
					if(gstate.equalTo(gstate1)) {
						flag = true;
						break;
					}
				}
				if(!flag)
					globalStates.add(gstate);
			}
			//globalStates.addAll(newGlobalStates);
		}
	}
	
	private ArrayList<GlobalState> duplicateGlobalStates(ArrayList<GlobalState> globalStates) {
		ArrayList<GlobalState> newGlobalStates = new ArrayList<GlobalState>();
		for(int i=0; i<globalStates.size(); i++) {
			GlobalState gstate = globalStates.get(i);
			newGlobalStates.add(gstate.Clone());
		}
		return newGlobalStates;
	}
	
	/**
	 * Add the assigned value to the variable specified in the logic expression, which is 
	 * in fact an assignment, to all global states.
	 * @param expr
	 * @param globalStates
	 */
	private void addLogicExpressionToGlobalStates(logic_expression expr, ArrayList<GlobalState> globalStates) {
		int i = 0; 
		while(i<globalStates.size()) {
			GlobalState gstate = globalStates.get(i);
			boolean result = processLogicExpression(expr, gstate);
			if(result)
				i++;
			else {
				globalStates.remove(i);
			}
		}
	}
	
	/**
	 * Create a state variable combined with its value assigned in a logic expression. 
	 * Then insert it into the local state of corresponding agent.
	 * @param lexpression
	 * @return true if the variable is added successfully, false otherwise
	 */
	private boolean processLogicExpression(logic_expression lexpression, GlobalState gstate) {
		// get the variable that is assigned a value in this logic expression
		variable var = (variable)lexpression.get_operand(0);
		
		// generate a state variable for the basictype
		OneVariable statevar = new OneVariable(var.get_var_type(), lexpression.get_operand(1));
		
		// get the agent name
		String agentName = var.get_agent_name();
		
		return gstate.addOneLocalVariable(agentName, statevar);
	}
	
	/**
	 * Merge the set states2 of states into the set states1 of states. If a state already exists in the other set, do not insert it. 
	 * @param states1 
	 * @param states2 
	 */
	private void mergeTwoSetsOfStates(ArrayList<GlobalState> states1, ArrayList<GlobalState> states2) {
		for(int i=0; i<states2.size(); i++) {
			GlobalState g2 = states2.get(i);
			boolean found = false;
			for(int j=0; j<states1.size(); j++) {
				GlobalState g1 = states1.get(j);
				if(g1.equalTo(g2)) {
					found = true;
					break;
				}
			}
			if(!found)
				states1.add(g2);
		}
	}
	
	private void addMissingVariables(ArrayList<GlobalState> globalStates) {
		ArrayList<GlobalState> newstates = new ArrayList<GlobalState>();
		for(int i=0; i<agents.size(); i++) {
			basic_agent agent = agents.get(i);
			Hashtable<String, basictype> obsvars = agent.get_obsvars();
			if(obsvars != null && obsvars.size() > 0) {
				for(Enumeration<String> j=obsvars.keys(); j.hasMoreElements();) {
					basictype bt = obsvars.get(j.nextElement());
					int type = bt.get_type();
					ArrayList<logic_expression> lexprs;
					if(type == 1)
						lexprs = bt.getLogicExpressions(agent.get_name());
					else if(type == 2)
						lexprs = ((rangedint)bt).getLogicExpressions(agent.get_name());
					else
						lexprs = ((enumerate)bt).getLogicExpressions(agent.get_name());
					newstates.clear();
					for(int k=0; k<globalStates.size(); k++) {
						ArrayList<GlobalState> tmpstates = addOneMissingVariable(bt.get_name(), agent.get_name(), lexprs, globalStates.get(k));
						//newstates.addAll(tmpstates);
						mergeTwoSetsOfStates(newstates, tmpstates);
					}
					globalStates.clear();
					globalStates.addAll(newstates);
				}
			}
			
			Hashtable<String, basictype> vars = agent.get_vars();
			if(vars != null && vars.size() > 0) {
				for(Enumeration<String> j=vars.keys(); j.hasMoreElements();) {
					basictype bt = vars.get(j.nextElement());
					int type = bt.get_type();
					ArrayList<logic_expression> lexprs;
					if(type == 1)
						lexprs = bt.getLogicExpressions(agent.get_name());
					else if(type == 2)
						lexprs = ((rangedint)bt).getLogicExpressions(agent.get_name());
					else
						lexprs = ((enumerate)bt).getLogicExpressions(agent.get_name());
					newstates.clear();
					for(int k=0; k<globalStates.size(); k++) {
						ArrayList<GlobalState> tmpstates = addOneMissingVariable(bt.get_name(), agent.get_name(), lexprs, globalStates.get(k));
						//newstates.addAll(tmpstates);
						mergeTwoSetsOfStates(newstates, tmpstates);
					}
					globalStates.clear();
					globalStates.addAll(newstates);
				}
			}
		}
	}
	
	private ArrayList<GlobalState> addOneMissingVariable(String name, String agentName, ArrayList<logic_expression> lexprs, GlobalState globalState) {
		ArrayList<GlobalState> allStates = new ArrayList<GlobalState>();
		
		if(!globalState.isVariableAssigned(name, agentName)) {
			for(int j=lexprs.size()-1; j>=0; j--) {
				ArrayList<GlobalState> gstates = new ArrayList<GlobalState>();
				gstates.add(globalState.Clone());
				logic_expression exp = lexprs.get(j);
				addLogicExpressionToGlobalStates(exp, gstates);
				allStates.addAll(gstates);
			}
		} else
			allStates.add(globalState);
		
		return allStates;
	}

	public boolean getEnabledActions(GlobalState gstate, Hashtable<String, TreeSet<String>> enabledActions) {
		for(int i=0; i<agents.size(); i++) {
			basic_agent agent = agents.get(i);
			TreeSet<String> actions = agent.get_actions();
			TreeSet<String> enabled = new TreeSet<String>();
			if((actions == null || actions.size() == 0)) {
				if(agent.get_name().compareTo("Environment") != 0) {
					enabledActions.clear();
					return false;
				} else
					continue;
			}
			ArrayList<protocol_line> protocol = agent.get_protocol();
			if((protocol == null || protocol.size() == 0)) {
				if(agent.get_name().compareTo("Environment") != 0) {
					enabledActions.clear();
					return false;
				} else
					continue;
			}
			bool_expression disjunction = null; 
			for(int j=0; j<protocol.size(); j++) {
				protocol_line line = protocol.get(j);
				bool_expression cond = line.get_condition();
				if(cond.get_op() != 4) {
					boolean isEnabled = cond.evaluate(agent.get_name(), gstate, null);
					if(j==0)
						disjunction = cond;
					else
						disjunction = new bool_expression(2, disjunction, cond, 0, 0, 0, 0);
					if(isEnabled) 
						addEnabledActions(enabled, line);
				} else { // handle "Other" branch
					if(disjunction != null) {
						if(enabled.size()==0) {
							bool_expression other = new bool_expression(3, disjunction, null, 0, 0, 0, 0);
							boolean isEnabled = other.evaluate(agent.get_name(), gstate, null);
							if(isEnabled) 
								addEnabledActions(enabled, line);
						}
					} else {
						Hashtable<String, basictype> obsvars = agent.get_obsvars();
						Hashtable<String, basictype> vars = agent.get_vars();
						if(!(obsvars == null || obsvars.size() == 0) || !(vars == null || vars.size() == 0)) 
							addEnabledActions(enabled, line);
						else {
							enabledActions.clear();
							return false;
						}
					}
				}
			}
			if(enabled.size() > 0)
				enabledActions.put(agent.get_name(), enabled);
			else {
				enabledActions.clear();
				return false;
			}
		}
		
		return true;
	}
	
	public boolean getInterleavedEnabledActions(GlobalState gstate, Hashtable<String, TreeSet<String>> enabledActions) {
		for(int i=0; i<agents.size(); i++) {
			basic_agent agent = agents.get(i);
			TreeSet<String> actions = agent.get_actions();
			TreeSet<String> enabled = new TreeSet<String>();
			if((actions == null || actions.size() == 0)) {
				if(agent.get_name().compareTo("Environment") != 0) {
					enabledActions.clear();
					return false;
				} else
					continue;
			}
			ArrayList<protocol_line> protocol = agent.get_protocol();
			if((protocol == null || protocol.size() == 0)) {
				if(agent.get_name().compareTo("Environment") != 0) {
					enabledActions.clear();
					return false;
				} else
					continue;
			}
			bool_expression disjunction = null; 
			for(int j=0; j<protocol.size(); j++) {
				protocol_line line = protocol.get(j);
				bool_expression cond = line.get_condition();
				if(cond.get_op() != 4) {
					boolean isEnabled = cond.evaluate(agent.get_name(), gstate, null);
					if(j==0)
						disjunction = cond;
					else
						disjunction = new bool_expression(2, disjunction, cond, 0, 0, 0, 0);
					if(isEnabled) 
						addEnabledActions(enabled, line);
				} else { // handle "Other" branch
					if(disjunction != null) {
						if(enabled.size()==0) {
							bool_expression other = new bool_expression(3, disjunction, null, 0, 0, 0, 0);
							boolean isEnabled = other.evaluate(agent.get_name(), gstate, null);
							if(isEnabled) 
								addEnabledActions(enabled, line);
						}
					} else {
						Hashtable<String, basictype> obsvars = agent.get_obsvars();
						Hashtable<String, basictype> vars = agent.get_vars();
						if(!(obsvars == null || obsvars.size() == 0) || !(vars == null || vars.size() == 0)) 
							addEnabledActions(enabled, line);
						else {
							enabledActions.clear();
							return false;
						}
					}
				}
			}
			if(enabled.size() > 0)
				enabledActions.put(agent.get_name(), enabled);
		}
		
		return true;
	}

	private void addEnabledActions(TreeSet<String> enabled, protocol_line line) {
		TreeSet<groupitem> acts = line.get_actions();
		for(Iterator<groupitem> k=acts.iterator(); k.hasNext();) {
			groupitem t = k.next();
			enabled.add(t.name);
		}
	}
	
	public ArrayList<Hashtable<String, String>> getPossibleTransitions(Hashtable<String, TreeSet<String>> enabledActions) {
		ArrayList<Hashtable<String, String>> trans = new ArrayList<Hashtable<String, String>>();
		
		for(int i=0; i<agents.size(); i++) {
			basic_agent agent = agents.get(i);
			TreeSet<String> enabled = enabledActions.get(agent.get_name());
			if(enabled != null && enabled.size() > 0)
				addOneAction(agent.get_name(), enabled, trans);
		}
		
		return trans;
	}
	
	private void addOneAction(String agentName, TreeSet<String> enabled, ArrayList<Hashtable<String, String>> trans) {
		ArrayList<Hashtable<String, String>> newtrans = new ArrayList<Hashtable<String, String>>();
		for(Iterator<String> i=enabled.iterator(); i.hasNext();) {
			String act = i.next();
			if(trans.size() == 0) {
				Hashtable<String, String> newelement = new Hashtable<String, String>();
				newelement.put(agentName, act);
				newtrans.add(newelement);
			} else 
				for(int j=0; j<trans.size(); j++) {
					Hashtable<String, String> element = trans.get(j);
					Hashtable<String, String> newelement = duplicateActions(element);
					newelement.put(agentName, act);
					newtrans.add(newelement);
				}
		}
		trans.clear();
		trans.addAll(newtrans);
	}
	
	private Hashtable<String, String> duplicateActions(Hashtable<String, String> element) {
		if(element == null)
			return null;
		
		Hashtable<String, String> newelement = new Hashtable<String, String>();
		for(Enumeration<String> i=element.keys(); i.hasMoreElements();) {
			String agent = i.nextElement();
			String action = element.get(agent);
			newelement.put(agent, action);
		}
		return newelement;
	}
	
	public ArrayList<GlobalState> getSuccessors(GlobalState gstate, Hashtable<String, String> actions) {
		ArrayList<GlobalState> successors = new ArrayList<GlobalState>();
		successors.add(gstate.Clone());
		
		for(int i=0; i<agents.size(); i++) {
			basic_agent agent = agents.get(i);
			ArrayList<evolution_line> evolution = agent.get_evolution();
			if(evolution != null) {
				ArrayList<GlobalState> allsucc = new ArrayList<GlobalState>();
				for(int j=0; j<evolution.size(); j++) {
					evolution_line line = evolution.get(j);
					bool_expression cond = line.get_condition();
					if(cond.evaluate(agent.get_name(), gstate, actions)) {
						ArrayList<GlobalState> tmpsucc = duplicateGlobalStates(successors);
						ArrayList<GlobalState> tmpsucc1 = new ArrayList<GlobalState>();
						ArrayList<assignment> assignments = line.get_assignments();
						for(int t=0; t<tmpsucc.size(); t++) {
							GlobalState tmpstate = tmpsucc.get(t);
							boolean result = executeAssignment(gstate, tmpstate, agent.get_name(), assignments);
							if(result)
								tmpsucc1.add(tmpstate);
						}
						//allsucc.addAll(tmpsucc1);
						mergeTwoSetsOfStates(allsucc, tmpsucc1);
					}
				}
				if(allsucc.size() > 0) {
					successors.clear();
					successors.addAll(allsucc);
				}
			}
		}
		
		currentStates = successors;
		return successors;
	}
	
	private boolean executeAssignment(GlobalState sourcestate, GlobalState gstate, 
			String agentName, ArrayList<assignment> assignments) {
		if(assignments != null)
			for(int t=0; t<assignments.size(); t++) {
				assignment assign = assignments.get(t);
				variable var = assign.get_var();
				String varName = var.get_variable_name();
				String agentName1 = var.get_agent_name();
				if(agentName1 == null || Util.isEmpty(agentName1))
					agentName1 = agentName;
		
				expression value = assign.get_var_value();
				int valuetype = value.get_type();
				if(valuetype == 0) { // a variable
					variable var1 = (variable)value;
					String varName1 = var1.get_variable_name();
					String agentName2 = var1.get_agent_name();
					if(agentName2 == null || Util.isEmpty(agentName2))
						agentName2 = agentName;
					expression tmpvalue = sourcestate.getValue(agentName2, varName1);
					gstate.replaceValue(agentName1, varName, tmpvalue);
				} else if(valuetype == 1) { // a boolean value
					gstate.replaceValue(agentName1, varName, value);
				} else if(valuetype == 2) { // an integer value
					int v = ((int_value)value).get_value();
					rangedint bt = (rangedint)var.get_var_type();
					if(bt.get_lowerbound() <= v && v <= bt.get_upperbound())
						gstate.replaceValue(agentName1, varName, value);
					else {
						/*MessageDialog.openError(null, "Int value out of bounds", 
								v + "is beyond the range [" + bt.get_lowerbound() + " .. " + 
								bt.get_upperbound() + "] of the variable" + agentName1 + "." +
								varName);*/
						System.out.println("Int value out of bounds: " +
								v + " is beyond the range [" + bt.get_lowerbound() + " .. " + 
								bt.get_upperbound() + "] of the variable " + agentName1 + "." +
								varName);
						return false;
					}
				} else if(valuetype == 3) { // an enumeration value
					gstate.replaceValue(agentName1, varName, value);
				} else { // an arithmetic expression
					int v = ((arithmetic_expression)value).evaluate(agentName1, sourcestate);
					rangedint bt = (rangedint)var.get_var_type();
					if(bt.get_lowerbound() <= v && v <= bt.get_upperbound()) {
						int_value tmpvalue = new int_value(v, 0, 0, 0, 0);
						gstate.replaceValue(agentName1, varName, tmpvalue);
					}
					else {
						/*MessageDialog.openError(null, "Int value out of bounds", 
								value.toString() + " = " + v + "is beyond the range [" + bt.get_lowerbound() + " .. " + 
								bt.get_upperbound() + "] of the variable" + agentName1 + "." +
								varName);*/
						System.out.println("Int value out of bounds: "+ 
								value.toString() + " = " + v + " is beyond the range [" + bt.get_lowerbound() + " .. " + 
								bt.get_upperbound() + "] of the variable " + agentName1 + "." +
								varName);
						return false;
					}
				}
			}
		
		return true;
	}
	
	public void addGlobalState(GlobalState gstates) {
		interactiveStack.add(gstates);
		int sp = interactiveStack.size();
		String s = "";
		if(sp == 1) 
			s = initprefix + gstates.toString(agents);
		else
			s = stateprefix1 + (sp - 1) + stateprefix2 + gstates.toString(agents);
		Position p = new Position(offset, s.length());
		textStack.add(p);
		offset += s.length();
	}
	
	public GlobalState getTopGlobalState() {
		if(interactiveStack.size()>0) {
			return interactiveStack.get(interactiveStack.size()-1);
		} else
			return null;
	}
	
	public Position getTopTextState() {
		if(textStack.size()>0) {
			return textStack.get(textStack.size()-1);
		} else
			return null;
	}
	
	public int stackSize() {
		return interactiveStack.size();
	}
	
	public GlobalState popupGlobalState() {
		if(interactiveStack.size()>0) {
			GlobalState gstates = interactiveStack.get(interactiveStack.size()-1);
			interactiveStack.remove(interactiveStack.size()-1);
			textStack.remove(textStack.size()-1);
			Position p = getTopTextState();
			offset = p == null ? 0 : p.offset + p.length;
			return gstates;
		} else
			return null;
	}
	
	public void quitInteractiveMode() {
		interactiveStack.clear();
		textStack.clear();
		offset = 0;
		currentStates = null;
		currentActions = null;
		chosenAction = null;
	}
	
	public boolean isStackEmpty() {
		return interactiveStack.size() == 0;
	}
	
	public String displayCandidateStates() {
		if(currentStates != null) {
			String strStates = "";
			for(int i=0; i<currentStates.size(); i++) {
				GlobalState gstate = currentStates.get(i);
				strStates += "----------------------------- State No. " + (i+1) + " -----------------------------\n";
				strStates += gstate.toString(agents);
			}
			return strStates;
		}
		return "";
	}
	
	public boolean chooseState(int number) {
		int stateCount = currentStates.size();
		if(number > 0 && number <=stateCount) {
			GlobalState gstate = currentStates.get(number-1);
			addGlobalState(gstate);
			chosenAction = null;
			return true;
		}
		return false;
	}
	
	public ArrayList<Hashtable<String, String>> generateEnabledActions() {
		Hashtable<String, TreeSet<String>> enabledActions = new Hashtable<String, TreeSet<String>>();
		getEnabledActions(getTopGlobalState(), enabledActions);
		currentActions = getPossibleTransitions(enabledActions);
		return currentActions;
	}
	
	public String displayCurrentActions() {
		if(currentActions != null) {
			String strActions = "";
			for(int i=0; i<currentActions.size(); i++) {
				Hashtable<String, String> gtran = currentActions.get(i);
				strActions += "----------------------------- Transition No. " + (i+1) + " -----------------------------\n";
				for(int j=0; j<agents.size(); j++) {
					String agentName = agents.get(j).get_name();
					String act = gtran.get(agentName);
					if(act != null) 
						strActions += agentName + ": " + act + "\n";
				}
			}
			return strActions;
		}
		return "";
	}

	public boolean chooseAction(int number) {
		int actionCount = currentActions.size();
		if(number > 0 && number <=actionCount) {
			chosenAction = currentActions.get(number-1);
			return true;
		}
		return false;
	}
	
	public ArrayList<GlobalState> generateSuccessorStates() {
		if(isStackEmpty()) 
			return generateInitialGlobalStates();
		else
			return getSuccessors(getTopGlobalState(), chosenAction);
	}
	
	public void backtrack() {
		if(chosenAction == null)
			popupGlobalState();
		else {
			chosenAction = null;
			Position p = getTopTextState();
			offset = p.offset+p.length;
		}
	}
	
	public String displayCurrentState() {
		GlobalState gstate = getTopGlobalState();
		String strState = "";
		if(stackSize() > 0) {
			if(stackSize() == 1)
				strState += initprefix;
			else 
				strState += stateprefix1 + (stackSize()-1) + stateprefix2;
			strState += gstate.toString(agents);
		}
		return strState;
	}
	
	public String displayChosenAction() {
		String strAction = "";
		if(chosenAction != null) {
			int i = 0;
			strAction += tranprefix;
			for(int j=0; j<agents.size(); j++) {
				String agentName = agents.get(j).get_name();
				String act = chosenAction.get(agentName);
				if(act != null) {
					if(i > 0)
						strAction += "\n";
					strAction += agentName + ": " + act;
					i++;
				}
			}
			strAction += "\n";
		}
		offset += strAction.length();
		return strAction;
	}
	
	public void storeIniStates() {
		if(iniStates == null || iniStates.size() == 0)
			iniStates = generateInitialGlobalStates();
	}
	
	private void clearIntArray(int[] source) {
		for(int i=0; i<source.length; i++)
			source[i] = 0;
	}
	
	private String getKey(int[] source) {
		String key = "";
		for(int k=source.length-1; k>=0; k--) 
			key += (k<source.length-1 ? "." : "") + Integer.toHexString(source[k]);
		return key;
	}
	
	public int[] exploration_int() {
		int bddlength = BDDlength();
		int statesize = (bddlength%32)==0 ? bddlength/32 : bddlength/32+1;
		int[] source = new int[statesize]; 
		
		long start = System.currentTimeMillis();
		HashSet<String> stateSpace = new HashSet<String>();
		//ArrayList<GlobalState> iniStates = generateInitialGlobalStates();
		Stack<GlobalState> searchStack = new Stack<GlobalState>();
		for(int i=0; i<iniStates.size(); i++) {
			GlobalState g = iniStates.get(i);
			//String key = g.toString(agents, localvariables);
			clearIntArray(source);
			g.toIntArray(agents, localvariables, source);
			String key = getKey(source);
			
			boolean found = stateSpace.contains(key);
			if(!found) {
				stateSpace.add(key);
				searchStack.push(g);
				
				visitState(searchStack, stateSpace, source);
			}
		}
		long end = System.currentTimeMillis();
		
		int[] results = new int[3];
		results[0] = stateSpace.size();
		results[1] = (int)((end-start)/1000);
		results[2] = (int)((end-start) % 1000);
		System.gc();
		return results;
	}
	
	public int[] exploration() {
		int bddlength = BDDlength();
		int statesize = (bddlength%32)==0 ? bddlength/32 : bddlength/32+1;
		
		long start = System.currentTimeMillis();
		HashSet<HashKey> stateSpace = new HashSet<HashKey>();
		//ArrayList<GlobalState> iniStates = generateInitialGlobalStates();
		Stack<GlobalState> searchStack = new Stack<GlobalState>();
		for(int i=0; i<iniStates.size(); i++) {
			GlobalState g = iniStates.get(i);
			//String key = g.toString(agents, localvariables);
			HashKey source = new HashKey(statesize);
			g.toIntArray(agents, localvariables, source.key);
			
			boolean found = stateSpace.contains(source);
			if(!found) {
				stateSpace.add(source);
				searchStack.push(g);
				
				visitState(searchStack, stateSpace, statesize);
			}
		}
		long end = System.currentTimeMillis();
		
		int[] results = new int[3];
		results[0] = stateSpace.size();
		results[1] = (int)((end-start)/1000);
		results[2] = (int)((end-start) % 1000);
		System.gc();
		return results;
	}
	
	
	public int[] por_exploration_int() {
		int bddlength = BDDlength();
		int statesize = (bddlength%32)==0 ? bddlength/32 : bddlength/32+1;
		int[] source = new int[statesize]; 

		Hashtable<String, bool_expression> tmp_evaluation = new Hashtable<String, bool_expression>(is_evaluation);
		for(Enumeration<String> i = is_evaluation.keys(); i.hasMoreElements(); ) {
			String key = i.nextElement();
			if(key.indexOf("GreenState") != -1 || key.indexOf("RedState") != -1)
				is_evaluation.remove(key);
		}
		long start = System.currentTimeMillis();
		HashSet<String> stateSpace = new HashSet<String>();
		//ArrayList<GlobalState> iniStates = generateInitialGlobalStates();
		Stack<ExtendedGlobalState> searchStack = new Stack<ExtendedGlobalState>();
		ArrayList<String> agnames = new ArrayList<String>();
		is_formulae.get(0).getAgents(agnames, is_agents, is_groups);
		
		for(int i=0; i<iniStates.size(); i++) {
			ExtendedGlobalState g = new ExtendedGlobalState(iniStates.get(i), is_evaluation);
			//String key = g.toString(agents, localvariables);
			clearIntArray(source);
			g.toIntArray(agents, localvariables, source);
			String key = getKey(source);

			boolean found = stateSpace.contains(key);
			if(!found) {
				stateSpace.add(key);
				searchStack.push(g);
				
				por_visitState(searchStack, stateSpace, agnames, source);
			}
		}
		long end = System.currentTimeMillis();
		
		int[] results = new int[3];
		results[0] = stateSpace.size();
		results[1] = (int)((end-start)/1000);
		results[2] = (int)((end-start) % 1000);
		System.gc();
		is_evaluation = tmp_evaluation;
		return results;
	}
	
	public int[] por_exploration() {
		int bddlength = BDDlength();
		int statesize = (bddlength%32)==0 ? bddlength/32 : bddlength/32+1;

		Hashtable<String, bool_expression> tmp_evaluation = new Hashtable<String, bool_expression>(is_evaluation);
		for(Enumeration<String> i = is_evaluation.keys(); i.hasMoreElements(); ) {
			String key = i.nextElement();
			if(key.indexOf("GreenState") != -1 || key.indexOf("RedState") != -1)
				is_evaluation.remove(key);
		}
		long start = System.currentTimeMillis();
		HashSet<HashKey> stateSpace = new HashSet<HashKey>();
		//ArrayList<GlobalState> iniStates = generateInitialGlobalStates();
		Stack<ExtendedGlobalState> searchStack = new Stack<ExtendedGlobalState>();
		ArrayList<String> agnames = new ArrayList<String>();
		is_formulae.get(0).getAgents(agnames, is_agents, is_groups);
		
		for(int i=0; i<iniStates.size(); i++) {
			ExtendedGlobalState g = new ExtendedGlobalState(iniStates.get(i), is_evaluation);
			//String key = g.toString(agents, localvariables);
			HashKey source = new HashKey(statesize);
			g.toIntArray(agents, localvariables, source.key);

			boolean found = stateSpace.contains(source);
			if(!found) {
				stateSpace.add(source);
				searchStack.push(g);
				
				por_visitState(searchStack, stateSpace, agnames, statesize);
			}
		}
		long end = System.currentTimeMillis();
		
		int[] results = new int[3];
		results[0] = stateSpace.size();
		results[1] = (int)((end-start)/1000);
		results[2] = (int)((end-start) % 1000);
		System.gc();
		is_evaluation = tmp_evaluation;
		return results;
	}
	
	private void visitState(Stack<GlobalState> searchStack, HashSet<String> stateSpace, int[] source) {
		if(searchStack.isEmpty())
			return;
		
		GlobalState g = searchStack.peek();
		//get enabled transition
		Hashtable<String, TreeSet<String>> enabledActions = new Hashtable<String, TreeSet<String>>();
		getInterleavedEnabledActions(g, enabledActions);
		ArrayList<GlobalState> successors = getInterleavedSuccessors(g, enabledActions);
		while(successors.size() > 0) {
			GlobalState newg = successors.remove(0);
			//String key = newg.toString(agents, localvariables);
			clearIntArray(source);
			newg.toIntArray(agents, localvariables, source);
			String key = getKey(source);

			boolean found = stateSpace.contains(key);
			if(!found) {
				stateSpace.add(key);
				searchStack.push(newg);
				//if(stateSpace.size() % 10000 == 0)
				//	System.gc();
				visitState(searchStack, stateSpace, source);
			} 
		}
		searchStack.pop();
	}
	
	private void visitState(Stack<GlobalState> searchStack, HashSet<HashKey> stateSpace, int statesize) {
		if(searchStack.isEmpty())
			return;
		
		GlobalState g = searchStack.peek();
		//get enabled transition
		Hashtable<String, TreeSet<String>> enabledActions = new Hashtable<String, TreeSet<String>>();
		getInterleavedEnabledActions(g, enabledActions);
		ArrayList<GlobalState> successors = getInterleavedSuccessors(g, enabledActions);
		while(successors.size() > 0) {
			GlobalState newg = successors.remove(0);
			//String key = newg.toString(agents, localvariables);
			HashKey source = new HashKey(statesize);
			newg.toIntArray(agents, localvariables, source.key);

			boolean found = stateSpace.contains(source);
			if(!found) {
				stateSpace.add(source);
				searchStack.push(newg);
				//if(stateSpace.size() % 10000 == 0)
				//	System.gc();
				visitState(searchStack, stateSpace, statesize);
			} 
		}
		searchStack.pop();
	}
	
	private void por_visitState(Stack<ExtendedGlobalState> searchStack, HashSet<String> stateSpace, 
			ArrayList<String> agnames, int[] source) {
		if(searchStack.isEmpty())
			return;
		
		ExtendedGlobalState g = searchStack.peek();
		//get enabled transition
		Hashtable<String, TreeSet<String>> enabledActions = new Hashtable<String, TreeSet<String>>();
		getInterleavedEnabledActions(g, enabledActions);
		ArrayList<GlobalState> tmpsuccessors = por_getInterleavedSuccessors(g, enabledActions);
		for(int i=0; i<tmpsuccessors.size(); i++) {
			ExtendedGlobalState newg = (ExtendedGlobalState)tmpsuccessors.get(i);
			newg.setVisible(g);
			newg.setChangedJ(g, agnames);
		}
		
		ArrayList<GlobalState> successors = new ArrayList<GlobalState>();
		for(int i=0; i<tmpsuccessors.size(); i++) {
			ExtendedGlobalState newg = (ExtendedGlobalState)tmpsuccessors.get(i);
			if(!newg.isChangedJ && !newg.isSynchronized && !newg.isVisible) {
				successors.add(newg);
				break;
			}
		} 
		if(successors.size() == 0)
			successors.addAll(tmpsuccessors);
		
		while(successors.size() > 0) {
			ExtendedGlobalState newg = (ExtendedGlobalState)successors.remove(0);
			//String key = newg.toString(agents, localvariables);
			clearIntArray(source);
			newg.toIntArray(agents, localvariables, source);
			String key = getKey(source);

			boolean found = stateSpace.contains(key);
			if(!found) {
				stateSpace.add(key);
				searchStack.push(newg);
								
				por_visitState(searchStack, stateSpace, agnames, source);
			}
		}
		searchStack.pop();
	}
	
	private void por_visitState(Stack<ExtendedGlobalState> searchStack, HashSet<HashKey> stateSpace, 
			ArrayList<String> agnames, int statesize) {
		if(searchStack.isEmpty())
			return;
		
		ExtendedGlobalState g = searchStack.peek();
		//get enabled transition
		Hashtable<String, TreeSet<String>> enabledActions = new Hashtable<String, TreeSet<String>>();
		getInterleavedEnabledActions(g, enabledActions);
		ArrayList<GlobalState> tmpsuccessors = por_getInterleavedSuccessors(g, enabledActions);
		for(int i=0; i<tmpsuccessors.size(); i++) {
			ExtendedGlobalState newg = (ExtendedGlobalState)tmpsuccessors.get(i);
			newg.setVisible(g);
			newg.setChangedJ(g, agnames);
		}
		
		ArrayList<GlobalState> successors = new ArrayList<GlobalState>();
		for(int i=0; i<tmpsuccessors.size(); i++) {
			ExtendedGlobalState newg = (ExtendedGlobalState)tmpsuccessors.get(i);
			if(!newg.isChangedJ && !newg.isSynchronized && !newg.isVisible) {
				successors.add(newg);
				break;
			}
		} 
		if(successors.size() == 0)
			successors.addAll(tmpsuccessors);
		
		while(successors.size() > 0) {
			ExtendedGlobalState newg = (ExtendedGlobalState)successors.remove(0);
			//String key = newg.toString(agents, localvariables);
			HashKey source = new HashKey(statesize);
			newg.toIntArray(agents, localvariables, source.key);

			boolean found = stateSpace.contains(source);
			if(!found) {
				stateSpace.add(source);
				searchStack.push(newg);
								
				por_visitState(searchStack, stateSpace, agnames, statesize);
			}
		}
		searchStack.pop();
	}
	
	public ArrayList<GlobalState> getInterleavedSuccessors(GlobalState gstate, 
			Hashtable<String, TreeSet<String>> actions) {
		ArrayList<GlobalState> successors = new ArrayList<GlobalState>();
		Hashtable<String, ArrayList<evolution_line>> enabledEVLine = new Hashtable<String, ArrayList<evolution_line>>();
		ArrayList<String> agnames = new ArrayList<String>();
		
		for(int i=0; i<agents.size(); i++) {
			basic_agent agent = agents.get(i);
			ArrayList<evolution_line> evolution = agent.get_evolution();
			if(evolution != null) {
				ArrayList<evolution_line> enabled_evolution = new ArrayList<evolution_line>();
				for(int j=0; j<evolution.size(); j++) {
					evolution_line line = evolution.get(j);
					bool_expression cond = line.get_condition();
					if(cond.evaluate1(agent.get_name(), gstate, actions)) {
						enabled_evolution.add(line);
					}
				}
				enabledEVLine.put(agent.get_name(), enabled_evolution);
				agnames.add(agent.get_name());
			}
		}
		
		for(int i=0; i<agnames.size(); i++) {
			String agname = agnames.get(i);
			ArrayList<evolution_line> enabled_evolution = enabledEVLine.get(agname);
			
			for(int j=0; j<enabled_evolution.size(); j++) {
				evolution_line line = enabled_evolution.get(j);
				ArrayList<logic_expression> synActions1 = new ArrayList<logic_expression>();			
				line.get_condition().getSynchronizedActions(agname, synActions1);
				if(synActions1.size() == 2) {
					logic_expression act11 = synActions1.get(0);
					logic_expression act12 = synActions1.get(1);
					String a1 = ((laction)act11.get_operand(0)).get_agent_name();
					String a2 = ((laction)act12.get_operand(0)).get_agent_name();
					String agname1 = (a1.compareTo(agname) == 0) ? a2 : a1;
					ArrayList<evolution_line> enabled_evolution1 = enabledEVLine.get(agname1);
						for(int k=0; k<enabled_evolution1.size(); k++) {
							evolution_line line1 = enabled_evolution1.get(k);
							ArrayList<logic_expression> synActions2 = new ArrayList<logic_expression>();			
							line1.get_condition().getSynchronizedActions(agname, synActions2);
							if(synActions2.size() == 2) {
								logic_expression act21 = synActions2.get(0);
								logic_expression act22 = synActions2.get(1);
								if((act11.isSameAction(act21) && act12.isSameAction(act22)) ||
										(act11.isSameAction(act22) && act12.isSameAction(act21))) {
									GlobalState tmpstate = gstate.Clone();
									ArrayList<assignment> assignments = line.get_assignments();
									boolean result = executeAssignment(gstate, tmpstate, agname, assignments);
									if(result) {
										GlobalState tmpstate1 = tmpstate.Clone();
										ArrayList<assignment> assignments1 = line1.get_assignments();
										boolean result1 = executeAssignment(tmpstate, tmpstate1, agname1, assignments1);
										if(result1) {
											//ExtendedGlobalState es = new ExtendedGlobalState(tmpstate1, is_evaluation);
											//es.setSynchronized(true);
											//successors.add(es);
											successors.add(tmpstate1);
											enabled_evolution1.remove(k);
										}
									}
								}
							}
						}
				} else {
					GlobalState tmpstate = gstate.Clone();
					ArrayList<assignment> assignments = line.get_assignments();
					boolean result = executeAssignment(gstate, tmpstate, agname, assignments);
					if(result) {
						//ExtendedGlobalState es = new ExtendedGlobalState(tmpstate, is_evaluation);
						//successors.add(es);
						successors.add(tmpstate);
					}
				}
					
			}
		}
		
		return successors;
	}
	
	public ArrayList<GlobalState> por_getInterleavedSuccessors(GlobalState gstate, 
			Hashtable<String, TreeSet<String>> actions) {
		ArrayList<GlobalState> successors = new ArrayList<GlobalState>();
		Hashtable<String, ArrayList<evolution_line>> enabledEVLine = new Hashtable<String, ArrayList<evolution_line>>();
		ArrayList<String> agnames = new ArrayList<String>();
		
		for(int i=0; i<agents.size(); i++) {
			basic_agent agent = agents.get(i);
			ArrayList<evolution_line> evolution = agent.get_evolution();
			if(evolution != null) {
				ArrayList<evolution_line> enabled_evolution = new ArrayList<evolution_line>();
				for(int j=0; j<evolution.size(); j++) {
					evolution_line line = evolution.get(j);
					bool_expression cond = line.get_condition();
					if(cond.evaluate1(agent.get_name(), gstate, actions)) {
						enabled_evolution.add(line);
					}
				}
				enabledEVLine.put(agent.get_name(), enabled_evolution);
				agnames.add(agent.get_name());
			}
		}
		
		for(int i=0; i<agnames.size(); i++) {
			String agname = agnames.get(i);
			ArrayList<evolution_line> enabled_evolution = enabledEVLine.get(agname);
			
			for(int j=0; j<enabled_evolution.size(); j++) {
				evolution_line line = enabled_evolution.get(j);
				ArrayList<logic_expression> synActions1 = new ArrayList<logic_expression>();			
				line.get_condition().getSynchronizedActions(agname, synActions1);
				if(synActions1.size() == 2) {
					logic_expression act11 = synActions1.get(0);
					logic_expression act12 = synActions1.get(1);
					String a1 = ((laction)act11.get_operand(0)).get_agent_name();
					String a2 = ((laction)act12.get_operand(0)).get_agent_name();
					String agname1 = (a1.compareTo(agname) == 0) ? a2 : a1;
					ArrayList<evolution_line> enabled_evolution1 = enabledEVLine.get(agname1);
						for(int k=0; k<enabled_evolution1.size(); k++) {
							evolution_line line1 = enabled_evolution1.get(k);
							ArrayList<logic_expression> synActions2 = new ArrayList<logic_expression>();			
							line1.get_condition().getSynchronizedActions(agname, synActions2);
							if(synActions2.size() == 2) {
								logic_expression act21 = synActions2.get(0);
								logic_expression act22 = synActions2.get(1);
								if((act11.isSameAction(act21) && act12.isSameAction(act22)) ||
										(act11.isSameAction(act22) && act12.isSameAction(act21))) {
									GlobalState tmpstate = gstate.Clone();
									ArrayList<assignment> assignments = line.get_assignments();
									boolean result = executeAssignment(gstate, tmpstate, agname, assignments);
									if(result) {
										GlobalState tmpstate1 = tmpstate.Clone();
										ArrayList<assignment> assignments1 = line1.get_assignments();
										boolean result1 = executeAssignment(tmpstate, tmpstate1, agname1, assignments1);
										if(result1) {
											ExtendedGlobalState es = new ExtendedGlobalState(tmpstate1, is_evaluation);
											es.setSynchronized(true);
											successors.add(es);
											enabled_evolution1.remove(k);
										}
									}
								}
							}
						}
				} else {
					GlobalState tmpstate = gstate.Clone();
					ArrayList<assignment> assignments = line.get_assignments();
					boolean result = executeAssignment(gstate, tmpstate, agname, assignments);
					if(result) {
						ExtendedGlobalState es = new ExtendedGlobalState(tmpstate, is_evaluation);
						successors.add(es);
					}
				}
					
			}
		}
		
		return successors;
	}

	public int BDDlength() {
		int k = 0;
		for(int i=0; i<agents.size(); i++) {
			basic_agent agent = agents.get(i);
			if(agent.obsvars != null)
				for(Enumeration<String> j=agent.obsvars.keys(); j.hasMoreElements(); ) {
					basictype t = agent.obsvars.get(j.nextElement());
					k += t.BDD_length();
				}
			if(agent.vars != null) 
				for(Enumeration<String> j=agent.vars.keys(); j.hasMoreElements(); ) {
					basictype t = agent.vars.get(j.nextElement());
					k += t.BDD_length();
				}
		}
		
		return k;
	}

}

class HashKey {
	public int[] key;
	
	HashKey(int size) {
		key = new int[size];
	}
	
	public String toString() {
		String str = "";
		for(int i=0; i<key.length; i++)
			str += (i>0 ? "." : "") + Integer.toHexString(key[i]);
		return str;
	}
	
	public boolean equals(Object obj) {
		if(this == obj)
			return true;
		if((obj == null) || (obj.getClass() != this.getClass()))
			return false;
		HashKey key1 = (HashKey)obj;
		if(key1.key.length != key.length)
			return false;
		for(int i=0; i<key.length; i++)
			if(key1.key[i]!=key[i])
				return false;
		return true;
	}
	
	public int hashCode() {
		if (key.length == 1)
			return key[0];
		int k = key[0];
		for(int i=1; i<key.length; i++)
			k ^= key[i];
		return k;
	}
}
