package org.mcmas.ui.syntax;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

public class LocalState {
	private Hashtable<String, OneVariable> stateVariables;
	
	public LocalState(Hashtable<String, OneVariable> stateVariables) {
		this.stateVariables = stateVariables;
	}
	
	public LocalState() {
		this.stateVariables = new Hashtable<String, OneVariable>();
	}
	
	public Hashtable<String, OneVariable> getStateVariables() {
		return stateVariables;
	}
	
	public LocalState Clone() {
		Hashtable<String, OneVariable> states = new Hashtable<String, OneVariable>();
		
		for(Enumeration<String> i = stateVariables.keys(); i.hasMoreElements();) {
			String name = i.nextElement();
			OneVariable var = stateVariables.get(name);
			states.put(name, var.Clone());
		}
		
		return new LocalState(states);
	}
	
	/**
	 * Assign a value to a variable. If the same value has already been assigned, nothing needs to be done.
	 * If the variable has been assigned to a different value, a conflict is reported and stateVariable is set to null.
	 * Otherwise, the value is assigned to the variable. 
	 * @param var
	 * @return
	 */
	public boolean addVariable(OneVariable var) {
		if(stateVariables != null) {
			if(stateVariables.containsKey(var.getName())) { // check if the variable has been already assigned a value
				OneVariable oldvar = stateVariables.get(var.getName());
				expression oldValue = oldvar.getValue();
				expression newValue = var.getValue();
				int type = var.getValue().get_type();
				
				// Now we need to check if the old value is equal to the new value
				if(type == 1) { // the variable is a boolean variable
					if(!((bool_value)oldValue).isEqualTo((bool_value)newValue)) {
						// the old value and the new value are not the same, stateVariables is set to null to indicate a conflict
						stateVariables = null;
						return false;
					} 
				} else if(type == 1) { // the variable is a boolean variable
					if(!((bool_value)oldValue).isEqualTo((bool_value)newValue)) {
						// the old value and the new value are not the same, stateVariables is set to null to indicate a conflict
						stateVariables = null;
						return false;
					} 
				} else if(type == 1) { // the variable is a boolean variable
					if(!((bool_value)oldValue).isEqualTo((bool_value)newValue)) {
						// the old value and the new value are not the same, stateVariables is set to null to indicate a conflict
						stateVariables = null;
						return false;
					} 
				}
				// The old and the new values are the same. We don't need to add the new value
				return true;
			} else {
				// The variable is not assigned a value yet. The new value is added.
				stateVariables.put(var.getName(), var);
				return true;
			}
		} else
			return false;
	}
	
	public boolean isVariableAssigned(String name) {
		if(stateVariables.containsKey(name))
			return true;
		else
			return false;
	}
	
	public String toString() {
		String str = "";
		for(Enumeration<String> i=stateVariables.keys(); i.hasMoreElements();) {
			OneVariable var = stateVariables.get(i.nextElement());
			str += var.toString();
		}
		if(Util.isEmpty(str))
			str = "\n";
		return str;
	}
	
	public String toString(ArrayList<String> lvariables) {
		String str = "";
		for(int i=0; i<lvariables.size(); i++) {
			OneVariable var = stateVariables.get(lvariables.get(i));
			str += var.toString1();
		}
		if(Util.isEmpty(str))
			str = "\n";
		return str;
	}
	
	public void toIntArray(int[] source, ArrayList<String> lvariables) {
		for(int i=0; i<lvariables.size(); i++) {
			OneVariable var = stateVariables.get(lvariables.get(i));
			var.toIntArray(source);
		}
	}

	public expression getValue(String varName) {
		OneVariable var = stateVariables.get(varName);
		if(var==null)
			return null;
		return var.getValue();
	}
	
	public void replaceValue(String varName, expression value) {
		OneVariable var = stateVariables.get(varName);
		if(var==null)
			return;
		var.replaceValue(value);
	}
	
	public boolean equalTo(LocalState lstate) {
		if(lstate != null && lstate.getStateVariables() != null && stateVariables != null) {
			Hashtable<String, OneVariable> vars = lstate.getStateVariables();
			if(stateVariables.size() == vars.size()) {
				for(Enumeration<String> i=stateVariables.keys(); i.hasMoreElements();) {
					String vname = i.nextElement();
					OneVariable value = stateVariables.get(vname);
					OneVariable value1 = vars.get(vname);
					if(value1 == null || !value.equalTo(value1))
						return false;
				}
				return true;
			}
		}
		
		return false;
	}
}
