package org.mcmas.ui.syntax;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeSet;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.rules.IToken;
import org.mcmas.ui.editors.IsplWordScanner;
import org.mcmas.ui.editors.Segment;

public class Util {
	private static String[] fgKeywords= { "Action", "Actions", "Agent", "end", "Environment", 
		"Evaluation", "Evolution", "Fairness", "Formulae", "Groups", "InitStates", "Obsvars", 
		"Protocol", "RedStates", "Vars", "Other", "if", "and", "or", "AG", "EG", "AX", "EX", 
		"X", "F", "G", "AF", "EF", "A", "E" , "U", "K", "GK", "GCK", "O", "KH", "DKH", "DK",
		"boolean", "false", "true" };

	public static int log_2(int n) {
		return (int) Math.ceil((Math.log((double) n) / Math.log((double) 2)));
	}

	public static String getErrorHeader(int line, int column, int length) {
		return "line " + line + ":" + column + ":" + length + " ";
	}
	
	public static String global_consistency_check(
			Hashtable<String, basic_agent> is_agents,
			ArrayList<basic_agent> agents,
			Hashtable<String, bool_expression> is_evaluation,
			bool_expression is_istates,
			Hashtable<String, TreeSet<groupitem>> is_groups,
			ArrayList<modal_formula> is_formulae,
			ArrayList<fairness_expression> is_fairness,
			IDocument document){
			//Hashtable<String, Integer> options) {
		String errorMessage = "";
		
		basic_agent is_environment = agents.get(0);
		int count = 0;

		// create global variable table
		Hashtable<String, Hashtable<String, basictype>> global_vars = new Hashtable<String, Hashtable<String, basictype>>();
		Hashtable<String, basictype> en_vars = new Hashtable<String, basictype>();
		Hashtable<String, basictype> enobsvars = is_environment.get_obsvars();
		Hashtable<String, basictype> envars = is_environment.get_vars();
		en_vars.putAll(enobsvars);
		en_vars.putAll(envars);
		global_vars.put(is_environment.get_name(), en_vars);
		for (int i = 1; i < agents.size(); i++) {
			global_vars.put(agents.get(i).get_name(), agents.get(i).get_vars());
		}

		/*
		 * collect enumerate values in observable variables
		 */
		TreeSet<String> obsenum = is_environment.get_obs_enum_values();

		/*
		 * Check consistency in Environment and other agents
		 */
		for (int t = 0; t < agents.size(); t++) {
			basic_agent agent = agents.get(t);
			String agent_name = agent.get_name();
			// Hashtable<String, basictype > agent_variables =
			// global_vars[agent_name];
			// if (options["verbosity"] > 4)
			// System.out.println("Check agent " + agent_name + " ...... ");
			if (t == 0) {
				// check if there is any variable defined in both vars and
				// obsvars
				// if (options["verbosity"] > 4)
				// System.out.println(" checking redefined variables ... ");
				for (Enumeration<String> i = envars.keys(); i.hasMoreElements();) {
					String key = i.nextElement();
					if (enobsvars.containsKey(key)) {// a variable is redefined.
						basictype bt = enobsvars.get(key);
						errorMessage += getErrorHeader(bt.line, bt.column, key.length()) + 
										"variable " + key + " has already been defined.\n";
					}
				}
			}

			// check if a variable has been used as a enumerate value
			/*if (!agent.check_var_against_enum_values(obsenum))
				is_consistent = false; */
			errorMessage += agent.check_var_against_enum_values(obsenum); 
			
			// check red states
			// if (options["verbosity"] > 4)
			// System.out.println(" checking red states ... ");
			bool_expression ag_red = agent.get_redstates();
			/*if (ag_red != null
					&& !ag_red.check_var_and_value(global_vars.get(agent_name)))
				is_consistent = false; */
			if(ag_red != null)
				errorMessage += ag_red.check_var_and_value(global_vars.get(agent_name), document);
			
			// check protocol
			// if (options["verbosity"] > 4)
			// System.out.println(" checking protocol ... ");
			count = 0;
			ArrayList<protocol_line> ag_prot = agent.get_protocol();
			TreeSet<String> ag_act = agent.get_actions();
			for (int i = 0; i < ag_prot.size(); i++) {
				// if (options["verbosity"] > 4)
				// System.out.println(" NO. " + ++count + " ... ");
				TreeSet<groupitem> tmp_act = ag_prot.get(i).get_actions();

				for (Iterator<groupitem> j = tmp_act.iterator(); j.hasNext();) {
					groupitem action = j.next();
					String key = action.name;
					if (!ag_act.contains(key)) {// an action is undefined;
						errorMessage += getErrorHeader(action.line, action.column, action.name.length()) + 
										"action " + key + " is not defined.\n";
					}
				}
				/*if (!(ag_prot.get(i).get_condition()
						.check_var_and_value(global_vars.get(agent_name))))
					is_consistent = false; */
				errorMessage += ag_prot.get(i).get_condition().check_var_and_value(global_vars.get(agent_name), document);
			}
			// check evolution function
			// if (options["verbosity"] > 4)
			// System.out.println(" checking evolution function ...");
			count = 0;
			ArrayList<evolution_line> ag_evol = agent.get_evolution();
			for (int i = 0; i < ag_evol.size(); i++) {
				// if (options["verbosity"] > 4)
				// System.out.println(" NO. " + ++count + ":");
				/* if (ag_evol.get(i).check_double_assignment())
					is_consistent = false;*/
				errorMessage += ag_evol.get(i).check_double_assignment();
				
				// check assignments (new states)
				ArrayList<assignment> va = ag_evol.get(i).get_assignments();
				for (int j = 0; j < va.size(); j++) {
					/* if (!(va.get(j).check_var_and_value(enobsvars, global_vars
							.get(agent_name))))
						is_consistent = false;*/
					errorMessage += va.get(j).check_var_and_value(enobsvars, global_vars.get(agent_name), document);
				}
				// check condition
				/* if (!(ag_evol.get(i).get_condition().check_evolution(enobsvars,
						global_vars.get(agent_name), agent_name, is_agents)))
					is_consistent = false;*/
				errorMessage += ag_evol.get(i).get_condition().check_evolution(enobsvars,
						global_vars.get(agent_name), agent_name, is_agents, document);
			}
		}

		/*
		 * Check group definition
		 */
		// if (options["verbosity"] > 4)
		// System.out.println(endl + "Checking group definition ...");
		for (Enumeration<String> i = is_groups.keys(); i.hasMoreElements();) {
			// if (options["verbosity"] > 4)
			// System.out.println(" group " + (i).first + " ...");
			TreeSet<groupitem> group = is_groups.get(i.nextElement());
			for (Iterator<groupitem> j = group.iterator(); j.hasNext();) {
				boolean flag = false;
				groupitem agent = j.next();
				if (!is_agents.containsKey(agent.name)) {
					errorMessage += getErrorHeader(agent.line, agent.column, agent.name.length()) + 
									"agent " + agent.name + " is not defined.\n";
				}
			}
		}

		/*
		 * Check evaluation function
		 */
		// if (options["verbosity"] > 4)
		// System.out.println(endl + "Checking evaluation function ...");
		for (Enumeration<bool_expression> i = is_evaluation.elements(); i
				.hasMoreElements();) {
			//if (options["verbosity"] > 4) 
			// System.out.println("  atomic proposition " + (i).first + " ...");
			/* if (!i.nextElement().check_global_consistency(global_vars))
				is_consistent = false; */
			errorMessage += i.nextElement().check_global_consistency(global_vars, document);
		}

		/*
		 * Check initstates
		 */
		// if (options["verbosity"] > 4)
		// System.out.println(endl + "Checking initial states ...");
		/* if (!is_istates.check_global_consistency(global_vars))
			is_consistent = false; */
		errorMessage += is_istates.check_global_consistency(global_vars, document);
		/*
		 * Check fairness expression
		 */
		// if (options["verbosity"] > 4)
		// System.out.println(endl + "Checking fairness definition ...");
		count = 0;
		for (int i = 0; i < is_fairness.size(); i++) {
			// if (options["verbosity"] > 4)
			// System.out.println(" fairness expression " + ++count + " ...
			// ");
			/* if (!is_fairness.get(i).check_atomic_proposition(0, is_agents,
					is_evaluation, is_groups))
				is_consistent = false; */
			errorMessage += is_fairness.get(i).check_atomic_proposition(0, is_agents,
					is_evaluation, is_groups);
		}

		/*
		   * add build-in atomic propositions: redstates and greenstates
		   */
		for (int t = 0; t < agents.size(); t++) {
			basic_agent agent = agents.get(t);
			String agent_name = agent.get_name();

		    bool_expression ag_red = agent.get_redstates();
		    String redstates = agent_name + ".RedStates";
		    String greenstates = agent_name + ".GreenStates";
		    if(ag_red!=null) {
		      bool_expression ag_green = new bool_expression(3, ag_red, -1, -1, -1, -1);
		      is_evaluation.put(redstates, ag_red);
		      is_evaluation.put(greenstates, ag_green);
		    } else {
		    	ag_red = new bool_expression(6, null, -1, -1, -1, -1);
		    	bool_expression ag_green = new bool_expression(5, null, -1, -1, -1, -1);
		    	is_evaluation.put(redstates, ag_red);
			    is_evaluation.put(greenstates, ag_green);
		    }
		  }
		  
		/*
		 * Check formulae
		 */
		// if (options["verbosity"] > 4)
		// System.out.println(endl + "Checking formula definition ...");
		count = 0;
		for (int i = 0; i < is_formulae.size(); i++) {
			// if (options["verbosity"] > 4)
			// System.out.println(" formula " + ++count + " ... ");
			/* if (!is_formulae.get(i).check_atomic_proposition(0, is_agents,
					is_evaluation, is_groups))
				is_consistent = false; */
			errorMessage += is_formulae.get(i).check_atomic_proposition(0, is_agents,
					is_evaluation, is_groups);
		}
		return errorMessage;
	}

	public static int getLength(IDocument document, int startline, int startcolumn, int endline, int endcolumn) {
		try{
			int offset1 = document.getLineOffset(startline)+startcolumn;
			int offset2 = document.getLineOffset(endline)+endcolumn;
			return offset2-offset1;
		} catch(BadLocationException e) {
			return 1;
		}
	}
	
	public static boolean isKeyword(String name) {
		for(int i=0; i<fgKeywords.length; i++)
			if(fgKeywords[i].compareTo(name) == 0)
				return true;
		return false;
	}
	public static Segment searchVariable(ArrayList<Segment> children, String name) {
		if(children != null) {
			for(int i=0; i<children.size(); i++) {
				Segment seg = children.get(i);
				int columnoffset = seg.name.indexOf(':');
				String varname = columnoffset==-1? seg.name : seg.name.substring(0, columnoffset-1);
				if(varname.compareTo(name) == 0) {
					return seg;
				}
			}
		}
		return null;
	}
	
	/**
	 * get the first non-space character backwards from the given offset in the document 
	 * @param document
	 * @param offset
	 * @return non-space character or '\0' if no such character
	 */
	public static char getCharBackward(IDocument document, int offset) {
		try {
			while (offset > 0) {
				offset --;
				char c = document.getChar(offset);
				if(!Character.isWhitespace(c))
					return c;
			}
		} catch(BadLocationException e) {
			
		}
		return '\0';
	}

	/**
	 * get the offset of the first non-space character from the given offset in the document
	 * @param document
	 * @param offset
	 * @return the offset of the first non-space character or -1 if no such character
	 */
	public static int getOffsetBackward(IDocument document, int offset) {
		try {
			while (offset > 0) {
				offset --;
				char c = document.getChar(offset);
				if(!Character.isWhitespace(c))
					return offset;
			}
		} catch(BadLocationException e) {
			
		}
		return -1;
	}

	/**
	 * get the offset of the first non-space character and this character must be the one specified by the parameter 
	 * from the given offset in the document
	 * @param document
	 * @param offset
	 * @param character
	 * @return the offset of the character or -1 otherwise
	 */
	public static int getCharOffsetBackward(IDocument document, int offset, int character) {
		try {
			while (offset > 0) {
				offset --;
				char c = document.getChar(offset);
				if(!Character.isWhitespace(c)) {
					if(c == character)
						return offset;
					else
						return -1;
				}
			}
		} catch(BadLocationException e) {
			
		}
		return -1;
	}
	
	/**
	 * get the offset of the first non-space character and this character must be '.' from the given
	 * offset in the document 
	 * @param document
	 * @param offset
	 * @return the offset of '.' or -1 otherwise
	 */
	public static int getDotOffset(IDocument document, int offset) {
		return getCharOffsetBackward(document, offset, '.');
	}
	
	/**
	 * get the offset of the first non-space character and this character must be '=' from the given
	 * offset in the document 
	 * @param document
	 * @param offset
	 * @return  the offset of '=' or -1 otherwise
	 */
	public static int getEqOffset(IDocument document, int offset) {
		return getCharOffsetBackward(document, offset, '=');
	}

	/**
	 * get the position of the first word backwards from the given offset in the document. 
	 * @param document
	 * @param offset
	 * @return the position or null if no such word exists
	 */
	public static Position getWordPositionBackward(IDocument document, int offset) {
		try {
			int length = -1;
			int start = -1;
			while (offset > 0) {
				offset --;
				char c = document.getChar(offset);
				if(!Character.isWhitespace(c)) {
					if(Character.isLetterOrDigit(c) || c == '_') {
						length = 1;
						start = offset;
						break;
					}
				}
			}
			if(length == -1)
				return new Position(offset, 0);
			while (offset > 0) {
				offset --;
				char c = document.getChar(offset);
				if(Character.isLetterOrDigit(c) || c == '_') {
					start = offset;
					length ++;
				} else
					break;
			}
			return new Position(start, length);
		} catch(BadLocationException e) {
			
		}
		return new Position(offset, 0);
	}

	public static Position getCurrentWord(IDocument document, int offset) {
		try {
			char c = document.getChar(offset);
			if(!Character.isLetterOrDigit(c) && c != '_')
				return new Position(offset, 0);
			int length = 1;
			int start = offset;
			int pos = offset;
			while (pos > 0) {
				pos --;
				c = document.getChar(pos);
				if(Character.isLetterOrDigit(c) || c == '_') {
					length++;
					start = pos;
				} else
					break;
			}
			pos = offset;
			while (pos < document.getLength()) {
				pos ++;
				c = document.getChar(pos);
				if(Character.isLetterOrDigit(c) || c == '_') {
					length ++;
				} else
					break;
			}
			return new Position(start, length);
		} catch(BadLocationException e) {
			
		}
		return new Position(offset, 0);
	}
	
	public static Position searchMatchedWord(IDocument document, String name, int offset) {
		try {
			IsplWordScanner WordScanner = new IsplWordScanner(); 
			String key = IsplWordScanner.getKeyWordID(name);
			int line = document.getLineOfOffset(offset);
			int off = -1;
			for(int i=line; i>=0; i--) {
				WordScanner.setRange(document, document.getLineOffset(i), document.getLineLength(i));
				IToken t = WordScanner.nextToken();
				while(!t.isEOF()) {
					if(t.getData()!=null && t.getData().toString().compareTo(key)==0 && WordScanner.getTokenOffset() < offset) {
						off = WordScanner.getTokenOffset();
					}
					t = WordScanner.nextToken();
				}
				if(off >= 0)
					return new Position(off, name.length());
			}
		} catch(BadLocationException e) {
			
		}
		
		return new Position(offset, 0);
	}
	
	public static Position searchMatchedWord(IDocument document, String[] names, int offset) {
		try {
			IsplWordScanner WordScanner = new IsplWordScanner(); 
			String[] keys = new String[names.length];
			for(int i=0; i<keys.length; i++)
				keys[i] = IsplWordScanner.getKeyWordID(names[i]);
			int line = document.getLineOfOffset(offset);
			int off = -1;
			int length = -1;
			for(int i=line; i>=0; i--) {
				WordScanner.setRange(document, document.getLineOffset(i), document.getLineLength(i));
				IToken t = WordScanner.nextToken();
				while(!t.isEOF()) {
					if(t.getData()!=null && WordScanner.getTokenOffset() < offset) {
						for(int j=0; j<keys.length; j++)
							if(t.getData().toString().compareTo(keys[j])==0) {
								off = WordScanner.getTokenOffset();
								length = WordScanner.getTokenLength();
								break;
							}
					}
					t = WordScanner.nextToken();
				}
				if(off >= 0)
					return new Position(off, length);
			}
		} catch(BadLocationException e) {
			
		}
		
		return new Position(offset, 0);
	}
	
	public static int findEndOfWhiteSpace(IDocument document, int offset, int end) throws BadLocationException {
		for(int i=offset; i<end; i++) {
			char c = document.getChar(i);
			if(c!=' ' && c !='\t')
				return i;
		}
		return end;
	}
	
    public static boolean isEmpty(String s) {
        return (s != null && s.length() == 0);
    }

    public static void leftShift(int[] source, int value, int count) {
    	source[source.length-1] <<= count;
    	int k = -1 << (32-count); // what is the size of int in 64 bit machine?
    	for(int i=source.length-2; i>=0; i--) {
    		int t = k & source[i];
    		t >>>= 32-count;
    		source[i+1] |= t;
    		source[i] <<= count;
    	}
    	source[0] |= value;
    }

}
