package org.mcmas.ui.syntax;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeSet;

public class basic_agent {
	String name; // agent name
	Hashtable<String, basictype> obsvars; // observable variables
	Hashtable<String, basictype> vars; // non-observable variables
	bool_expression redstates; // red states
	TreeSet<String> actions; // local actions
	ArrayList<protocol_line> protocol; // local protocol
	ArrayList<evolution_line> evolution; // evolution function
	int action_index_begin;
	int action_index_end;
	Hashtable<String, boolean[]> action_indices; // store index of variables for actions

	public basic_agent(String n, Hashtable<String, basictype> v0,
			Hashtable<String, basictype> v1, bool_expression r,
			TreeSet<String> v2, ArrayList<protocol_line> v3,
			ArrayList<evolution_line> v4) {
		name = n;
		obsvars = v0;
		vars = v1;
		actions = v2;
		protocol = v3;
		evolution = v4;
		redstates = r;
		action_indices = new Hashtable<String, boolean[]>();
		generate_action_bdd_value();
	}

	public String get_name() {
		if (name == null) {
			System.out.println(" error: name in agent is undefined!");
		}
		return name;
	}

	public Hashtable<String, basictype> get_obsvars() {
		return obsvars;
	}

	public Hashtable<String, basictype> get_vars() {
		return vars;
	}

	public TreeSet<String> get_actions() {
		return actions;
	}

	public ArrayList<protocol_line> get_protocol() {
		return protocol;
	}

	public ArrayList<evolution_line> get_evolution() {
		return evolution;
	}

	public bool_expression get_redstates() {
		return redstates;
	}

	public String toString() {
		String str = "Agent " + name + "\n";

		if (obsvars != null && obsvars.size() > 0) {
			str += "Obsvars:\n";
			for (Enumeration<basictype> i = obsvars.elements(); i
					.hasMoreElements();)
				str += "    " + (i).nextElement() + ";\n";
			str += "end Obsvars\n";
		}

		str += "Vars:\n";
		for (Enumeration<basictype> i = vars.elements(); i.hasMoreElements();)
			str += "    " + (i).nextElement() + ";\n";
		str += "end Vars\nRedStates:\n    ";
		if (redstates == null)
			str += "";
		else
			str += redstates.toString() + ";";
		str += "\nend RedStates\nActions = { ";
		int k = 0;
		int bound = actions.size() - 1;
		Iterator<String> i1 = actions.iterator();
		while (i1.hasNext()) {
			String element = (String) i1.next();
			if (k == bound)
				str += element + " ";
			else
				str += element + ", ";
			k++;
		}
		str += "}\nProtocol:\n";
		for (int i = 0; i < protocol.size(); i++)
			str += "    " + protocol.get(i).toString() + ";\n";
		str += "end Protocol\nEvolution:\n";
		for (int i = 0; i < evolution.size(); i++)
			str += "    " + evolution.get(i).toString() + ";\n";
		str += "end Evolution\n";
		str += "end Agent\n";
		return str;
	}

	public int state_BDD_length() {
		int count = 0;
		if (obsvars != null)
			for (Enumeration<basictype> i = obsvars.elements(); i
					.hasMoreElements();)
				count += (i).nextElement().BDD_length();
		if (vars != null)
			for (Enumeration<basictype> i = vars.elements(); i
					.hasMoreElements();)
				count += (i).nextElement().BDD_length();
		return count;
	}

	public int actions_BDD_length() {
		if (actions == null || actions.size() == 0)
			return 0;
		return Util.log_2(actions.size());
	}

	public void set_action_index_begin(int i) {
		action_index_begin = i;
	}

	public void set_action_index_end(int i) {
		action_index_end = i;
	}

	public int get_action_index_begin() {
		return action_index_begin;
	}

	public int get_action_index_end() {
		return action_index_end;
	}

	public void generate_action_bdd_value() {
		int size = actions_BDD_length();
		if (size == 0)
			return;
		boolean[] base = new boolean[size];
		for (int i = 0; i < size; i++)
			base[i] = false;
		Iterator<String> i = actions.iterator();
		while (i.hasNext()) {
			String element = (String) i.next();
			boolean[] temp = new boolean[base.length];
			for (int j = 0; j < base.length; j++)
				temp[j] = base[j];
			action_indices.put(element, temp);
			for (int j = size - 1; j >= 0; j--) {
				if (base[j])
					base[j] = false;
				else {
					base[j] = true;
					break;
				}
			}
		}
	}

	public void print_value_index() {
		for (Enumeration<String> i = action_indices.keys(); i.hasMoreElements();) {
			String key = i.nextElement();
			System.out.print(key + ": ");
			boolean[] value = action_indices.get(key);
			for (int j = 0; j < value.length; j++)
				System.out.print(value[j] ? 1 : 0);
			System.out.println();
		}
	}

	public Hashtable<String, boolean[]> get_action_indices() {
		return action_indices;
	}

	public int allocate_BDD_2_variables(int start) {
		int count = start;
		if (obsvars != null)
			for (Enumeration<String> i = obsvars.keys(); i.hasMoreElements();) {
				basictype value = obsvars.get(i.nextElement());
				int l = value.BDD_length();
				value.set_index_begin(count);
				value.set_index_end(count + l - 1);
				count += l;
			}
		if (vars != null)
			for (Enumeration<String> i = vars.keys(); i.hasMoreElements();) {
				basictype value = vars.get(i.nextElement());
				int l = value.BDD_length();
				value.set_index_begin(count);
				value.set_index_end(count + l - 1);
				count += l;
			}
		return count;
	}

	public int allocate_BDD_2_actions(int start) {
		set_action_index_begin(start);
		int end = start + actions_BDD_length();
		set_action_index_end(end - 1);
		return end;
	}

	public void print_variable_BDD_encoding() {
		if (obsvars != null)
			for (Enumeration<String> i = obsvars.keys(); i.hasMoreElements();) {
				String key = i.nextElement();
				basictype value = obsvars.get(key);
				int begin = value.get_index_begin();
				int end = value.get_index_end();
				System.out.print("----- " + key + ": " + begin + " .. " + end
						+ " -----");
				basictype b = value;
				if (b.get_type() == 2) //rangeint
					((rangedint) b).print_value_index();
				else if (b.get_type() == 3) //enumerate
					((enumerate) b).print_value_index();
				else
					b.print_value_index();
			}
		if (vars != null)
			for (Enumeration<String> i = vars.keys(); i.hasMoreElements();) {
				String key = i.nextElement();
				basictype value = vars.get(key);
				int begin = value.get_index_begin();
				int end = value.get_index_end();
				System.out.print("----- " + key + ": " + begin + " .. " + end
						+ " -----");
				basictype b = value;
				if (b.get_type() == 2) //rangeint
					((rangedint) b).print_value_index();
				else if (b.get_type() == 3) //enumerate
					((enumerate) b).print_value_index();
				else
					b.print_value_index();
			}
	}

	public TreeSet<String> get_obs_enum_values() {
		if (obsvars != null) {
			TreeSet<String> tmpset = new TreeSet<String>();
			for (Enumeration<String> i = obsvars.keys(); i.hasMoreElements();) {
				basictype value = obsvars.get(i.nextElement());
				if (value.get_type() == 3) {
					TreeSet<String> varset = ((enumerate) value)
							.get_enumvalue();
					tmpset.addAll(varset);
				}
			}
			return tmpset;
		}
		return null;
	}

	public String check_var_against_enum_values(TreeSet<String> obsenum) {
		String errorMessage = "";
		TreeSet<String> tmpset = new TreeSet<String>();
		if (vars != null)
			for (Enumeration<String> i = vars.keys(); i.hasMoreElements();) {
				basictype value = vars.get(i.nextElement());
				if (value.get_type() == 3) {
					TreeSet<String> varset = ((enumerate) value)
							.get_enumvalue();
					tmpset.addAll(varset);
				}
			}
		if (obsenum != null)
			tmpset.addAll(obsenum);
		if (obsvars != null)
			for (Enumeration<String> i = obsvars.keys(); i.hasMoreElements();) {
				String key = i.nextElement();
				if (tmpset.contains(key)) {
					basictype bt = obsvars.get(key);
					errorMessage += Util.getErrorHeader(bt.line, bt.column, bt.name.length()) + 
								"variable " + key + " has been used as a enumerate value.\n";
				}
			}
		if (vars != null)
			for (Enumeration<String> i = vars.keys(); i.hasMoreElements();) {
				String key = i.nextElement();
				if (tmpset.contains(key)) {
					basictype bt = obsvars.get(key);
					errorMessage += Util.getErrorHeader(bt.line, bt.column, bt.name.length()) +
									"variable " + key + " has been used as a enumerate value.\n";
				}
			}
		return errorMessage;
	}

}
