package org.mcmas.ui.syntax;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.TreeSet;

import org.eclipse.jface.text.IDocument;

public class bool_expression {
	int op; // 0: logic_expression, 1: and, 2: or, 3: not, 4: other, 5: true, 6: false
	// only the operands[0] is used if op == 0 or 3
	Object[] operands;
	public int startline;
	public int startcolumn;
	public int endline;
	public int endcolumn;

	public bool_expression(int startline, int startcolumn, int endline, int endcolumn) {
		op = 4; // encode "other" branch in protocol
		this.startline = startline;
		this.startcolumn = startcolumn;
		this.endline = endline;
		this.endcolumn = endcolumn;
	}

	public bool_expression(int o, Object obj1, int startline, int startcolumn, int endline, int endcolumn) { // single operand
		op = o;
		operands = new Object[1];
		operands[0] = obj1;
		this.startline = startline;
		this.startcolumn = startcolumn;
		this.endline = endline;
		this.endcolumn = endcolumn;
	}

	public bool_expression(int o, Object obj1, Object obj2, int startline, int startcolumn, int endline, int endcolumn) { // two operands
		op = o;
		operands = new Object[2];
		operands[0] = obj1;
		operands[1] = obj2;
		this.startline = startline;
		this.startcolumn = startcolumn;
		this.endline = endline;
		this.endcolumn = endcolumn;
	}

	public int get_op() {
		return op;
	}

	public boolean is_other_branch() {
		if (op == 4)
			return true;
		else
			return false;
	}

	public Object get_operand(int i) {
		if (i > 1 || (i == 1 && (op == 0 || op == 3))) {
			System.out.println("error: boolean operand " + i
					+ " does not exist!");
		} else if (operands[i] == null) {
			System.out
					.println("error: boolean operand " + i + " is undefined!");
		}
		return (operands[i]);
	}

	public String toString() {
		if (op == 0)
			return ((logic_expression) (operands[0])).toString();
		else if (op == 1)
			return ((bool_expression) (operands[0])).toString() + " and "
					+ ((bool_expression) (operands[1])).toString();
		else if (op == 2)
			return ((bool_expression) (operands[0])).toString() + " or "
					+ ((bool_expression) (operands[1])).toString();
		else if (op == 3)
			return "! " + ((bool_expression) (operands[0])).toString();
		else if (op == 4)
			return "Other";
		  else if(op==5)
			    return "TRUE";
			  else if(op==6)
			    return "FALSE";
		return "error: should not happen.";
	}

	public String check_var_and_value(Hashtable<String, basictype> vars, IDocument document) {
		String tmp = check_null_operand();
		if(!Util.isEmpty(tmp))
			return tmp;
		if (op == 0) // logic_expression
			return ((logic_expression) operands[0]).check_var_and_value(vars, document);
		else if (op == 3) // not operaror
			return ((bool_expression) operands[0]).check_var_and_value(vars, document);
		else if (op == 1 || op == 2) {
			String b1 = ((bool_expression) operands[0])
					.check_var_and_value(vars, document);
			String b2 = ((bool_expression) operands[1])
					.check_var_and_value(vars, document);
			return (b1 + b2);
		} else if (op == 4)
			return "";
		else
			return Util.getErrorHeader(0, 0, 1) + "undefined error in " + toString() + ". report it to the developers.\n";
	}

	public String check_evolution(Hashtable<String, basictype> obsvars,
			Hashtable<String, basictype> vars, String agent_name,
			Hashtable<String, basic_agent> global_actions, IDocument document) {
		String tmp = check_null_operand();
		if(!Util.isEmpty(tmp))
			return tmp;
		if (op == 0) // logic_expression
			return ((logic_expression) operands[0]).check_evolution(obsvars,
					vars, agent_name, global_actions, document);
		else if (op == 3) // not operaror
			return ((bool_expression) operands[0]).check_evolution(obsvars,
					vars, agent_name, global_actions, document);
		else if (op == 1 || op == 2) {
			String b1 = ((bool_expression) operands[0]).check_evolution(
					obsvars, vars, agent_name, global_actions, document);
			String b2 = ((bool_expression) operands[1]).check_evolution(
					obsvars, vars, agent_name, global_actions, document);
			return (b1 + b2);
		} else
			return Util.getErrorHeader(0, 0, 1) + "undefined error in " + toString() + "evolution function of Agent " + agent_name + "\n";
	}

	public String check_global_consistency(Hashtable<String, Hashtable<String, basictype>> vars, IDocument document) {
		String tmp = check_null_operand();
		if(!Util.isEmpty(tmp))
			return tmp;
		if (op == 0) // logic_expression
			return ((logic_expression) operands[0])
					.check_global_consistency(vars, document);
		else if (op == 3) // not operaror
			return ((bool_expression) operands[0])
					.check_global_consistency(vars, document);
		else if (op == 1 || op == 2) {
			String b1 = ((bool_expression) operands[0])
					.check_global_consistency(vars, document);
			String b2 = ((bool_expression) operands[1])
					.check_global_consistency(vars, document);
			return (b1 + b2);
		} else
			return Util.getErrorHeader(0, 0, 1) + "undefined error in " + toString() + ". report it to the developers.\n";
	}

	private String check_null_operand() {
		String tmp = "";

		if (op == 0 || op == 3) {// logic_expression
			if(operands[0] == null)
				tmp = Util.getErrorHeader(startline, startcolumn, endcolumn-startcolumn) + "Null left boolean operand.\n";
		}
		else if (op == 1 || op == 2) {
			if(operands[0] == null)
				tmp = Util.getErrorHeader(startline, startcolumn, endcolumn-startcolumn) + "Null left boolean operand.\n";
			if(operands[1] == null)
				tmp += Util.getErrorHeader(startline, startcolumn, endcolumn-startcolumn) + "Null right boolean operand.\n";
		}
		return tmp;
	}
	
	bool_expression negation_down() {
		if (op == 0) {// logic_expression
			return new bool_expression(3, this, 0, 0, 0, 0);
		} else if (op == 3) // not operaror
			return ((bool_expression) operands[0]).push_negation_inside();
		else if (op == 1) { // and
			bool_expression b1 = ((bool_expression) operands[0])
					.negation_down();
			bool_expression b2 = ((bool_expression) operands[1])
					.negation_down();
			return new bool_expression(2, b1, b2, 0, 0, 0, 0);
		} else { // or
			bool_expression b1 = ((bool_expression) operands[0])
					.negation_down();
			bool_expression b2 = ((bool_expression) operands[1])
					.negation_down();
			return new bool_expression(1, b1, b2, 0, 0, 0, 0);
		}
	}

	bool_expression push_negation_inside() {
		if (op == 0) // logic_expression
			return this;
		else if (op == 3) // not operaror
			return ((bool_expression) operands[0]).negation_down();
		else { // and , or
			bool_expression b1 = ((bool_expression) operands[0])
					.push_negation_inside();
			bool_expression b2 = ((bool_expression) operands[1])
					.push_negation_inside();
			return new bool_expression(op, b1, b2, 0, 0, 0, 0);
		}
	}
	
	public boolean evaluate(String agentName, GlobalState gstate, Hashtable<String, String> actions) {
		if (op == 0) {// logic_expression
			return ((logic_expression)operands[0]).evaluate(agentName, gstate, actions);
		} else if (op == 3) {// not operaror
			return !((bool_expression) operands[0]).evaluate(agentName, gstate, actions);
		}
		else if (op == 1) { // and
			if(!((bool_expression) operands[0]).evaluate(agentName, gstate, actions))
				return false;
			return ((bool_expression) operands[1]).evaluate(agentName, gstate, actions);
		} else { // or
			if(((bool_expression) operands[0]).evaluate(agentName, gstate, actions))
				return true;
			return ((bool_expression) operands[1]).evaluate(agentName, gstate, actions);
		}
	}
	
	public boolean evaluate1(String agentName, GlobalState gstate, Hashtable<String, TreeSet<String>> actions) {
		if (op == 0) {// logic_expression
			return ((logic_expression)operands[0]).evaluate1(agentName, gstate, actions);
		} else if (op == 3) {// not operaror
			return !((bool_expression) operands[0]).evaluate1(agentName, gstate, actions);
		}
		else if (op == 1) { // and
			if(!((bool_expression) operands[0]).evaluate1(agentName, gstate, actions))
				return false;
			return ((bool_expression) operands[1]).evaluate1(agentName, gstate, actions);
		} else { // or
			if(((bool_expression) operands[0]).evaluate1(agentName, gstate, actions))
				return true;
			return ((bool_expression) operands[1]).evaluate1(agentName, gstate, actions);
		}
	}

	public boolean evaluate2(GlobalState gstate) {
		if (op == 0) {// logic_expression
			return ((logic_expression)operands[0]).evaluate2(gstate);
		} else if (op == 3) {// not operaror
			return !((bool_expression) operands[0]).evaluate2(gstate);
		}
		else if (op == 1) { // and
			if(!((bool_expression) operands[0]).evaluate2(gstate))
				return false;
			return ((bool_expression) operands[1]).evaluate2(gstate);
		} else { // or
			if(((bool_expression) operands[0]).evaluate2(gstate))
				return true;
			return ((bool_expression) operands[1]).evaluate2(gstate);
		}
	}

	
	/**
	 * Here we assume only "and" is used in the expression
	 */
	public void getSynchronizedActions(String agname, ArrayList<logic_expression> synActions) {
		if(op == 0) {
			logic_expression l = (logic_expression)operands[0];
			int left = l.operands[0].get_type();
			if(left == 4) {
				synActions.add(l);
			}
		} if(op == 1) {
			bool_expression b1 = (bool_expression)operands[0];
			bool_expression b2 = (bool_expression)operands[1];
			b1.getSynchronizedActions(agname, synActions);
			b2.getSynchronizedActions(agname, synActions);
		}
	}
}
