package org.mcmas.ui.syntax;

import java.util.ArrayList;
import java.util.Iterator;

public class rangedint extends basictype {
	private int lowerbound;
	private int upperbound;
	private int power_two;
	private int half_power;
	private int bdd_length;

	public rangedint(String n, int l, int u, int line, int column) {
		super(n, 2, line, column);
		lowerbound = l;
		upperbound = u;
		bdd_length = Util.log_2(upperbound - lowerbound + 1);
		power_two = 1;
		power_two = power_two << bdd_length;
		half_power = 1;
		half_power = half_power << (bdd_length - 1);
		half_power += lowerbound;
	}

	public int get_lowerbound() {
		return lowerbound;
	}

	public int get_upperbound() {
		return upperbound;
	}

	public boolean is_valid_value(int i) {
		return (i <= upperbound && i >= lowerbound);
	}

	public String toString() {
		return get_name() + ": " + lowerbound + " .. " + upperbound;
	}

	public int BDD_length() {
		return bdd_length;
	}

	public void print_value_index() {
		for (int k = lowerbound; k <= upperbound; k++) {
			System.out.print(k + ": ");
			int[] v = get_value_index(k);
			for (int j = 0; j < v.length; j++)
				System.out.println(((v)[j] == 1 ? "1" : ((v)[j] == -1 ? "0"
						: "_")));
			System.out.println();
		}
	}

	public int[] get_value_index(int value) {
		if (value >= lowerbound && value <= upperbound) {
			//int index_end = get_index_end();
			int[] temp = new int[bdd_length];
			for (int i = 0; i < bdd_length; i++)
				temp[i] = -1;
			if (power_two + lowerbound == upperbound + 1) {
				int distance = value - lowerbound;
				for (int i = bdd_length - 1; i >= 0; i--) {
					if ((distance & 1) == 1)
						(temp)[i] = 1;
					distance >>= 1;
				}
			} else {
				int index = value - half_power;
				if (index < 0) {
					int distance = value - lowerbound;
					for (int i = bdd_length - 2; i >= 0; i--) {
						if ((distance & 1) == 1)
							(temp)[i] = 1;
						distance >>= 1;
					}
					if (value - lowerbound > upperbound - half_power)
						(temp)[bdd_length - 1] = 0;
				} else {
					int distance = index;
					for (int i = bdd_length - 2; i >= 0; i--) {
						if ((distance & 1) == 1)
							(temp)[i] = 1;
						distance >>= 1;
					}
					(temp)[bdd_length - 1] = 1;
				}
			}
			return temp;
		}
		return null;
	}

	public int get_value_index_int(int value) {
		return value-lowerbound;
	}
	
	public int find_value_by_index(boolean[] index) {
		int value = 0;
		if (power_two + lowerbound == upperbound + 1) {
			for (int i = 0; i < index.length; i++) {
				value <<= 1;
				value += (index[i] == true ? 1 : 0);
			}
			return value + lowerbound;
		} else {
			for (int i = 0; i < index.length - 1; i++) {
				value <<= 1;
				value += (index[i] == true ? 1 : 0);
			}
			if (value > upperbound - half_power
					|| index[bdd_length - 1] == false)
				return value + lowerbound;
			else
				return value + half_power;
		}
	}

	/**
	 * Generate logic expressions for all possible values
	 */
	public ArrayList<logic_expression> getLogicExpressions(String agentName) {
		if(lexprs == null) {
			variable var = new variable(agentName, get_name(), 0, 0, 0, 0);
			var.set_var_type(this);
			lexprs = new ArrayList<logic_expression>();
			for(int i = lowerbound; i<=upperbound; i++) {
				int_value b = new int_value(i, 0, 0, 0, 0);
				lexprs.add(new logic_expression(0, var, b));
			}
		}
		return lexprs;
	}
}
