package org.mcmas.ui.wizards;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowData;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.dialogs.ContainerSelectionDialog;

/**
 * The "New" wizard page allows setting the container for the new file as well
 * as the file name. The page will only accept file name without the extension
 * OR with the extension that matches the expected one (ispl).
 */

public class IsplNewFilePage extends WizardPage {
	private Text fileText;

	private Button checkButton;
	private Button checkButton1;
	private Label label;
	
	/**
	 * Constructor for SampleNewWizardPage.
	 * 
	 * @param pageName
	 */
	public IsplNewFilePage() {
		super("wizardPage");
		setTitle("New MCMAS Project");
		setDescription("Create a new file with *.ispl extension that can be opened by a multi-page editor.");
	}

	/**
	 * @see IDialogPage#createControl(Composite)
	 */
	public void createControl(Composite parent) {
		Composite container = new Composite(parent, SWT.NULL);
		GridLayout layout = new GridLayout();
		container.setLayout(layout);
		layout.numColumns = 1;
		layout.verticalSpacing = 9;
		//RowLayout rowLayout = new RowLayout();
		//rowLayout.type = SWT.VERTICAL;
		//rowLayout.marginLeft = 5;
 		//rowLayout.marginRight = 5;		
 		//container.setLayout(rowLayout);
		
 		Point size = parent.getSize();
		
		checkButton1 = new Button(container, SWT.CHECK);
		checkButton1.setText("&Add a new file to the project");
		checkButton1.setSelection(true);
		GridData gd = new GridData(GridData.FILL_HORIZONTAL);
		//gd.horizontalSpan = 2;
		checkButton1.setLayoutData(gd);
		checkButton1.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				//handleBrowse();
				if(checkButton1.getSelection())
					setGroupEnabled(true);
				else
					setGroupEnabled(false);
			}
		});		
		
		//Label label = new Label(container, SWT.NULL);
        //label.setText("");

		Group group = new Group(container, SWT.NULL);
		gd = new GridData(GridData.FILL_HORIZONTAL);
		//gd.horizontalSpan = 2;
		group.setLayoutData(gd);
		layout = new GridLayout();
		group.setLayout(layout);
		layout.numColumns = 2;
		layout.verticalSpacing = 9;
		
		label = new Label(group, SWT.NULL);
		label.setText("&File name:");

		fileText = new Text(group, SWT.BORDER | SWT.SINGLE);
		gd = new GridData(GridData.FILL_HORIZONTAL);
		fileText.setLayoutData(gd);
		fileText.addModifyListener(new ModifyListener() {
			public void modifyText(ModifyEvent e) {
				dialogChanged();
			}
		});
		
		//label = new Label(container, SWT.NULL);
        //label.setText("");

		checkButton = new Button(group, SWT.CHECK);
		checkButton.setText("&Generate the basic framework");
		checkButton.setSelection(true);
		gd = new GridData(GridData.FILL_HORIZONTAL);
		gd.horizontalSpan = 2;
		checkButton.setLayoutData(gd);
		/*checkButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				handleBrowse();
			}
		});*/
		
		initialize();
		dialogChanged();
		setControl(container);
	}

	/**
	 * Tests if the current workbench selection is a suitable container to use.
	 */

	private void initialize() {
		fileText.setText("new_file.ispl");
	}


	/**
	 * Ensures that both text fields are set.
	 */

	private void dialogChanged() {
		String fileName = getFileName();

		if (fileName.length() == 0) {
			updateStatus("File name must be specified");
			return;
		}
		if (fileName.replace('\\', '/').indexOf('/', 1) > 0) {
			updateStatus("File name must be valid");
			return;
		}
		int dotLoc = fileName.lastIndexOf('.');
		if (dotLoc != -1) {
			String ext = fileName.substring(dotLoc + 1);
			if (ext.equalsIgnoreCase("ispl") == false) {
				updateStatus("File extension must be \"ispl\"");
				return;
			}
		}
		updateStatus(null);
	}

	private void updateStatus(String message) {
		setErrorMessage(message);
		setPageComplete(message == null);
	}

	public String getFileName() {
		return fileText.getText();
	}
	
	public boolean getSelection() {
		return checkButton.getSelection();
	}
	
	
	private void setGroupEnabled(boolean flag) {
		label.setEnabled(flag);
		fileText.setEnabled(flag);
		checkButton.setEnabled(flag);
	}
	
	public boolean isAddFileSelected() {
		return checkButton1.getSelection();
	}
	
	public boolean isAddContentSelected() {
		return checkButton.getSelection();
	}
}